/*******************************************************************/
/*      Copyright (C) 1989 by                                      */
/*            The Regents of the University of California          */
/*                                                                 */
/*                             DISCLAIMER                          */
/*                                                                 */
/*      "Although   each   program   has   been  tested  by  its   */
/*      contributor, no warranty, express or implied, is made by   */
/*      the  contributor  or the University of California, as to   */
/*      the accuracy and functioning of the program and  related   */
/*      program material, nor shall the fact of the distribution   */
/*      constitute any such warranty, and no  responsibility  is   */
/*      assumed   by   the  contributor  or  the  University  of   */
/*      California in connection herewith."                        */
/*******************************************************************/

/* BACKUP.C: Main for backup.
 *
 * $Header: /cd/ua/sam/src/ss/backup/RCS/backup.c,v 1.3 89/06/09 16:05:52 sam REL-6-9-89 $
 *
 * $Log:	backup.c,v $
 * Revision 1.3  89/06/09  16:05:52  sam
 * Changed so that output is no less verbose when the chat var is set.
 * 
 * Revision 1.2  89/04/03  21:47:36  sam
 * Changed to ignore a SIGPIPE
 * 
 * Revision 1.1  89/03/08  20:48:57  sam
 * Initial revision
 * 
 */
#ifndef lint
static char rcsid[] = "$Header: /cd/ua/sam/src/ss/backup/RCS/backup.c,v 1.3 89/06/09 16:05:52 sam REL-6-9-89 $";
#endif

#include "../global.h"
#include "backup.h"
#include <sys/param.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/ioctl.h>
#include <sys/mtio.h>
#include <sys/socket.h>
#include <fcntl.h>
#include <stdio.h>
#include <ctype.h>
#include <netdb.h>
#include <setjmp.h>
#include <varargs.h>

/* PROC_STAT: Status about an exited process.
 */
struct proc_stat {
    int pid;
    union wait status;
};

/* Local Vars */
static int
    recov_tape_num,		/* Vars to hold old values in case we need*/
    recov_bytes_left,		/* to return to the beginning of a save */
    recov_file_num,
    read_errs_pid;		/* Pid of the read_errs process */
static jmp_buf
    jumpbuf;			/* Jump buf for returning at end of a save */


/* Global Vars */
int
    nsaves,			/* Total number of saves */
    nsets,			/* Number of save sets */
    nexited,			/* Size of exited[] */
    chat,			/* Chat with an operator? */
    tapefd = -1,		/* File to send tape output to */
    isatape,			/* True iff output is a tape */
    pid,			/* Our pid */
    shell_port,			/* Port num for rsh */
    file_num,			/* File number we're on on tape */
    tape_num,			/* Number of tape in this volume */
    bytes_left,			/* Number of bytes left on tape */
    bytes_per_tape,		/* Number of bytes on a tape */
    blk_size,			/* Block size for writes to tape */
    debug,			/* Debug level */
    dtablesize,			/* Value from getdtablesize() */
    timeout;			/* Time-out (in seconds) for each save */
struct tm
    today;			/* Start time of program */
char
    **sets,			/* Name of save sets */
    *save_format,		/* Format string for savelist name */
    *tapedev,			/* Name of tape device */
    *mailto,			/* Default person to mail to */
    *data_file,			/* Name of dumpdates file */
    *save_list_dir,		/* Directory to put savelists */
    todaystr[30];		/* Today's date in ascii */
struct saverec
    **saves;			/* List of saves to do */
struct proc_stat
    exited[10];			/* Procs that have exited */

/* Function Decls */
extern time_t
    time();
extern char
    *new_str(), *strcpy(), *sprintf(), *strncpy();
union wait
    *has_exited(), *wait_for();
    


/* MAIN: Initialize, run and exit.
 */
main(argc, argv)
    char **argv;
{
    initialize(argc, argv);
    do_all_saves();
    clean_up_and_exit(0);
} /* main */


/* DO_ALL_SAVES: Sort the saves by tape, then start backups on all saves that
 * go to the same tape and same machine.
 */
do_all_saves()
{
    int i, j;
    char *tpname, buf[500];
    int tpname_comp();

    /* Go through and expand all of the tape names */
    for (i = 0; i < nsaves; ++i) {
	build_name(buf, saves[i]->tape_name, saves[i]);
	free(saves[i]->tape_name);
	saves[i]->tape_name = new_str(buf);
	if (saves[i]->tape_name == NULL) {
	    pr("\nOut of memory\n");
	    exit(1);
	}
    }

    /* Sort the saves by tapename */
    qsort((char*)saves, nsaves, sizeof(struct saverec*), tpname_comp);

    /* Go through and call do_one_tape on each set going to the same tape */
    for (i = 0; i < nsaves; i = j) {
	tpname = saves[i]->tape_name;
	for (j = i+1; j <= nsaves; ++j) {
	    if (j == nsaves || strcmp(tpname, saves[j]->tape_name)) {
		do_one_tape(tpname, &saves[i], j - i);
		break;
	    }
	}
    }

    pr("\n** All saves have been completed!! **\n");
} /* do_all_saves */


/* TPNAME_COMP: Compare to sort the saverecs by tape name.
 */
tpname_comp(r1, r2)
    struct saverec **r1, **r2;
{
    return strcmp((*r1)->tape_name, (*r2)->tape_name);
} /* tpname_comp */


/* DO_ONE_TAPE: Do all the saves that go to the same tape.
 */
do_one_tape(tape_name, savv, savc)
    char *tape_name;
    struct saverec **savv;
    int savc;
{
    int i, j, mch_comp();
    struct machrec *mch;

    /* Get the tape mounted */
    if (!debug) {
	if (mount_tape(tape_name, 0) == -1) {
	    pr("\nSkipping saves on the %s tape\n", tape_name);
	    return;
	}
    }

    /* Sort the saves by machine */
    qsort((char*)savv, savc, sizeof(struct saverec*), mch_comp);

    /* Go through and call do_save_set on all saves on the same machine */
    for (i = 0; i < savc; i = j) {
	mch = savv[i]->mach;
	for (j = i+1; j <= savc; ++j) {
	    if (j == savc || mch != savv[j]->mach) {
		do_save_set(mch, &savv[i], j - i);
		break;
	    }
	}
    }

    /* Close the tape */
    if (!debug) if (tp_close() == -1) abort();
} /* do_one_tape */


/* MCH_COMP: Compare two machines for sorting.
 */
mch_comp(r1, r2)
    struct saverec **r1, **r2;
{
    return strcmp((*r1)->mach->name, (*r2)->mach->name);
} /* mch_comp */


/* DO_SAVE_SET: Do the saves coming from the same machine and going to the
 * same tape.  Mach describes the machine, savv contains the saves and savc
 * contains a count of the saves.
 */
do_save_set(mach, savv, savc)
    struct machrec *mach;
    struct saverec **savv;
    int savc;
{
    int i, j, l, gotone;
    time_t tval, begin_time, get_last_save();
    char save_list_file[MAXPATHLEN], rargs[MAXPATHLEN*2], *s;

    /* Change tapes if needed */
    if (!debug && !bytes_left) {
	if (tp_change() == -1) return;
    }

    /* Build the save list file name */
    l = strlen(save_list_dir);
    bcopy(save_list_dir, save_list_file, l);
    save_list_file[l] = '/';
    build_name(save_list_file + l + 1, save_format, savv[0]);

    /* Build the saver command */
    (void) sprintf(rargs, "%s -t %s -x %d %d %d %s", STAND_RCMD, save_list_file,
	    file_num, bytes_left, tape_num-1, savv[0]->tape_name);

    /* Tell op what's going on */
    /* Get a nice date string */
    (void) time(&tval);
    (s = ctime(&tval))[24] = '\0';
    pr("\nSave on host *%s* beginning %s\n", mach->name, s);

    /* Add the paths to the saver command */
    for (i = gotone = 0; i < savc; ++i) {
	if (should_doit(savv[i])) {
	    gotone = 1;
	    for (j = 0; j < savv[i]->npaths; ++j) {
		tval = (savv[i]->type == INC)
			? get_last_save(mach->name, savv[i]->paths[j])
			: (time_t)0;
		(void) sprintf(rargs, "%s -a %d %s",
			       rargs, tval, savv[i]->paths[j]);
		pr("    Doing %s on path %s\n",
		   (savv[i]->type == INC)? "INCREMENTAL" : "FULLSAVE",
		   savv[i]->paths[j]);
	    }
	}
    }

    /* If no saves to do, return */
    if (!gotone || debug) {
	pr("No saves to be done\n");
	return;
    }

    /* Clock beginning of save */
    begin_time = time((time_t*)0);

    /* If not successful, print a message and return */
    if (do_save(mach->name, rargs) == -1) {
	pr("\nSkipping save for %s\n\n", mach->name);
	return;
    }

    /* Update datafile with time of fullsave dumps */
    for (i = 0; i < savc; ++i) {
	if (should_doit(savv[i]) && savv[i]->type == FULL) {
	    for (j = 0; j < savv[i]->npaths; ++j) {
		update_time(mach->name, savv[i]->paths[j], begin_time);
	    }
	}
    }

    /* Tell op this save is done */
    /* Get a nice date string */
    (void) time(&tval);
    (s = ctime(&tval))[24] = '\0';
    pr("    Save completed %s\n", s);
} /* do_save_set */


/* QUIT_HANDLER: Handle a quit signal.
 */
sigtype quit_handler()
{
    char buf[10];

    pr("\nAre you sure you want to abort (y/n)? ");
    input_line(buf, sizeof(buf));
    if (buf[0] != 'y') return;
    fast_copy_cleanup();
    abort();
} /* quit_handler */


/* USR1_HANDLER: Catch a sigusr1 from read_errs and do a longjmp with an
 * error.
 */
sigtype usr1_handler()
{
    longjmp(jumpbuf, CP_ERR_RD+1);
} /* child_handler */


/* ALARM_HANDLER: Called when we've timed out.
 */
sigtype alarm_handler()
{
    if (chat) {
	if (err_timeout() != -1) {
	    pr("\nContinuing with save\n");
	    (void) alarm(0);
	    return;
	}
    }
    longjmp(jumpbuf, CP_ERR_TIME+1);
} /* alarm_handler */


/* DO_SAVE: Executes everything that needs to be done for one save and
 * handles any errors that come up.
 */
do_save(mach_name, saver_args)
    char *mach_name, *saver_args;
{
    int st, sv_out = -1, sv_err, i;

    /* Save recovery info in case of error */
    recov_tape_num = tape_num;
    recov_bytes_left = bytes_left;
    recov_file_num = file_num;

    nexited = read_errs_pid = 0;

    /* Everything returns here when done */
    st = setjmp(jumpbuf);

    /* If returning, st will be real status plus one */
    if (st--) {
	(void) alarm(0);
	(void) close(sv_out);
	if (st == CP_OK && read_errs_pid > 0) {
	    /* Give read_errs a few seconds to exit in case it has an error */
	    (void) signal(SIGUSR1, usr1_handler);
	    for (i = 20; !has_exited(read_errs_pid) && i; i--) sleep(1);
	}
	(void) signal(SIGUSR1, SIG_IGN);
	(void) signal(SIGALRM, SIG_IGN);
	fast_copy_cleanup();
	if (st == CP_OK) return 0;
	if (st == CP_ERR_TIME || handle_err(st) == -1) return -1;
	pr("\nRetrying save on *%s*\n", mach_name);
    }

    /* Start the saver in the background */
    if (start_saver(mach_name, saver_args, &sv_out, &sv_err) == -1) return -1;

    /* Start a background process to read errors from the saver */
    (void) signal(SIGUSR1, usr1_handler);
    read_errs(sv_err);
    (void) close(sv_err);

    /* Install an alarm */
    if (timeout > 0) {
	(void) signal(SIGALRM, alarm_handler);
	(void) alarm((unsigned)timeout);
    }

    /* Copy from saver to tape. */
    longjmp(jumpbuf, do_copy(sv_out)+1);

    /*NOTREACHED*/
} /* do_save */


/* START_SAVER: Run the saver, either remotely, or locally.
 */
start_saver(mach_name, rargs, outfd, errfd)
    char *mach_name, *rargs;
    int *outfd, *errfd;
{
    struct hostent *he, *gethostbyname();
    char hostbuf[MAXHOSTNAMELEN];
    int i;

    /* Get name of host to save on */
    he = gethostbyname(mach_name);
    if (he == NULL) {
	pr("\nUnknown host %s\n", mach_name);
	return -1;
    }

    /* Get the local hostname */
    (void) gethostname(hostbuf, sizeof(hostbuf));

    /* Start up saver either locally or remotely */
    if (!strcmp(hostbuf, he->h_name)) {
	if (do_local_cmd(rargs, outfd, errfd) == -1) return -1;
    } else {
	if (do_rcmd(mach_name, rargs, outfd, errfd) == -1) return -1;
    }

#ifdef  SO_RCVBUF
    /* Increase socket rcv buf size to speed things up */
    for (i = SENDSIZE;
	 setsockopt(*outfd, SOL_SOCKET, SO_RCVBUF, &i, sizeof(i)) < 0 && i > 0;
	 i -= 1024);
#endif
    return 0;
} /* do_saver */


/* DO_LOCAL_CMD: Do a command on this machine.
 */
do_local_cmd(args, outfd, errfd)
    char *args;
    int *outfd, *errfd;
{
    int i, pipe_out[2], pipe_err[2];

    /* Make pipes for stdout and stderr */
    (void) pipe(pipe_out);
    (void) pipe(pipe_err);

    /* Fork off */
    switch(fork()) {
	case -1:
	    prerror("fork");
	    return -1;
	case 0:
	    /* Start up saver */
	    (void) close(0);
	    (void) dup2(pipe_out[W_END], 1);
	    (void) dup2(pipe_err[W_END], 2);
	    for (i = 3; i < dtablesize; ++i) (void) close(i);
	    reset_sigs();
	    exec_str(args);
	    perror("args");
	    (void) write(2, "~1\n", 3);
	    exit(1);
	default:
	    break;
    }

    /* Close unneeded fd's */
    (void) close(pipe_out[W_END]);
    (void) close(pipe_err[W_END]);

    /* Return the stdout and stderr fd's */
    *outfd = pipe_out[R_END];
    *errfd = pipe_err[R_END];

    return 0;
} /* do_local_cmd */


/* DO_RCMD: Do a remote command.
 */
do_rcmd(mach, args, outfd, errfd)
    char *mach, *args;
    int *outfd, *errfd;
{
    char hostbuf[MAXHOSTNAMELEN], *hname = hostbuf;

    /* Copy hostname into buf */
    (void) NCOPY(hostbuf, mach, sizeof(hostbuf));

    *outfd = rcmd(&hname, shell_port, "root", "root", args, errfd);

    if (*outfd < 0 || *errfd < 0) {
	pr("\nCan't do an rsh on %s\n", hname);
	return -1;
    }
    return 0;
} /* do_rcmd */
    

/* DO_COPY: Do the copy from the saver output to the tape.
 */
do_copy(in)
    int in;
{
    int st;

    /* Do the copy */
    st = fast_copy(in);

    /* Adjust bytes_left to be a multiple of blk_size */
    bytes_left -= (bytes_left % blk_size);

    /* Write an eof if no errors */
    if (st == CP_OK) {
	if (isatape && tp_ioctl(MTWEOF, 1) == -1) {
	    st = CP_ERR_WR;
	} else {
	    ++file_num;
	}
    }
    return st;
} /* do_copy */


/* READ_ERRS: Fork off a process that will read the error channel from the
 * remote saver and send the errors to out output.  This process will also exit
 * with the same status as the remote saver.  The function returns the id of
 * this process.
 */
read_errs(sock)
    int sock;
{
    char buf[50];
    int i, n, gottilde = 0, mailfd;

    /* Go into the background */
    switch(read_errs_pid = fork()) {
	case -1:
	    prerror("fork");
	    return;
	case 0:
	    break;
	default:
	    return;
    }

    /* Init */
    reset_sigs();
    (void) close(0);
    mailfd = get_mailfd();
    for (i = 3; i < dtablesize; ++i) {
	if (i != mailfd && i != sock) (void) close(i);
    }

    /* Read something from the socket */
    while ((n = read(sock, buf, sizeof(buf))) > 0) {
	for (i = 0; i < n; ++i) {
	    if (gottilde) {
		/* Got tilde, plus digit, return remote status */
		if (isdigit(buf[i])) {
		    if (i) {
			buf[i-1] = '\0';
			pr("%s", buf);
		    }
		    if (atoi(buf+i)) (void) kill(getppid(), SIGUSR1);
		    exit(0);
		}
		gottilde = 0;
		pr("~");
	    }
	    if (buf[i] == '~') {
		gottilde = 1;
	    }
	}
	buf[n] = '\0';
	pr("%s", buf);
    }
    if (n < 0) prerror("read");
    exit(0);
    /* NOTREACHED */
} /* read_errs */


/* HANDLE_ERR: Handle some error that occurred during the save.  If we can't
 * do it,
 */
handle_err(err_num)
    int err_num;
{
    int choice = CH_ABORT;

    /* Let the operator decide what to do */
    if (chat) {
	switch (err_num) {
	    case CP_ERR_RD:
		choice = err_rd();
		break;
	    case CP_ERR_WR:
		choice = err_wr();
		break;
	    case CP_ERR_OTHER:
		choice = err_other();
		break;
	}
    }
    /* Carry out the operator's choice */
    switch(choice) {
	case CH_NEW:
	    if (tp_change() == -1) abort();
	    break;
	case CH_RETRY:
	    if (recov_pos() == -1) abort();
	    break;
	case CH_SKIP:
	    return -1;
	case CH_ABORT:
	    abort();
    }
    return 0;
} /* handle_err */


/* RECOV_POS: Restore the tape to the position it was in when we started the
 * save by using the recov_* variables.
 */
recov_pos()
{
    /* Can't do this if not using a tape */
    if (!isatape) {
	pr("\nCan't restore tape position on non-tape\n");
	return -1;
    }

    /* If we've changed tapes, go back to the correct tape */
    if (tape_num != recov_tape_num) {
	tape_num = recov_tape_num - 1;
	if (mount_new_tape() == -1) return -1;
    }

    /* Position to the correct file */
    if (tp_ioctl(MTREW, 0) == -1) return -1;
    if (file_num = recov_file_num) {
	if (tp_ioctl(MTFSF, file_num) == -1) return -1;
    }

    /* Set bytes_left to the old value */
    bytes_left = recov_bytes_left;

    return 0;
} /* recov_pos */


/* ERR_RD: Talk to operator about read error.  Return operator's choice.
 */
err_rd()
{
    static int tr[3] = {CH_RETRY, CH_SKIP, CH_ABORT};

    return tr[input_op("\n\
An error occurred while reading the files to be saved.\n\
Type \"retry\" to try this save again.\n\
Type \"skip\" to skip this save\n\
Type \"abort\" to abort all saves\n",
	    "retry", "skip", "abort", 0)];
} /* err_rd */


/* ERR_WR: Talk to operator about write error.  Return operator's choice.
 */
err_wr()
{
    static int tr[3] = {CH_RETRY, CH_NEW, CH_ABORT};

    return tr[input_op("\n\
An error occurred while writing to the tape.\n\
Type \"retry\" to try this save again on the same tape.\n\
Type \"new\" to try this save again on a new tape.\n\
Type \"abort\" to abort all saves\n",
	    "retry", "new", "abort", 0)];
} /* err_wr */


/* ERR_OTHER: Talk to operator about other error.  Return operator's choice.
 */
err_other()
{
    return err_rd();
} /* err_other */


/* ERR_TIMEOUT: Talk to operator about the timeout.  Return -1 if operator
 * wants to kill the job.
 */
err_timeout()
{
    return(input_op("\n\
This save is taking too much time.  This may mean that some part\n\
of the save has failed.\n\
Type \"skip\" to skip this save.\n\
Type \"continue\" to continue with this save.\n",
	    "skip", "continue", 0) ? 0 : -1);
} /* err_timeout */


/* SHOULD_DOIT: Decide whether a save should be done by comparing the current
 * time to the times given by the user in the config file.
 */
should_doit(save)
    struct saverec *save;
{
    return save->days[today.tm_wday];
} /* should_do_it */


/* BUILD_NAME: Make a tapename or savelist name for the given machine
   and save using the passed format.
 */
build_name(buf, fmt, save)
    register char *buf, *fmt;
    struct saverec *save;
{
    char *end;
    int days;
    static char dow[7][10] = {"Sunday", "Monday", "Tuesday", "Wednesday",
	"Thursday", "Friday", "Saturday"};

    end = fmt + strlen(fmt);
    for (; fmt < end; ++fmt) {
	if (*fmt == '%') switch(*++fmt) {
	    /* Hostname */
	    case 'h':
		(void) strcpy(buf, save->mach->name);
		while (*buf) ++buf;
		break;
	    /* Day of week */
	    case 'd':
		bcopy(todaystr, buf, 3);
		buf += 3;
		break;
	    /* Full name for day of the week */
	    case 'D':
		(void) strcpy(buf, dow[today.tm_wday]);
		while (*buf) ++buf;
		break;
	    /* Set name */
	    case 's':
		/* Works through 1999 (at least) */
		days = today.tm_year*365 + (today.tm_year-1)/4 + today.tm_yday;
		(void) strcpy(buf, sets[(days / 7) % nsets]);
		while (*buf) ++buf;
		break;
	    /* Name of the tape */
	    case 't':
		(void) strcpy(buf, save->tape_name);
		while (*buf) ++buf;
		break;
	    default:
		*buf++ = *fmt;
	} else {
	    *buf++ = *fmt;
	}
    }
    *buf = '\0';
} /* build_name */


/* CLEAN_UP_AND_EXIT: Call this to quit the program
 */
clean_up_and_exit(status)
{
    if (read_errs_pid > 0) (void) kill(read_errs_pid, 9);
    setup_mail((char*)0);
    pr("\nClosing tape drive (should take less than 5 minutes)....\n");
    (void) tp_close();
    exit(status);
} /* clean_up_and_exit */
    

/* ABORT: Call this to abort everything.
 */
abort()
{
    pr("\n\n*************** SAVE ABORTED ******************\n");
    clean_up_and_exit(1);
} /* abort */


/* EXEC_STR: Exec a command given by a string.
 */
exec_str(str)
    char *str;
{
    int argc;
    char *argv[200];

#define isspc(c) ((c) == ' ' || (c) == '\t')
    /* Parse it up */
    for (argc = 0; *str != '\n' && *str != '\0';) {
	if (isspc(*str)) {
	    *str = '\0';
	    do { ++str; } while (isspc(*str));
	} else {
	    argv[argc++] = str;
	    do { ++str; } while (!isspc(*str) && *str && *str != '\n');
	}
    }
    *str = '\0';
    argv[argc] = NULL;

    (void) execv(argv[0], argv);
} /* exec_str */


/* RESET_SIGS: Set all "important" sigs to the values we want.  This should
 * be called by the child after a fork.
 */
reset_sigs()
{
    (void) signal(SIGHUP, SIG_IGN);
    (void) signal(SIGINT, SIG_IGN);
    (void) signal(SIGQUIT, SIG_IGN);
    (void) signal(SIGTERM, SIG_DFL);
    (void) signal(SIGALRM, SIG_IGN);
    (void) signal(SIGUSR1, SIG_IGN);
    (void) signal(SIGPIPE, SIG_IGN);
} /* reset_sigs */


/* HAS_EXITED: Is proc in the exited list.
 */
union wait *has_exited(proc)
    int proc;
{
    int i;

    for (i = 0; i < nexited; ++i) {
	if (exited[i].pid == proc) return &exited[i].status;
    }
    while ((i = wait3(&exited[nexited].status, WNOHANG, (struct rusage*)0)) > 0)
    {
	exited[nexited++].pid = i;
	if (i == proc) return &exited[nexited-1].status;
    }
    return (union wait*)0;
} /* has_exited */


/* WAIT_FOR: Do a virtual wait for some proc.  Return the status buf.
 */
union wait *wait_for(proc)
    int proc;
{
    int i;

    for (i = 0; i < nexited; ++i) {
	if (exited[i].pid == proc) return &exited[i].status;
    }
    for (;;) {
	i = wait(&exited[nexited].status);
	exited[nexited++].pid = i;
	if (i == proc) return &exited[nexited-1].status;
    }
} /* wait_for */
