/*******************************************************************/
/*      Copyright (C) 1989 by                                      */
/*            The Regents of the University of California          */
/*                                                                 */
/*                             DISCLAIMER                          */
/*                                                                 */
/*      "Although   each   program   has   been  tested  by  its   */
/*      contributor, no warranty, express or implied, is made by   */
/*      the  contributor  or the University of California, as to   */
/*      the accuracy and functioning of the program and  related   */
/*      program material, nor shall the fact of the distribution   */
/*      constitute any such warranty, and no  responsibility  is   */
/*      assumed   by   the  contributor  or  the  University  of   */
/*      California in connection herewith."                        */
/*******************************************************************/

/* FAST_COPY.C: Do a fast copy from saver output to tape.
 *
 * $Header: /cd/ua/sam/src/ss/backup/RCS/fast_copy.c,v 1.3 89/06/09 16:07:41 sam REL-6-9-89 $
 *
 * $Log:	fast_copy.c,v $
 * Revision 1.3  89/06/09  16:07:41  sam
 * No changes.
 * 
 * Revision 1.2  89/04/03  21:48:49  sam
 * *** empty log message ***
 * 
 * Revision 1.1  89/03/08  20:49:28  sam
 * Initial revision
 * 
 */
#ifndef lint
static char rcsid[] = "$Header: /cd/ua/sam/src/ss/backup/RCS/fast_copy.c,v 1.3 89/06/09 16:07:41 sam REL-6-9-89 $";
#endif

#include "../global.h"
#include "backup.h"
#include <stdio.h>
#include <sys/file.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <signal.h>

/* Pipe ends */
#define R_END 0
#define W_END 1

/* Pipe names */
#define RDLK	0
#define WRLK	1
#define SHARE	2

/* SHAR: Info to share between procs.
 */
struct _shar {
    int left;
};

/* Local vars */
static char
    *buf;		/* Buffer to hold i/o block */
static int
    num_in_buf,		/* Number of bytes in buf */
    cpid,		/* Child's pid */
    pipes[3][2];	/* Pipes!! */
static struct _shar
    shar;		/* Shared structure */

/* Externs */
extern int
    dtablesize,		/* Value from getdtablesize */
    tapefd,		/* File tape output is sent to (used indirectly) */
    blk_size,		/* Block size */
    bytes_left;		/* Number of bytes left on tape */

/* Function Decls */
extern char
    *mktemp(), *malloc();
extern int
    mount_new_tape();
extern union wait
    *has_exited(), *wait_for();


/* FAST_COPY: Copy from the fd in to the tape.
 */
int fast_copy(in)
    int in;
{
    int i, j, status, mailfd;

    /* Make buffer */
    buf = NEW(char, blk_size);
    if (buf == NULL) {
	pr("Out of memory\n");
	return CP_ERR_OTHER;
    }

    do {
	/* Make all the pipes */
	for (i = 0; i < 3; ++i) {
	    if (pipe(pipes[i]) == -1) {
		prerror("pipe");
		free(buf);
		return CP_ERR_OTHER;
	    }
	}

	/* Set up shared info and locks */
	shar.left = bytes_left;
	send_shared();
	give_lock(RDLK, CP_CONT);
	give_lock(WRLK, CP_CONT);

	/* Fork off and do the copy */
	switch(cpid = fork()) {
	    case -1:
		perror("fork");
		break;
	    case 0:
		mailfd = get_mailfd();
		for (i = 0; i < dtablesize; ++i) {
		    if (i != tapefd && i != in && i != mailfd) {
			(void) close(i);
		    }
		}
		reset_sigs();
		break;
	    default:
		break;
	}
	status = copy_loop(in);

	/* Send the returned status to the other process */
	give_lock(RDLK, status);
	give_lock(WRLK, status);

	/* Child dies here. After this, only the parent is left.
         */
	if (cpid == 0) exit(0);

	/* Wait for the child */
	if (cpid > 0) {
	    (void) wait_for(cpid);
	}

	/* Copy out new value of bytes_left */
	bytes_left = shar.left;

	for (i = 0; i < 3; ++i) for (j = 0; j < 2; ++j) {
	    (void) close(pipes[i][j]);
	    pipes[i][j] = 0;
	}

	if (status == CP_CHANGE) {
	    if (tp_change() == -1) {
		status = CP_ERR_OTHER;
	    }
	}
    } while (status == CP_CHANGE);

    free(buf);
    buf = NULL;
    cpid = 0;
    return status;
} /* fast_copy */


/* FAST_COPY_CLEANUP: Clean things up if we're interrupted in the middle of a
 * save.
 */
fast_copy_cleanup()
{
    int i, j;

    if (buf) free(buf);
    buf = NULL;
    /* Give the child 30 seconds to exit before killing it */
    if (cpid > 0) {
	for (i = 30; !has_exited(cpid) && i; --i) sleep(1);
        if (!has_exited(cpid)) (void) kill(cpid, 9);
    }
    for (i = 0; i < 3; ++i) for (j = 0; j < 2; ++j) {
	if (pipes[i][j]) (void) close(pipes[i][j]);
	pipes[i][j] = 0;
    }
} /* fast_copy_cleanup */


/* COPY_LOOP: Control reads and writes
 */
static int copy_loop(in)
    int in;
{
#define GET_A_LOCK(what) st1 = get_lock(what); if (st1 != CP_CONT) return st1
    int st1, st;

    for (;;) {
	GET_A_LOCK(RDLK);		/* Get perm to read */
	get_shared();			/* Get shared info */
	st = do_read(in);		/* Read */
	send_shared();			/* Send shared info to next proc */
	if (!num_in_buf) return st;
	GET_A_LOCK(WRLK);		/* Get perm to write */
	give_lock(RDLK, st);		/* Let next proc read */
	st = do_write();		/* Write */
	if (st != CP_CONT) return st;
	give_lock(WRLK, CP_CONT);	/* Let next proc write */
    }
} /* copy_loop */


/* GIVE_LOCK: Give away a lock.
 */
give_lock(which, code)
    int which, code;
{
    char c = code;

    (void) write(pipes[which][W_END], &c, 1);
} /* give_lock */


/* GET_LOCK: Get a lock.
 */
int get_lock(which)
    int which;
{
    char c;

    (void) read(pipes[which][R_END], &c, 1);
    return c;
} /* get_lock */


/* GET_SHARED: Get the shared struct from the pipe.
 */
static get_shared()
{
    (void) read_all(pipes[SHARE][R_END], (char*)&shar, sizeof(shar));
} /* get_shared */


/* SEND_SHARED: Send the shared struct across the pipe.
 */
static send_shared()
{
    (void) write_all(pipes[SHARE][W_END], (char*)&shar, sizeof(shar));
} /* send_shared */


/* DO_READ: Read in one block.
 */
static int do_read(in)
    int in;
{
    int n, res, lft = blk_size;
    char *bf = buf;

    num_in_buf = 0;
    while (lft) {
	if (shar.left == 0) return CP_CHANGE;
	n = min(lft, shar.left);
	res = read(in, bf, n);
	switch(res) {
	    case -1:
		prerror("read");
		return CP_ERR_RD;
	    case 0:
		return CP_OK;
	    default:
		if (res > n) return CP_ERR_RD;
		bf += res;
		lft -= res;
		shar.left -= res;
		num_in_buf += res;
		break;
	}
    }
    return CP_CONT;
} /* do_read */


/* DO_WRITE: Write out one block.
 */
static int do_write()
{
    if (tp_write(buf, num_in_buf) != num_in_buf) return CP_ERR_WR;
    return CP_CONT;
} /* do_write */
 

/* READ_ALL: Like read(2), but read until everything is in.
 */
read_all(fd, buf, nbytes)
    int fd, nbytes;
    char *buf;
{
    int retval = nbytes, res;

    while (nbytes) {
        res = read(fd, buf, nbytes);
        if (res == -1 || res == 0) {
            if (res == -1) {
                prerror("read");
            } else {
                pr("eof on read");
            }
            return -1;
        }
        nbytes -= res;
        buf += res;
    }
    return retval;
} /* read_all */


/* WRITE_ALL: Like write(2), but write until everything is out.
 */
write_all(fd, buf, nbytes)
    int fd, nbytes;
    char *buf;
{
    int retval = nbytes, res;

    while (nbytes) {
        res = write(fd, buf, nbytes);
        if (res == -1 || res == 0) {
            if (res == -1) {
                prerror("write");
            } else {
                pr("eof on write");
            }
            return -1;
        }
        nbytes -= res;
        buf += res;
    }
    return retval;
} /* write_all */


#ifdef TEST

int outfd, bytes_left, blk_size = 64512;

/* Code to test fast_copy.
 */


tp_write(buf, size)
    char *buf;
    int size;
{
    return write_all(outfd, buf, size);
} /* tp_write */


tp_change()
{
    char buf[10];

    pr("Change the tape:");
    input_line(buf, sizeof(buf));
    bytes_left = 64512*30000;
    return 0;
} /* tp_change */


child_handler()
{
    (void) wait((union wait*)0);
} /* child_handler */


main(argc, argv)
    char **argv;
{
    int infd;

    bytes_left = 30000*64512;

    if (argc < 3) usage();

    if (!strcmp(argv[1], "-")) {
	infd = 0;
    } else {
	infd = open(argv[1], O_RDONLY);
	if (infd == -1) {
	    prerror(argv[1]);
	    exit(1);
	}
    }

    if (!strcmp(argv[2], "-")) {
	outfd = 1;
    } else {
	outfd = open(argv[2], O_WRONLY|O_CREAT|O_TRUNC, 0666);
	if (outfd == -1) {
	    prerror(argv[2]);
	    exit(1);
	}
    }

    if (argc > 3) {
	if (argv[3][0] < '0' || argv[3][0] > '9') usage();
	bytes_left = atoi(argv[3]);
    }

    signal(SIGCHLD, child_handler);
    pr("fast_copy returns %d\n", fast_copy(infd, 64512));

    close(outfd);
    close(infd);
}

usage()
{
    pr("Usage: fast_copy in out [bytes_left]\n");
    exit(1);
}
#endif TEST
