/*******************************************************************/
/*      Copyright (C) 1989 by                                      */
/*            The Regents of the University of California          */
/*                                                                 */
/*                             DISCLAIMER                          */
/*                                                                 */
/*      "Although   each   program   has   been  tested  by  its   */
/*      contributor, no warranty, express or implied, is made by   */
/*      the  contributor  or the University of California, as to   */
/*      the accuracy and functioning of the program and  related   */
/*      program material, nor shall the fact of the distribution   */
/*      constitute any such warranty, and no  responsibility  is   */
/*      assumed   by   the  contributor  or  the  University  of   */
/*      California in connection herewith."                        */
/*******************************************************************/

/* FILEIO.C: Routines to manipulating the dumpdates file.
 *
 * $Header: /cd/ua/sam/src/ss/backup/RCS/fileio.c,v 1.3 89/06/09 16:08:24 sam REL-6-9-89 $
 *
 * $Log:	fileio.c,v $
 * Revision 1.3  89/06/09  16:08:24  sam
 * No changes.
 * 
 * Revision 1.2  89/04/03  21:48:59  sam
 * No changes
 * 
 * Revision 1.1  89/03/08  20:49:34  sam
 * Initial revision
 * 
 */
#ifndef lint
static char rcsid[] = "$Header: /cd/ua/sam/src/ss/backup/RCS/fileio.c,v 1.3 89/06/09 16:08:24 sam REL-6-9-89 $";
#endif

#include "../global.h"
#include "backup.h"
#include <sys/types.h>
#include <sys/param.h>
#include <stdio.h>

/* Externs */
extern int
    pid;		/* Our process id */
extern char
    *data_file;		/* Name of dumpdates files */

/* Function Decls */
extern char
    *sprintf();


/* GET_LAST_SAVE: Read the data file and return the last time this path on
 * this host was dumped.  If datafile doesn't exist, or partition not found then
 * return 0.
 */
time_t get_last_save(host, path)
    char *host, *path;
{
    FILE *datafp;
    char rdhost[MAXHOSTNAMELEN], rdpath[MAXPATHLEN];
    time_t rdtime;

    datafp = fopen(data_file, "r");
    if (datafp == NULL) {
	return (time_t)0;
    }

    while (read_data_line(datafp, rdhost, rdpath, &rdtime)) {
	if (!strcmp(rdhost, host) && !strcmp(rdpath, path)) {
	    (void) fclose(datafp);
	    return rdtime;
	}
    }
    (void) fclose(datafp);
    return (time_t) 0;
} /* get_last_save */


/* UPDATE_TIME: Update the last time of a save for the path on the given host
 */
update_time(host, path, time)
    char *host, *path;
    time_t time;
{
    FILE *datafp, *tmpfp;
    char tmpfilename[MAXPATHLEN];
    char rdhost[MAXHOSTNAMELEN], rdpath[MAXPATHLEN];
    time_t rdtime;
    int wrote_it = 0;

    (void) sprintf(tmpfilename, "%s.%d", data_file, pid);
    if ((tmpfp = fopen(tmpfilename, "w")) == NULL) {
	pr("Can't write to temp file %s\n", tmpfilename);
	return;
    }

    if ((datafp = fopen(data_file, "r")) == NULL) {
	pr("Warning: creating a new ");
	prerror(data_file);
    } else {
	/* Update old entry if found, copy over all others */
	while (read_data_line(datafp, rdhost, rdpath, &rdtime)) {
	    if (!strcmp(rdhost, host) && !strcmp(rdpath, path)) {
		write_data_line(tmpfp, host, path, time);
		++wrote_it;
	    } else {
		write_data_line(tmpfp, rdhost, rdpath, rdtime);
	    }
	}
    }

    /* Adding a new entry */
    if (!wrote_it) {
	write_data_line(tmpfp, host, path, time);
    }

    (void) fclose(tmpfp);
    if (datafp) (void) fclose(datafp);

    if (rename(tmpfilename, data_file) == -1) {
	prerror("update_time: rename");
    }
} /* update_time */


/* READ_DATA_LINE: Read in one data line, break it into host, path and time,
 * and return.  If at end of file or error, returns 0, else 1.
 */
read_data_line(fp, host, path, time)
    FILE *fp;
    char *host, *path;
    time_t *time;
{
    switch(fscanf(fp, " %[^: \t\n] : %s %ld\n", host, path, time)) {
	case EOF:
	    return 0;
	case 0:
	    if (!feof(fp)) {
		pr("Error reading data file\n");
	    }
	    return 0;
	default:
	    return 1;
    }
} /* read_data_line */


/* WRITE_DATA_LINE: Write out one data line to the data file.
 */
write_data_line(fp, host, path, time)
    FILE *fp;
    char *host, *path;
    time_t time;
{
    char s[MAXPATHLEN+MAXHOSTNAMELEN+2];
    int pad;

    (void) sprintf(s, "%s:%s", host, path);
    fputs(s, fp);
    (void) putc(' ', fp);
    for (pad = 39 - strlen(s); pad > 0; --pad)
	(void) putc(' ', fp);
    (void) fprintf(fp, "%ld\n", time);
}
