/*******************************************************************/
/*      Copyright (C) 1989 by                                      */
/*            The Regents of the University of California          */
/*                                                                 */
/*                             DISCLAIMER                          */
/*                                                                 */
/*      "Although   each   program   has   been  tested  by  its   */
/*      contributor, no warranty, express or implied, is made by   */
/*      the  contributor  or the University of California, as to   */
/*      the accuracy and functioning of the program and  related   */
/*      program material, nor shall the fact of the distribution   */
/*      constitute any such warranty, and no  responsibility  is   */
/*      assumed   by   the  contributor  or  the  University  of   */
/*      California in connection herewith."                        */
/*******************************************************************/

/* INIT.C: Initialization Routines.
 *
 * $Header: /cd/ua/sam/src/ss/backup/RCS/init.c,v 1.3 89/06/09 16:08:49 sam REL-6-9-89 $
 *
 * $Log:	init.c,v $
 * Revision 1.3  89/06/09  16:08:49  sam
 * No changes.
 * 
 * Revision 1.2  89/04/03  21:49:06  sam
 * No changes
 * 
 * Revision 1.1  89/03/08  20:49:46  sam
 * Initial revision
 * 
 */
#ifndef lint
static char rcsid[] = "$Header: /cd/ua/sam/src/ss/backup/RCS/init.c,v 1.3 89/06/09 16:08:49 sam REL-6-9-89 $";
#endif

#include "../global.h"
#include "backup.h"
#include <sys/types.h>
#include <sys/file.h>
#include <stdio.h>
#include <time.h>
#include <netdb.h>
#include <sys/stat.h>
#include <signal.h>

/* Local vars */
static char
    *conf_file;	/* Name of config file */

/* Externs */
extern int
    pid,		/* Our process id */
    shell_port,		/* Port num for rsh */
    blk_size,		/* Block size */
    chat,		/* Talk to the operator? */
    isatape,		/* Set this if output is a tape */
    bytes_per_tape,	/* Num of bytes per tape */
    debug,		/* Debug level */
    dtablesize;		/* Filde descriptor table size */
extern struct tm
    today;		/* Start time of save */
extern char
    *tapedev,		/* Name of tape device */
    *mailto,		/* Name of person to mail to */
    *save_format,	/* Format string to form a save name */
    *data_file,		/* Name of dumpdates file */
    *save_list_dir,	/* Save list directory */
    todaystr[];		/* Today's date in ascii */
extern FILE
    *yyin;		/* Input file for parser */

/* Function decls */
extern char
    *strcpy();
extern sigtype
    quit_handler();

    

/* INITIALIZE: Set up vars, read command line and config file.
 */
initialize(argc, argv)
    char **argv;
{
    struct servent *sp, *getservbyname();
    time_t tdy, time();
    struct stat stbuf;

    /* Miscellaneous stuff */
    dtablesize = getdtablesize();
    pid = getpid();
    (void) umask(022);
    tdy = time((time_t*)0);
    bcopy((char*)localtime(&tdy), (char*)&today, sizeof(today));
    (void) strcpy(todaystr, asctime(&today));
    chat = isatty(0) && isatty(1);

    /* Parse command line */
    parse_args(argc, argv);

    /* Get shell port.
    /* For some reason this is very inefficient on the sun4.  If it really
     * bogs down, maybe shell_port should just be a constant.
     */
    if (sp = getservbyname("shell", "tcp")) {
	shell_port = sp->s_port;
    } else {
	pr("Can't getservbyname(shell,tcp)");
    }

    /* Read config file */
    if ((yyin = fopen(conf_file, "r")) == NULL) {
	pr("Can't read config file ");
	prerror(conf_file);
	exit(1);
    }
    (void) yyparse();
    (void) fclose(yyin);

    /* Variables not set on command line and not set in config file get set
     * here (if needed)
     */
    if (!tapedev) tapedev = DEFTAPEDEV;
    if (!data_file) data_file = DATAFILE;
    if (!save_format) save_format = DEF_SAVE_FORMAT;
    if (!save_list_dir) save_list_dir = SAVELISTDIR;
    if (!blk_size) blk_size = DEF_BLK_SIZE;
    if (!bytes_per_tape) bytes_per_tape = DEF_BYTES_PER_TAPE;

    /* Check that bytes_per_tape is a multiple of blk_size */
    if (bytes_per_tape % blk_size) {
	pr("Truncating bytes per tape to be a multiple of the block size\n");
	bytes_per_tape -= (bytes_per_tape % blk_size);
    }

    /* Set isatape if output is a tape */
    if (access(tapedev, W_OK) == -1 || stat(tapedev, &stbuf)) {
	prerror(tapedev);
	exit(1);
    }
    isatape = (stbuf.st_mode & S_IFMT) == S_IFCHR;

    /* Set up signals */
    reset_sigs();
    (void) signal(SIGINT, quit_handler);

    /* Start up the mailer */
    if (mailto) setup_mail(mailto);

    /* Setup for doing output */
    setlinebuf(stdout);
} /* initialize */


/* PARSE_ARGS: Parse up the command line.
 */
parse_args(argc, argv)
    char **argv;
{
    char *opt;

    tapedev = NULL;
    mailto = NULL;
    data_file = NULL;
    save_format = NULL;
    conf_file = CONFIGFILE;
    blk_size = 0;
    bytes_per_tape = 0;
    debug = 0;

    if (argc == 1) return;

    opt = &argv[1][0];
    if (*opt == '-') ++opt;
    argc -= 2;
    argv += 2;

    while (*opt) switch(*opt++) {
	case 'f':
	    if (!argc--) usage();
	    tapedev = *argv++;
	    break;
	case 'm':
	    if (!argc--) usage();
	    mailto = *argv++;
	    break;
	case 'c':
	    if (!argc--) usage();
	    conf_file = *argv++;
	    break;
	case 'F':
	    if (!argc--) usage();
	    data_file = *argv++;
	    break;
	case 's':
	    if (!argc--) usage();
	    save_format = *argv++;
	    break;
	case 'l':
	    if (!argc--) usage();
	    save_list_dir = *argv++;
	    break;
	case 'b':
	    if (!argc--) usage();
	    blk_size = atoi(*argv++);
	    break;
	case 'd':
	    if (!argc--) usage();
	    bytes_per_tape = atoi(*argv++);
	    break;
	case 'D':
	    debug = 1;
	    break;
	default:
	    usage();
    }
} /* parse_args */


/* USAGE: Print a nasty message and exit.
 */
usage()
{
    pr("\
usage: backup [-][fmcFslbd] [tapedev] [mailto] [config-file] [dump-file]\n\
       [save-name] [savelist-dir] [block-size] [bytes-per-tape]\n");
    exit(1);
} /* usage */
