/*******************************************************************/
/*      Copyright (C) 1989 by                                      */
/*            The Regents of the University of California          */
/*                                                                 */
/*                             DISCLAIMER                          */
/*                                                                 */
/*      "Although   each   program   has   been  tested  by  its   */
/*      contributor, no warranty, express or implied, is made by   */
/*      the  contributor  or the University of California, as to   */
/*      the accuracy and functioning of the program and  related   */
/*      program material, nor shall the fact of the distribution   */
/*      constitute any such warranty, and no  responsibility  is   */
/*      assumed   by   the  contributor  or  the  University  of   */
/*      California in connection herewith."                        */
/*******************************************************************/

/* OPIO.C: Routines to do I/O with the operator
 *
 * $Header: /cd/ua/sam/src/ss/backup/RCS/opio.c,v 1.3 89/06/09 16:09:55 sam REL-6-9-89 $
 *
 * $Log:	opio.c,v $
 * Revision 1.3  89/06/09  16:09:55  sam
 * Do a perror if exec_str returns.
 * 
 * Revision 1.2  89/04/03  21:49:50  sam
 * No longer close stdout and stderr when forking off mail program.
 * 
 * Revision 1.1  89/03/08  20:50:02  sam
 * Initial revision
 * 
 */
#ifndef lint
static char rcsid[] = "$Header: /cd/ua/sam/src/ss/backup/RCS/opio.c,v 1.3 89/06/09 16:09:55 sam REL-6-9-89 $";
#endif

#include "../global.h"
#include "backup.h"
#include <sys/param.h>
#include <varargs.h>
#include <stdio.h>
#include <fcntl.h>

/* Locals */
static FILE
    *mailfp;			/* Fp to send mail to */

/* Externs */
extern int
    dtablesize,			/* Value from getdtablesize */
    errno;
extern char
    *sys_errlist[];

/* Funciton Decls */
extern char
    *sprintf(), *strncpy();


/* PR: Tell the operator something.  Uses same args as printf.
 */
/*VARARGS0*/
pr(va_alist)
    va_dcl
{
    va_list args;
    char *fmt;

    va_start(args);
    fmt = va_arg(args, char *);
#ifdef sequent
    (void) _doprnt(fmt, args, stdout);
    if (mailfp) (void) _doprnt(fmt, args, mailfp);
#else
    (void) vfprintf(stdout, fmt, args);
    if (mailfp) (void) vfprintf(mailfp, fmt, args);
#endif
    va_end(args);
} /* pr */


/* INPUT_OP: Get a response from the operator.
 */
/*VARARGS0*/
input_op(va_alist)
    va_dcl
{
    va_list args;
    char buf[200], *argv[20], *prompt;
    int i, argc;

    va_start(args);
    prompt = va_arg(args, char *);
    for (argc = 0; argv[argc] = va_arg(args, char *); ++argc);
    va_end(args);

    if (argc == 0) {
	pr("bad args to input_op\n");
	abort();
    }

    for (;;) {
	pr("%s", prompt);

	input_line(buf, sizeof(buf));

	for (i = 0; i < argc; ++i) {
	    if (!strcmp(argv[i], buf)) {
		return i;
	    }
	}

	pr("Please type ");
	pr("\"%s\"", argv[0]);
	for (i = 1; i < argc; ++i) {
	    if (argv[i+1] == NULL) {
		pr(" or \"%s\"", argv[i]);
	    } else {
		pr(", \"%s\"", argv[i]);
	    }
	}
	pr(".\n");
    }
} /* input_op */


/* INPUT_LINE: Get a line of input from the operator.
 */
input_line(buf, size)
    char *buf;
    int size;
{
    int i;

    if (feof(stdin)) clearerr(stdin);
    for (i = 0; i < size-1; ++i) {
	if ((buf[i] = getchar()) == '\n')
	    break;
    }
    buf[i] = '\0';

    /* Send op's input to mail */
    if (mailfp) (void) fprintf(mailfp, "%s\n", buf);
} /* input_line */


/* PRERROR: Like perror, but make output go to the right place.
 */
prerror(buf)
    char *buf;
{
    extern int sys_nerr;

    pr("%s: %s\n", buf,
	errno >= sys_nerr ? "Unknown error" : sys_errlist[errno]);
} /* prerror */


/* SETUP_MAIL: Set up a person to mail to.  If the person being mailed to
 * hasn't changed since the last call here, don't do anything.  If called
 * with a NULL argument, close the mail channel.
 */
setup_mail(whom)
    char *whom;
{
    static char saved_whom[100];
    char buf[MAXPATHLEN];
    int pip[2], i;

    /* Return if not changing */
    if (whom && !strcmp(whom, saved_whom)) return;

    /* Close old connection */
    if (mailfp) (void) fclose(mailfp);
    saved_whom[0] = '\0';
    mailfp = NULL;

    /* Done if whom is null */
    if (whom == NULL) return;

    (void) NCOPY(saved_whom, whom, sizeof(saved_whom));
    (void) sprintf(buf, "%s %s", MAIL_CMD, whom);
    (void) pipe(pip);

    switch(fork()) {
	case -1:
	    prerror("fork");
	    return;
	case 0:
	    (void) dup2(pip[R_END], 0);
	    for (i = 3; i < dtablesize; ++i) (void) close(i);
	    reset_sigs();
	    exec_str(buf);
	    perror(buf);
	    exit(0);
	default:
	    break;
    }
    (void) close(pip[R_END]);
    mailfp = fdopen(pip[W_END], "w");
    setlinebuf(mailfp);
    return;
} /* setup_mail */


/* GET_MAILFD: Return the fd being used for mail.
 */
get_mailfd()
{
    return mailfp ? fileno(mailfp) : -1;
} /* get_mail_fd */
