%{
/*******************************************************************/
/*      Copyright (C) 1989 by                                      */
/*            The Regents of the University of California          */
/*                                                                 */
/*                             DISCLAIMER                          */
/*                                                                 */
/*      "Although   each   program   has   been  tested  by  its   */
/*      contributor, no warranty, express or implied, is made by   */
/*      the  contributor  or the University of California, as to   */
/*      the accuracy and functioning of the program and  related   */
/*      program material, nor shall the fact of the distribution   */
/*      constitute any such warranty, and no  responsibility  is   */
/*      assumed   by   the  contributor  or  the  University  of   */
/*      California in connection herewith."                        */
/*******************************************************************/

/* PARSER.Y: Parser for the config file
 *
 * $Header: /cd/ua/sam/src/ss/backup/RCS/parser.y,v 1.3 89/06/09 16:10:22 sam REL-6-9-89 $
 *
 * $Log:	parser.y,v $
 * Revision 1.3  89/06/09  16:10:22  sam
 * No changes.
 * 
 * Revision 1.2  89/04/03  21:50:33  sam
 * No changes
 * 
 * Revision 1.1  89/03/08  20:50:09  sam
 * Initial revision
 * 
 */
#ifndef lint
static char rcsid[] = "$Header: /cd/ua/sam/src/ss/backup/RCS/parser.y,v 1.3 89/06/09 16:10:22 sam REL-6-9-89 $";
#endif

#include "../global.h"
#include "backup.h"

/* Locals */
static int
    hadwhen;			/* Flag that we had a when clause */
static struct machrec
    *curmach;			/* Pointer to current machine */
static struct saverec
    *cursave;			/* Pointer to current save */

/* Externs */
extern int
    nsaves,			/* Total number of saves */
    nsets,			/* Number of save sets */
    timeout;			/* Time out value */
extern char
    *mailto,			/* Default person to mail to */
    *data_file,			/* Name of dumpdates file */
    *save_format,		/* Name of savelist file */
    *save_list_dir,		/* Dir where savelists go */
    *tapedev,			/* Name of the tape device */
    **sets;			/* Save sets array */
extern struct saverec
    **saves;			/* Array of saves */

/* Function Decls */
extern char
    *realloc(), *malloc();

#define _A(type, n) (unsigned)sizeof(type)*(n)
#define ALL(type, n) \
	((type*)check_alloc(malloc(_A(type, n))))
#define ALLV(p, type, n) \
	((p)=(type*)check_alloc((p) ? realloc((char*)p, _A(type,n)) \
				    : malloc(_A(type,n))))

%}

%token T_MAIL
%token T_DATA_FILE
%token T_TIMEOUT
%token T_SAVE_NAME
%token T_SAVELISTDIR
%token T_TAPEDEV
%token T_MACHINE
%token T_FULLSAVE
%token T_INCREMENTAL
%token T_PATH
%token T_TAPE
%token T_WHEN
%token T_SETS
%token T_SUN
%token T_MON
%token T_TUE
%token T_WED
%token T_THU
%token T_FRI
%token T_SAT
%token <label>	T_PATHNAME
%token <label>	T_TEXT
%token <num>	T_NUMBER

%type <stypes> save_type

%left ':'
%left '='
%left '-'

%% 

start:		{
		    nsets = 0;
		}
		all
		{ 
		    if (!nsaves) yyerror("No saves in config file!\n");
		}
	;

all:		toplevel all
	|	epsilon
	;

toplevel:	one_mach
	|	T_MAIL separators T_TEXT
		    { if (!mailto) mailto = new_str($3); }
	|	T_SAVE_NAME separators T_TEXT
		    { if (!save_format) save_format = new_str($3); }
	|	T_DATA_FILE separators T_PATHNAME
		    { if (!data_file) data_file = new_str($3); }
	|	T_TIMEOUT separators T_NUMBER
		{
		    timeout = $3 * 60;
		    if (timeout <= 0) yyerror("Bad timeout value\n");
		}
	|	T_SETS separators setnames
	|	T_SAVELISTDIR separators T_PATHNAME
		    { if (!save_list_dir) save_list_dir = new_str($3); }
	|	T_TAPEDEV separators T_PATHNAME
		    { if (!tapedev) tapedev = new_str($3); }
	;

setnames:	setname setnames
	|	epsilon
	;

setname:	T_TEXT
		{ 
		    ++nsets;
		    ALLV(sets, char*, nsets);
		    sets[nsets-1] = new_str($1);
		}
	;

one_mach:	T_MACHINE separators T_TEXT
		{
		    curmach = ALL(struct machrec, 1);
		    curmach->name = new_str($3);
		}
		saves
	;

saves:		one_save saves
	|	one_save
	;

one_save:	save_type
		{
		    cursave = ALL(struct saverec, 1);
		    ++nsaves;
		    ALLV(saves, struct saverec *, nsaves);
		    saves[nsaves-1] = cursave;
		    bzero((char*)cursave, sizeof(struct saverec));
		    cursave->type = $1;
		    cursave->mach = curmach;
		    hadwhen = 0;
		}
		clauses
		{
		    if (!cursave->npaths) yyerror("No paths\n");
		    if (!hadwhen) yyerror("No \"when\" clause\n");
		    if (!cursave->tape_name) yyerror("No tape name\n");
		}
	;

save_type:	T_INCREMENTAL
		{ $$ = INC; }
	|	T_FULLSAVE
		{ $$ = FULL; }
	;

clauses:	path clauses
	|	when clauses
	|	tape clauses
	|	epsilon
	;

path:		T_PATH separators pathnames
	;

pathnames:	onepathname pathnames
	|	onepathname
	;

onepathname:	T_PATHNAME
		{
		    ++cursave->npaths;
		    ALLV(cursave->paths, char*, cursave->npaths);
		    cursave->paths[cursave->npaths-1] = new_str($1);
		}

when:		T_WHEN separators days
		    { ++hadwhen; }
	;

days:		oneday days
	|	oneday
	;

oneday:		T_SUN { cursave->days[0] = 1; }
	|	T_MON { cursave->days[1] = 1; }
	|	T_TUE { cursave->days[2] = 1; }
	|	T_WED { cursave->days[3] = 1; }
	|	T_THU { cursave->days[4] = 1; }
	|	T_FRI { cursave->days[5] = 1; }
	|	T_SAT { cursave->days[6] = 1; }
	;

tape:		T_TAPE separators T_TEXT
		    { cursave->tape_name = new_str($3); }
	;

separators:	':' | '=' | '-' | epsilon
	;

epsilon:	/* nothing */
	;

%%

#include <stdio.h>

static yyerror(s)
    char *s;
{
    extern int yylineno;

    pr("Error on line %d of config file: %s\n", yylineno, s);
    exit(1);
}


/* CHECK_ALLOC: Check the return of a call to malloc and die if it's null.
 */
char *check_alloc(p)
    char *p;
{
    if (p == NULL) yyerror("Out of memory");
    return p;
} /* check_alloc */


/* NEW_STR: Alloc room for a string and copy the passed string into it.
 */
char *new_str(s)
    char *s;
{
    char *retval;
    int l;

    l = strlen(s)+1;
    retval = ALL(char, l);
    bcopy(s, retval, l);
    return retval;
} /* new_str */
