/*******************************************************************/
/*      Copyright (C) 1989 by                                      */
/*            The Regents of the University of California          */
/*                                                                 */
/*                             DISCLAIMER                          */
/*                                                                 */
/*      "Although   each   program   has   been  tested  by  its   */
/*      contributor, no warranty, express or implied, is made by   */
/*      the  contributor  or the University of California, as to   */
/*      the accuracy and functioning of the program and  related   */
/*      program material, nor shall the fact of the distribution   */
/*      constitute any such warranty, and no  responsibility  is   */
/*      assumed   by   the  contributor  or  the  University  of   */
/*      California in connection herewith."                        */
/*******************************************************************/

/* TPIO.C: Tape I/O (mostly O)
 *
 * $Header: /cd/ua/sam/src/ss/backup/RCS/tpio.c,v 1.3 89/06/09 16:10:47 sam REL-6-9-89 $
 *
 * $Log:	tpio.c,v $
 * Revision 1.3  89/06/09  16:10:47  sam
 * No changes
 * 
 * Revision 1.2  89/04/03  21:50:18  sam
 * No changes
 * 
 * Revision 1.1  89/03/08  20:50:14  sam
 * Initial revision
 * 
 */
#ifndef lint
static char rcsid[] = "$Header: /cd/ua/sam/src/ss/backup/RCS/tpio.c,v 1.3 89/06/09 16:10:47 sam REL-6-9-89 $";
#endif

#include "../global.h"
#include "backup.h"
#include <sys/param.h>
#include <sys/file.h>
#include <sys/ioctl.h>
#include <sys/mtio.h>
#include <sys/types.h>
#include <stdio.h>

/* Locals */
char
    tapename[MAXTAPENAME];	/* Name of the tape */

/* Externs */
extern char
    *tapedev;			/* Name of tape device */
extern int
    tape_num,			/* Number of tape we're on */
    tapefd,			/* Fd to read from tape */
    bytes_left,			/* Number of bytes left on this tape */
    bytes_per_tape,		/* Number of bytes on one tape */
    chat,			/* Talk with the operator? */
    isatape,			/* Is output a tape? */
    tape_num,			/* Number of tape */
    file_num;			/* Number of file we're on */

/* Extern Funcs */
extern off_t
    lseek();
extern char
    *strncpy(), *strcat(), *sprintf();



/* CHECK_ERR: Do a perror if code is -1.  Always return code.
 */
check_err(code, name)
    int code;
    char *name;
{
    if (code == -1) prerror(name);
    return code;
} /* check_err */


/* TP_IOCTL: Do an ioctl on the tape.
 */
int tp_ioctl(cmd, count)
    int cmd, count;
{
    struct mtop op;

    op.mt_op = cmd;
    op.mt_count = count;
    return check_err(ioctl(tapefd, MTIOCTOP, &op), "ioctl");
} /* tp_ioctl */
	

/* TP_CLOSE: Close the tape device.
 */
int tp_close()
{
    int retval = (tapefd == -1) ? 0 : check_err(close(tapefd), "close");
    tapefd = -1;
    return retval;
} /* tp_close */


/* TP_OPEN: Open the tape device.
 */
int tp_open(mode)
    int mode;
{
    return tapefd = check_err(open(tapedev, mode, 0666), "open");
} /* tp_open */


/* TP_SEEK: Do an lseek on the tape.
 */
int tp_seek(offset, whence)
    int offset, whence;
{
    return check_err((int)lseek(tapefd, (off_t)offset, whence), "lseek");
} /* tp_seek */


/* TP_READ: Low level read of tape.  Should only read full blocks here.
 */
int tp_read(blk, size)
    char *blk;
    int size;
{
    return read_all(tapefd, blk, size);
} /* tp_read */


/* TP_WRITE: Low level write to tape.
 */
int tp_write(blk, size)
    char *blk;
    int size;
{
    return write_all(tapefd, blk, size);
} /* tp_write */


/* TP_CHANGE: Change to next tape.
 */
tp_change()
{
    if (tp_close() == -1) return -1;
    if (mount_new_tape() == -1) return -1;
    return 0;
} /* tp_change */


/* MOUNT_TAPE: Mount the first tape from a restore.  Fsf to the right file on
 * the tape.
 */
mount_tape(tname, fnum)
    char *tname;
    int fnum;
{
    tape_num = 0;
    if (!tname || tname[0] == '\0') {
	tapename[0] = '\0';
    } else {
	(void) NCOPY(tapename, tname, sizeof(tapename));
    }

    /* Open the tape */
    if (mount_new_tape() == -1) return -1;

    /* Fsf to the right file */
    file_num = fnum;
    if (fnum) {
	if (!isatape) {
	    pr("Can't do a file skip when not using a tape.\n");
	    return -1;
	}
	return tp_ioctl(MTFSF, file_num);
    }
    return 0;
} /* mount_tape */


/* MOUNT_NEW_TAPE: Change the tape.
 */
mount_new_tape()
{
    static char nice_num[5][10] =
	{"first", "second", "third", "fourth", "fifth"};
    char name[MAXTAPENAME+2];
    char prompt[200];

    if (chat) {
	if (*tapename) {
	    (void) NCOPY(name, tapename, MAXTAPENAME);
	    (void) strcat(name, " ");
	} else {
	    *name = '\0';
	}
	if (tape_num > 4) {
	    (void) sprintf(prompt,
		"\nInsert the %dth %stape into the tape drive\n",
		tape_num+1, name);
	} else {
	    (void) sprintf(prompt,
		"\nInsert the %s %stape into the tape drive\n",
		nice_num[tape_num], name);
	}
	(void) strcat(prompt, "Type \"ok\" when the tape drive is ready\n");
	(void) input_op(prompt, "ok", 0);
    }

    /* Open the tape */
    while (tp_open(isatape ? O_WRONLY : O_WRONLY|O_TRUNC) == -1) {
	if (chat) {
	    if (input_op("\n\
The tape can't be opened for writing.\n\
Type \"retry\" to try again\n\
Type \"abort\" to abort all saves\n", "retry", "abort", 0))
	    {
		abort();
	    }
	} else {
	    return -1;
	}
    }
    ++tape_num;
    file_num = 0;
    bytes_left = bytes_per_tape;
    return 0;
} /* mount_new_tape */
