/* GLOBAL.H: Global definitions
 *
 * $Header: /cd/ua/sam/src/ss/RCS/global.h,v 1.3 89/06/09 16:13:05 sam REL-6-9-89 $
 *
 * $Log:	global.h,v $
 * Revision 1.3  89/06/09  16:13:05  sam
 * No changes.
 * 
 * Revision 1.2  89/04/03  21:17:42  sam
 * Mostly just moved everything around.  Added one or two constants.
 * 
 * Revision 1.1  89/03/08  20:29:47  sam
 * Initial revision
 * 
 */

/* Includes */

#include <sys/types.h>
#include <sys/param.h>



/* Constants */

/* SUN3X: Set this if running Sun OS 3.5 or earlier */
#ifdef sun
/* #define SUN3X */
#endif

/* SUN40: Set if running OS 4.0 or later (won't be set if SUN3X already is) */
#if defined(sun) && !defined(SUN3X)
#define SUN40
#endif

#define	MAXFILENAME	256		/* Supposedly the limit in the kernel */
#define	MAXDIRLENGTH	(MAXPATHLEN*5)	/* Real MAXPATH */
#define MAXTAPENAME	48		/* Max chars in a tape name */
#define MAXSLISTNAME	48		/* Max chars in a savelist name */
#define	SENDSIZE	(52*1024)	/* Max socket send size */
#define BYTESPT		(64512*30000)	/* Max bytes per tape */
#define BLKSIZ		64512		/* Default block size */
#define SAVELISTDIR	"/usr/local/etc/ss/savelists"
					/* Default dir to put savelists */
#define RMT_CMD		"/etc/rmt"	/* How to run rmt */
#define	TRUE		1
#define	FALSE		0
#define FILE_INFO_SIZE	20		/* Our size for file_info struct */
#define SAVE_HEADER_SIZE (20+MAXTAPENAME+MAXSLISTNAME)
					/* Our size for save header struct */
#define VERSION		"10"		/* File format version */

/* Save types. (can be or'ed together)
 */
#define SAVE_INC	1	/* Is an incremental */
#define SAVE_SLIST	2	/* Savelist only */

/* Record types in file_info */
#define TYPE_END	0	/* End of savelist */
#define	A_FILE		1	/* A file with data */
#define	CD_UP		2	/* Move up one directory level */
#define	CD_DOWN		3	/* Move down into a new subdirectory */
#define CD_ROOT		4	/* Move to root of save */
#define	SYMLINK		5	/* A symlink */
#define	L_FILE		6	/* A hardlink to a previously-saved file */
#define	I_FILE		7	/* A file not saved on an incremental */
#define	M_FILE		8	/* Same as A_FILE, but has multiple links */
#define PATH_DIR	9	/* Like CD_DOWN, but not fullsave of this dir */

#define NTYPES		10	/* Number of the above file types */



/* Typedefs */

/* SIGTYPE:  The type of a signal handler */
#ifdef SUN40
    typedef void sigtype;
#else
    typedef int sigtype;
#endif



/* Structures */

/* SAVE_HEADER: Save or savelist header  (length is multiple of 4).  If
 * compiler puts holes in this, we're sunk.
 * IF YOU UPDATE THIS, MAKE SURE TO UPDATE SAVE_HEADER_SIZE!
 */
struct save_header {
    char version[2];		/* Version number */
    unsigned char type;		/* Type of save (from above list) */
    unsigned char tapenum;	/* Which tape in set this save is on */
    u_long begin_time;		/* Time save began */
    u_long end_time;		/* End time of save */
    u_long file_offset;		/* Starting filenum on tape */
    u_long bytes_left;		/* How many bytes left on tape starting here */
    char tapename[MAXTAPENAME];	/* Name of tape set */
    char slistname[MAXSLISTNAME];	/* Name of savelist file */
};

/* FILE_INFO: Info for each file in save.  Make sure the compiler packs this
 * (no gaps), so that it can be used for look ups of stuff saved in a file.
 * If you change this, don't forget to change FILE_INFO_SIZE!
 */
struct file_info {
    u_long st_size;	/* Size in bytes, or offset on tape (L_FILE) */
    u_long st_atime;	/* Last access time */
    u_long st_mtime;	/* Last modification time */
    u_short st_mode;	/* Complete mode of the file */
    u_short st_uid;		/* Obvious */
    u_short st_gid;
    unsigned char name_len;	/* Length of filename */
    unsigned char record_type;	/* One of the following record types */
    /* File name goes here -- not null terminated */
};



/* Functional macros */

/* NCOPY: Do a strncpy, but make sure the string is null terminated.
 */
#define NCOPY(d, s, n) strncpy(d, s, n); d[(n)-1] = '\0'

/* NEW: Malloc n of the type specified.
 */
#define NEW(type, n) (type *)malloc((unsigned)(sizeof(type)*(n)))

/* ALLOC: Allocate space for a pointer.  If it's nil now, call malloc.  If
 * non-nill, call realloc.
 */
#define ALLOC(p, type, n)\
    ((p) ? (type*)realloc((char*)p, (unsigned)sizeof(type)*(n)) : NEW(type, n))

/* NUM_ARR: Tell the number of elements in an array.
 */
#define NUM_ARR(a) (sizeof(a)/sizeof(a[0]))

/* MIN and MAX: Return the min or max of two numbers.
 */
#define min(x,y) ((x) < (y) ? (x) : (y))
#define max(x,y) ((x) > (y) ? (x) : (y))



/* Macros that should be set in types.h and param.h, but may not be */

#ifndef MAXHOSTNAMELEN
#define MAXHOSTNAMELEN	32
#endif

#ifndef FD_SETSIZE
#define FD_SETSIZE	32
#endif

#ifndef NFDBITS
#define NFDBITS		(sizeof(long) * NBBY)
#endif

#ifndef FD_SET
#define FD_SET(n, p)	((p)->fds_bits[(n)/NFDBITS] |= (1 << ((n) % NFDBITS)))
#endif

#ifndef FD_ZERO
#define FD_ZERO(p)	bzero((char *)(p), sizeof(*(p)))
#endif
