/*******************************************************************/
/*      Copyright (C) 1989 by                                      */
/*            The Regents of the University of California          */
/*                                                                 */
/*                             DISCLAIMER                          */
/*                                                                 */
/*      "Although   each   program   has   been  tested  by  its   */
/*      contributor, no warranty, express or implied, is made by   */
/*      the  contributor  or the University of California, as to   */
/*      the accuracy and functioning of the program and  related   */
/*      program material, nor shall the fact of the distribution   */
/*      constitute any such warranty, and no  responsibility  is   */
/*      assumed   by   the  contributor  or  the  University  of   */
/*      California in connection herewith."                        */
/*******************************************************************/

/* BROWSE.C: Save browser
 *
 * $Header: /cd/ua/sam/src/ss/rest/RCS/browse.c,v 1.2 89/06/09 15:25:37 sam REL-6-9-89 $
 *
 * $Log:	browse.c,v $
 * Revision 1.2  89/06/09  15:25:37  sam
 * Made ls put output into columns
 * 
 * Revision 1.1  89/04/03  20:57:57  sam
 * Initial revision
 * 
 */
#ifndef lint
static char rcsid[] = "$Header: /cd/ua/sam/src/ss/rest/RCS/browse.c,v 1.2 89/06/09 15:25:37 sam REL-6-9-89 $";
#endif

#include "../global.h"
#include "rest.h"
#include <stdio.h>
#include <sys/stat.h>

/* Structs */
struct ls_info {		/* Info to save in an ls command */
    struct file_ent *fe;	/* Ptr to info about this file */
    char *name;			/* Name to display, not null terminated*/
    int len;			/* Length of above name */
};
#define LS_INC 15
#define SCRNWIDTH 80

/* Local vars */
static struct file_ent
    *curdir, *root;
struct slist_info
    *list_info;

/* Function decls */
static int
    cd_cmd(), ls_cmd(), quit_cmd(), help_cmd(), echo_cmd(), add_cmd(),
    extr_cmd();
struct file_ent
    *find_filename(), *find_path();
extern char
    *index();

struct cmd_rec cmds[] = {
    {"cd           Change the current directory", cd_cmd},
    {"ls           List files", ls_cmd},
    {"echo         Echo back the command line", echo_cmd},
    {"add          Add files to be restored", add_cmd},
    {"extract      Extract the files marked in this savelist", extr_cmd},
    {"quit         Quit back to top level", quit_cmd},
    {"help         Print this message", help_cmd},
    {"?            Print this message", help_cmd},
};


/* BROWSE: Start browsing through the given savelist.
 */
browse(_info)
    struct slist_info *_info;
{
    list_info = _info;
    root = list_info->root;
    curdir = root;
    shell("browse> ", NUM_ARR(cmds), cmds, (int(*)())NULL);
} /* browse */


/* CD_CMD: Change to a new directory.
 */
/*ARGSUSED*/
static cd_cmd(argc, argv, out)
    int argc;
    char **argv;
    FILE *out;
{
    struct file_ent *fe;

    if (argc != 2) {
	pr("Usage: cd directory\n");
	return 0;
    }
    if ((fe = find_path(curdir, argv[1], (struct file_ent**)NULL)) == NULL
	|| !(fe->flags & FE_ISDIR))
    {
	pr("%s: No such directory\n", argv[1]);
	return 0;
    }
    curdir = fe;
    return 0;
} /* cd_cmd */


/* QUIT_CMD: Quit the browser.
 */
/*ARGSUSED*/
static quit_cmd(argc, argv, out)
    int argc;
    char **argv;
    FILE *out;
{
    return -1;
} /* quit_cmd */


/* HELP_CMD: Print a help message.
 */
/*ARGSUSED*/
static help_cmd(argc, argv, out)
    int argc;
    char **argv;
    FILE *out;
{
    int i;

    for (i = 0; i < NUM_ARR(cmds); ++i) {
	(void) fprintf(out, "%s\n", cmds[i].name);
    }
    return 0;
} /* help_cmd */


/* ECHO_CMD: Echo the command line.
 */
static int echo_cmd(argc, argv, out)
    int argc;
    char **argv;
    FILE *out;
{
    int i;

    for (i = 1; i < argc; ++i) {
	fputs(argv[i], out);
    }
    if (argc > 1) (void) putc('\n', out);
    return 0;
} /* echo_cmd */


/* ADD_CMD: Add files to be restored.
 */
/*ARGSUSED*/
static int add_cmd(argc, argv, out)
    int argc;
    char **argv;
    FILE *out;
{
    int i;
    struct file_ent *fe, *parent;

    for (i = 1; i < argc; ++i) {
	fe = find_path(curdir, argv[i], &parent);
	if (fe == NULL) {
	    pr("%s: No such file or directory\n", argv[i]);
	} else {
	    /* Add this fe and anything under it */
	    add_file(fe);

	    /* Make a path from the root to the parent of this file */
	    while (parent != root) {
		if (!(parent->flags & FE_REST)) {
		    parent->flags |= FE_REST;
		    list_info->num_marked++;
		}
		parent = parent->info_u.dir->parent;
	    }
	}
    }
} /* add_cmd */


/* ADD_FILE: Set the REST flag on the file given it and any files under it.
 */
add_file(fe)
    register struct file_ent *fe;
{
    register struct gen_list *flist;
    register int i;

    if (!(fe->flags & FE_REST)) {
	fe->flags |= FE_REST;
	list_info->num_marked++;
    }
    if (fe->flags & FE_ISDIR) {
	flist = &fe->info_u.dir->file_list;
	for (i = 0; i < LIST_SZ(*flist); ++i) {
	    fe = LIST_GET(*flist, struct file_ent*, i);
	    if (fe->flags & FE_ISDIR) {
		add_file(fe);
	    } else {
		if (!(fe->flags & FE_REST)) {
		    fe->flags |= FE_REST;
		    list_info->num_marked++;
		}
	    }
	}
    }
} /* add_file */


/* EXTR_CMD: Start an extract for this savelist.
 */
/*ARGSUSED*/
static int extr_cmd(argc, argv, out)
    int argc;
    char **argv;
    FILE *out;
{
    if (list_info->num_marked) {
	inter_rest(list_info);
    } else {
	pr("No files marked for restore\n");
    }
} /* extr_cmd */


/* LS_CMD: List files
 */
/*ARGSUSED*/
static int ls_cmd(argc, argv, out)
    int argc;
    char **argv;
    FILE *out;
{
    struct gen_list llist;
    struct ls_info *lsi;
    struct file_ent *fe;
    char *s;
    int i, j, dopt, lopt, uopt, topt;

    dopt = lopt = uopt = topt = 0;
    if (argc > 1 && argv[1][0] == '-') {
	for (s = &argv[1][1]; *s; ++s) switch(*s) {
	    case 'l': lopt++; break;
	    case 'd': dopt++; break;
	    case 'u': uopt++; break;
	    case 't': topt++; break;
	    default:
		pr("Unkown option %c\n", *s);
		break;
	}
	argc--; argv++;
    }

    if (argc == 1) {
	ls_dir(out, curdir, dopt, lopt, uopt, topt);
    } else {
	/* Do non-directories (and dirs if dopt is set) first */
	init_list(&llist);
	for (i = 1, j = 1; i < argc; ++i) {
	    fe = find_path(curdir, argv[i], (struct file_ent**)NULL);
	    if (fe && (fe->flags & FE_ISDIR)) {
		argv[j++] = argv[i];
		continue;
	    }
	    if (fe == NULL) {
		pr("%s: No such file or directory\n", argv[i]);
	    } else {
		lsi = LIST_NEW(llist, struct ls_info, LS_INC);
		if (lsi == NULL) return 0;
		lsi->fe = fe;
		lsi->name = argv[i];
		lsi->len = strlen(argv[i]);
	    }
	}
	argc = j;
	if (LIST_SZ(llist)) {
	    ls_list(out, &llist, dopt, lopt, uopt, topt);
	    if (argc > 1) putc('\n', out);
	}
	delete_list(&llist);

	/* Do the directories here */
	for (i = 1; i < argc; ++i) {
	    fe = find_path(curdir, argv[i], (struct file_ent**)NULL);
	    if (!fe) continue;			/* This shouldn't happen */
	    if (argc != 2) fprintf(out, "%s:\n", argv[i]);
	    ls_dir(out, fe, dopt, lopt, uopt, topt);
	    if (i != argc-1) putc('\n', out);
	}
    }
    return 0;
} /* ls_cmd */


/* LS_DIR: Read all the files in a dir and pass to ls_list.
 */
ls_dir(out, fe, dopt, lopt, uopt, topt)
    FILE *out;
    struct file_ent *fe;
    int dopt, lopt, uopt, topt;
{
    struct gen_list llist, *flist;
    struct file_info *info;
    struct ls_info *lsi;
    int i;

    if (!(fe->flags & FE_ISDIR)) return;
    flist = &fe->info_u.dir->file_list;

    init_list(&llist);
    for (i = 0; i < LIST_SZ(*flist); ++i) {
	lsi = LIST_NEW(llist, struct ls_info, LS_INC);
	if (lsi == NULL) return;
	lsi->fe = LIST_GET(*flist, struct file_ent*, i);
	info = FETOINFO(lsi->fe);
	lsi->name = INFONAME(info);
	lsi->len = info->name_len;
    }
    ls_list(out, &llist, dopt, lopt, uopt, topt);
} /* ls_dir */


/* LS_LIST: Sort the list of ls_info's and pass to either ls_long or ls_col.
 */
ls_list(out, list, dopt, lopt, uopt, topt)
    FILE *out;
    struct gen_list *list;
    int dopt, lopt, uopt, topt;
{
    int ls_cmp_mtime(), ls_cmp_atime(), ls_cmp_name(), ls_long(), ls_col();

    /* Sort the list */
    qsort((char*)&LIST_GET(*list, struct ls_info, 0), LIST_SZ(*list),
	  sizeof(struct ls_info),
	  topt ? (uopt ? ls_cmp_atime : ls_cmp_mtime) : ls_cmp_name);

    /* Pass to either ls_long or ls_col */
    (lopt ? ls_long : ls_col)(out, list, dopt, lopt, uopt, topt);
} /* ls_list */


/* LS_CMP_MTIME: Compare to sort by mtime.
 */
int ls_cmp_mtime(lsi1, lsi2)
    struct ls_info *lsi1, *lsi2;
{
    register struct file_info *fi1 = FETOINFO(lsi1->fe);
    register struct file_info *fi2 = FETOINFO(lsi2->fe);
    return convl(&fi2->st_mtime) - convl(&fi1->st_mtime);
} /* ls_cmp_mtime */


/* LS_CMP_ATIME: Compare to sort by atime.
 */
int ls_cmp_atime(lsi1, lsi2)
    struct ls_info *lsi1, *lsi2;
{
    register struct file_info *fi1 = FETOINFO(lsi1->fe);
    register struct file_info *fi2 = FETOINFO(lsi2->fe);
    return convl(&fi2->st_atime) - convl(&fi1->st_atime);
} /* ls_cmp_atime */


/* LS_CMP_NAME: Compare to sort alphabetically by name.
 */
int ls_cmp_name(lsi1, lsi2)
    register struct ls_info *lsi1, *lsi2;
{
    int i = strncmp(lsi1->name, lsi2->name, min(lsi1->len, lsi2->len));
    return i ? i : lsi1->len - lsi2->len;
} /* ls_cmp_name */


ls_long(out, list, dopt, lopt, uopt, topt)
    FILE *out;
    struct gen_list *list;
    int dopt, lopt, uopt, topt;
{
    ls_col(out, list, dopt, lopt, uopt, topt);
} /* ls_long */


/*ARGSUSED*/
ls_col(out, list, dopt, lopt, uopt, topt)
    FILE *out;
    struct gen_list *list;
    int dopt, lopt, uopt, topt;
{
    struct ls_info *lsi;
    int i, j, n, idx, col_width, ncols, nrows;

    /* Find column width */
    for (i = col_width = 0; i < LIST_SZ(*list); ++i) {
	j = LIST_GET(*list, struct ls_info, i).len;
	if (j > col_width) col_width = j;
    }
    col_width = (col_width + 10) & ~(8-1);

    /* Find number of cols and rows */
    ncols = SCRNWIDTH / col_width;
    if (!ncols) ncols = 1;
    nrows = ((LIST_SZ(*list)-1) / ncols) + 1;

    /* Print it all out */
    for (i = 0; i < nrows; ++i) {
	for (j = 0; j < ncols; ++j) {
	    idx = (j * nrows) + i;
	    if (idx >= LIST_SZ(*list)) continue;
	    lsi = &LIST_GET(*list, struct ls_info, idx);
	    n = ls_pr_name(out, lsi, FETOINFO(lsi->fe));
	    if ((idx + nrows) >= LIST_SZ(*list) || j == ncols-1) continue;
	    while (n++ < col_width) putc(' ', out);
	}
	putc('\n', out);
    }
} /* ls_col */


/* LS_PR_NAME: Print a file name, return the number of chars output.
 */
ls_pr_name(out, lsi, fi)
    FILE *out;
    struct ls_info *lsi;
    struct file_info *fi;
{
    register int c, mode;

    putc((lsi->fe->flags & FE_REST) ? '+' : ' ', out);
    fwrite(lsi->name, lsi->len, 1, out);
    mode = convs(&fi->st_mode);
    switch(mode & S_IFMT) {
	case S_IFDIR: c = '/'; break;
	case S_IFLNK: c = '@'; break;
	case S_IFSOCK: c = '='; break;
	default:
	    c = (mode & (S_IXUSR|S_IXGRP|S_IXOTH)) ? '*' : 0;
	    break;
    }
    if (c) {
	putc(c, out);
	return lsi->len + 2;
    } else {
	return lsi->len + 1;
    }
} /* ls_pr_name */


/* FIND_FILENAME: Search a given directory for a name.
 */
static struct file_ent *find_filename(dir, name)
    struct file_ent *dir;
    char *name;
{
    struct file_ent **fe, **end_fe;
    struct gen_list *list = &dir->info_u.dir->file_list;
    struct file_info *info;
    int len = strlen(name);

    end_fe = &LIST_GET(*list, struct file_ent*, LIST_SZ(*list));
    for (fe = &LIST_GET(*list, struct file_ent*, 0); fe < end_fe; ++fe)
    {
	info = FETOINFO(*fe);
	if (len == info->name_len && !bcmp(name, INFONAME(info), len)) {
	    return *fe;
	}
    }
    return NULL;
} /* find_filename */


/* FIND_PATH: Find a full path in the fs.
 */
static struct file_ent *find_path(dir, name, ret_parent)
    struct file_ent *dir;
    char *name;
    struct file_ent **ret_parent;
{
    char *s, *p;
    struct file_ent *cdir = dir, *fe;

    /* Go to root if abs path */
    while (*name == '/') {
	cdir = root;
	++name;
    }
    fe = cdir;

    /* Step across each part of the path */
    for (p = index(s = name, '/');;p = index(s = p+1, '/'), cdir = fe) {
	if (!*s) break;
	if (p) *p = '\0';
	if (ISDOTFILE(s)) {
	    if (s[1]) {
		fe = cdir->info_u.dir->parent;
	    } else {
		fe = cdir;
	    }
	} else {
	    fe = find_filename(cdir, s);
	}
	if (p) *p = '/';
	if (!fe) return NULL;
	if (!p) break;
	if (!(fe->flags & FE_ISDIR)) return NULL;
    }
    if (ret_parent) *ret_parent = cdir;
    return fe;
} /* find_path */
