/*******************************************************************/
/*      Copyright (C) 1989 by                                      */
/*            The Regents of the University of California          */
/*                                                                 */
/*                             DISCLAIMER                          */
/*                                                                 */
/*      "Although   each   program   has   been  tested  by  its   */
/*      contributor, no warranty, express or implied, is made by   */
/*      the  contributor  or the University of California, as to   */
/*      the accuracy and functioning of the program and  related   */
/*      program material, nor shall the fact of the distribution   */
/*      constitute any such warranty, and no  responsibility  is   */
/*      assumed   by   the  contributor  or  the  University  of   */
/*      California in connection herewith."                        */
/*******************************************************************/

/* DOREST.C: Do a restore.
 *
 * $Header: /cd/ua/sam/src/ss/rest/RCS/dorest.c,v 1.3 89/06/09 15:49:59 sam REL-6-9-89 $
 *
 * $Log:	dorest.c,v $
 * Revision 1.3  89/06/09  15:49:59  sam
 * Now sets the mode/owner of the top directory restored.
 * 
 * Revision 1.2  89/04/03  21:01:06  sam
 * Changed to use new memory routines
 * If file is less than IOBLKSIZ, do a read/write instead of mmap
 * Fixed it so that directories now have their set_attr done after
 * everything in them has been restored (at the CD_UP).  This fixes the 
 * problem with dirs having the wrong mtime.
 * Changed set_attr to use fchmod and fchown whenever possible.
 * 
 * Revision 1.1  89/03/08  20:38:21  sam
 * Initial revision
 * 
 */
#ifndef lint
static char rcsid[] = "$Header: /cd/ua/sam/src/ss/rest/RCS/dorest.c,v 1.3 89/06/09 15:49:59 sam REL-6-9-89 $";
#endif

#include "../global.h"
#include "rest.h"
#include <sys/stat.h>
#include <stdio.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <errno.h>

/* Return codes */
#define RST_OK		0
#define RST_ALL_DONE	1
#define RST_SOME_ERR	2

/* IOBUFSIZ: Size to do disk writes.
 */
#define IOBUFSIZ 8192

/* Local vars */
static char
    *slist_ptr,			/* Pointer into savelist (if using one) */
    path_name[MAXDIRLENGTH];	/* Our current path */
static int
    del_dir_cnt,		/* Nest count of non-restored dirs */
    will_rest,			/* Set this if at least one file will be done */
    tp_offset,			/* Byte offset into tape */
    path_len,			/* Length of path_name */
    (*should_rest)();		/* Points to caller's should_rest routine */
struct gen_list
    dir_list,			/* List of directory file_info's */
    mf_list;			/* List of mf_info rec's */

/* Externs */
extern char
    wd[];			/* Our wd when we started program */
extern int
    wd_len,			/* strlen(wd) */
    uid;			/* Our uid */

/* Function decl's */
static int
    pass1_type_end(), pass1_afile(), pass1_cd_up(), pass1_cd_down(),
    pass1_cd_root(), pass1_symlink(), pass1_lfile(), pass1_mfile(),
    pass2_type_end(), pass2_afile(), pass2_cd_up(), pass2_cd_down(),
    pass2_cd_root(), pass2_symlink(), pass2_lfile(), pass2_mfile();
struct mf_info
    *find_mfile();
extern char
    *bsearch(), *mktemp(), *rindex(), *malloc(), *realloc(), *sprintf(),
    *strcpy();

/* More local vars */
static int
    (*(pass1_int[NTYPES]))() = {
	pass1_type_end, pass1_afile, pass1_cd_up, pass1_cd_down,
	pass1_cd_root, pass1_symlink, pass1_lfile, NULL, pass1_mfile,
	pass1_cd_down},
    (*(pass2_int[NTYPES]))() = {
	pass2_type_end, pass2_afile, pass2_cd_up, pass2_cd_down,
	pass2_cd_root, pass2_symlink, pass2_lfile, NULL, pass2_mfile,
	pass2_cd_down};


/* RESTORE: Do a restore.
 */
restore(_slist_ptr, pass1_ext, pass2_ext, _should_rest)
    char *_slist_ptr;
    void (**pass1_ext)(), (**pass2_ext)();
    int (*_should_rest)();
{
    int pass1_rd_hd(), pass2_rd_hd(), tape_rd_hd();
    extern int hash;

    slist_ptr = _slist_ptr;
    should_rest = _should_rest;
    init_list(&mf_list);
    init_list(&dir_list);

    /* If we have a savelist, do two passes */
    if (slist_ptr) {
	will_rest = 0;
	if (rest_pass(pass1_rd_hd, pass1_int, pass1_ext) == -1) return -1;
	if (!will_rest) return 0;
	slist_ptr = _slist_ptr;
	if (rest_pass(pass2_rd_hd, pass2_int, pass2_ext) == -1) return -1;
    } else {
	/* Restore completely from tape */
	if (rest_pass(tape_rd_hd, pass2_int, pass2_ext) == -1) return -1;
    }
    cleanup_mflist();
    delete_list(&dir_list);
    if (hash) pr("\n");
    return 0;
} /* restore */


/* REST_PASS: One restore pass.
 */
rest_pass(rd_hd, int_routs, ext_routs)
    int (*rd_hd)(), (**int_routs)();
    void (**ext_routs)();
{
    static int add_size[NTYPES] =
	       {0,1,0,0,0,1,0,0,1,0}; /* {A,M}_FILE, SYMLINK */
    struct file_info *file_hd;
    char file_name[MAXFILENAME];
    register int (*fp)();

    /* This is the offset the tape should be at when entering here */
    tp_offset = SAVE_HEADER_SIZE;
    del_dir_cnt = 0;
    path_len = 0;
    *path_name = '\0';

    for (;;) {
	/* Read header for this file */
	if (rd_hd(&file_hd, file_name) == -1) return -1;

	/* Check the record type before calling a function with it */
	if ((unsigned)file_hd->record_type >= NTYPES) {
	    pr("Record type out of range\n");
	    return -1;
	}

	/* Do external routine for this file */
	if (fp = (int(*)())ext_routs[file_hd->record_type]) {
	    (*fp)(file_hd, path_name, file_name);
	}

	/* Do internal routine for this file */
	if (fp = int_routs[file_hd->record_type]) {
	    switch((*fp)(file_hd, file_name)) {
		case RST_ALL_DONE:
		    return 0;
		case RST_SOME_ERR:
		    return -1;
		default:
		    break;
	    }
	}

	/* Add st_size to offsets if needed */
	if (add_size[file_hd->record_type]) {
	    tp_offset += convl(&file_hd->st_size);
	}
    }
} /* rest_pass */


/* PASS1_RD_HD: Read the file header from the savelist and return in hd and
 * name.
 */
pass1_rd_hd(hd, name)
    struct file_info **hd;
    char *name;
{
    /* Get header */
    *hd = (struct file_info *)slist_ptr;
    slist_ptr += FILE_INFO_SIZE;
    tp_offset += FILE_INFO_SIZE;

    /* Get name if needed */
    if (NEEDS_NAME((*hd)->record_type)) {
	bcopy(slist_ptr, name, (int)(*hd)->name_len);
	name[(*hd)->name_len] = '\0';
	slist_ptr += (*hd)->name_len;
	tp_offset += (*hd)->name_len;
    }
    return RST_OK;
} /* pass1_rd_hd */


/* PASS2_RD_HD: Read the file header from the savelist and return in hd and
 * name.  Also skip the appropriate amount on tape.
 */
pass2_rd_hd(hd, name)
    struct file_info **hd;
    char *name;
{
    /* Get header */
    *hd = (struct file_info *)slist_ptr;
    slist_ptr += FILE_INFO_SIZE;
    tp_offset += FILE_INFO_SIZE;
    (void) tp_seek(FILE_INFO_SIZE, 1);

    /* Get name if needed */
    if (NEEDS_NAME((*hd)->record_type)) {
	bcopy(slist_ptr, name, (int)(*hd)->name_len);
	name[(*hd)->name_len] = '\0';
	slist_ptr += (*hd)->name_len;
	tp_offset += (*hd)->name_len;
	(void) tp_seek((int)(*hd)->name_len, 1);
    }
    return RST_OK;
} /* pass2_rd_hd */


/* TAPE_RD_HD: Read the file header from the tape and return in hd and
 * name.  Return error codes on error.
 */
tape_rd_hd(hd, name)
    struct file_info **hd;
    char *name;
{
    static struct file_info hd_buf;

    /* Get header */
    if (tp_read((char*)&hd_buf, FILE_INFO_SIZE) != FILE_INFO_SIZE) return -1;
    tp_offset += FILE_INFO_SIZE;
    *hd = &hd_buf;

    /* Get name if needed */
    if (NEEDS_NAME(hd_buf.record_type)) {
	if (tp_read(name, (int)hd_buf.name_len) != (int)hd_buf.name_len) {
	    return -1;
	}
	name[hd_buf.name_len] = '\0';
	tp_offset += hd_buf.name_len;
    }
    return 0;
} /* tape_rd_hd */


/*****************************************************************************/
/* PASS1 Routines                                                            */
/*****************************************************************************/


/* PASS1_CD_ROOT: Delete current pathname.
 */
/*ARGSUSED*/
static int pass1_cd_root(hd, name)
    struct file_info *hd;
    char *name;
{
    del_dir_cnt = 0;
    path_len = 0;
    *path_name = '\0';
    return RST_OK;
} /* pass1_cd_root */


/* PASS1_CD_UP: Delete one file from pathname.
 */
/*ARGSUSED*/
static int pass1_cd_up(hd, name)
    struct file_info *hd;
    char *name;
{
    if (del_dir_cnt) {
	--del_dir_cnt;
    } else {
	remove_path();
    }
    return RST_OK;
} /* pass1_cd_up */


/* PASS1_CD_DOWN: Add a file to the path.
 */
static int pass1_cd_down(hd, name)
    struct file_info *hd;
    char *name;
{
    if (del_dir_cnt || !(*should_rest)(hd, path_name, name)) {
	++del_dir_cnt;
    } else {
	add_path(name, (int)hd->name_len);
    }
    return RST_OK;
} /* pass1_cd_down */


/* PASS1_MFILE: Add the mfile to the list of mfiles.
 */
static int pass1_mfile(hd, name)
    struct file_info *hd;
    char *name;
{
    int doit = !del_dir_cnt && (*should_rest)(hd, path_name, name);

    if (add_mfile(hd, name, doit) == -1) return RST_SOME_ERR;
    if (doit) will_rest = 1;
    return RST_OK;
} /* pass1_mfile */


/* PASS1_AFILE: Skip afile.
 */
static int pass1_afile(hd, name)
    struct file_info *hd;
    char *name;
{
    if (!will_rest) {
	will_rest = !del_dir_cnt && (*should_rest)(hd, path_name, name);
    }
    return RST_OK;
} /* pass1_afile */


/* PASS1_LFILE: If will be restoring this, mark corresponding mfile.
 */
static int pass1_lfile(hd, name)
    struct file_info *hd;
    char *name;
{
    struct mf_info *mfi;

    if (!del_dir_cnt && (*should_rest)(hd, path_name, name)) {
	mfi = find_mfile(convl(&hd->st_size));
	if (mfi) {
	    if (!mfi->restore) {
		/* Make sure the mfile will be restored into a temp file */
		mfi->istmp = mfi->restore = 1;
		(void) mktemp(mfi->name);
	    }
	    will_rest = 1;
	}
    }
    return RST_OK;
} /* pass1_lfile */


/* PASS1_SYMLINK: Skip a symlink.
 */
static int pass1_symlink(hd, name)
    struct file_info *hd;
    char *name;
{
    if (!will_rest) {
	will_rest = !del_dir_cnt && (*should_rest)(hd, path_name, name);
    }
    return RST_OK;
} /* pass1_symlink */


/* PASS1_TYPE_END: All done.
 */
/*ARGSUSED*/
static int pass1_type_end(hd, name)
    struct file_info *hd;
    char *name;
{
    return RST_ALL_DONE;
} /* pass1_type_end */


/*****************************************************************************/
/* PASS2 Routines                                                            */
/*****************************************************************************/


/* PASS2_CD_ROOT: Do a chdir to the root dir
 */
/*ARGSUSED*/
static int pass2_cd_root(hd, name)
    struct file_info *hd;
    char *name;
{
    /* Clean up dir's left on the dir_list */
    clean_dir_list();
    path_len = 0;
    *path_name = '\0';
    del_dir_cnt = 0;
    /* We should be at wd, but just in case */
    if (chdir(wd) == -1) {
	pr("Can't cd to ");
	prerror(wd);
	return RST_SOME_ERR;
    }
    return RST_OK;
} /* pass2_cd_root */


/* PASS2_CD_UP: Chdir up one level.
 */
/*ARGSUSED*/
static int pass2_cd_up(hd, name)
    struct file_info *hd;
    char *name;
{
    if (del_dir_cnt) {
	--del_dir_cnt;
    } else {
	/* Remove dir from current path name */
	remove_path();

	/* Set this dir's attributes and cd up */
	if (pop_dir() == -1) {
	    return RST_SOME_ERR;
	}
    }
    return RST_OK;
} /* pass2_cd_up */


/* PASS2_CD_DOWN: Make a directory and then cd into it.
 */
static int pass2_cd_down(hd, name)
    struct file_info *hd;
    char *name;
{
    int retval;
    struct file_info *fi;

    if (del_dir_cnt || !(*should_rest)(hd, path_name, name)) {
	++del_dir_cnt;
	return RST_OK;
    }

    /* Make the directory, remove file with same name if needed */
    for (retval = -100; retval == -100 && mkdir(name, 0700) == -1;) {
	switch(safe_rm(name)) {
	    /* Nothing to remove */
	    case -1:
		pr("Can't make directory %s/", path_name);
		prerror(name);
		retval = RST_SOME_ERR;
		break;
	    /* Already a directory */
	    case -2:
		retval = RST_OK;
		break;
	    /* Able to remove it */
	    default:
		break;
	}
    }
    /* Cd into the dir */
    if (retval != RST_SOME_ERR) {
	retval = RST_OK;
	if (chdir(name) == -1) {
	    pr("Can't chdir to %s/", path_name);
	    prerror(name);
	    retval = RST_SOME_ERR;
	}
    }
    /* Add dir to dir_list stack so we'll set it's attributes later */
    if (retval == RST_OK) {
	add_path(name, (int)hd->name_len);
	if ((fi = LIST_NEW(dir_list, struct file_info, 10)) == NULL) {
	    retval = RST_SOME_ERR;
	} else {
	    bcopy((char*)hd, (char*)fi, sizeof(struct file_info));
	}
    } else {
	++del_dir_cnt;
	retval = RST_OK;
    }
    return retval;
} /* cd_down */


/* PASS2_MFILE: Look up the mfile for this, and restore it if flagged.
 */
static int pass2_mfile(hd, name)
    struct file_info *hd;
    char *name;
{
    int doit;
    struct mf_info *mfi;
    char *s;

    mfi = find_mfile((tp_offset - FILE_INFO_SIZE) - (int)hd->name_len);
    if (mfi) {
	if (doit = mfi->restore) {
	    /* Do restore */
	    s = rindex(mfi->name, '/');
	    if (rest_afile(hd, s ? s+1 : mfi->name) == -1) return RST_SOME_ERR;
	}
    } else {
	/* Assume we're only doing pass2, add it to list now. */
	doit = !del_dir_cnt && (*should_rest)(hd, path_name, name);
	if (add_mfile(hd, name, doit) == -1) return RST_SOME_ERR;

	if (doit) {
	    /* Do restore */
	    if (rest_afile(hd, name) == -1) return RST_SOME_ERR;
	}
    }
    if (!doit) {
	(void) tp_seek(convl(&hd->st_size), 1);
    }
    return RST_OK;
} /* pass2_mfile */


/* PASS2_AFILE: Decide whether to restore a file and do it if needed.
 */
static int pass2_afile(hd, name)
    struct file_info *hd;
    char *name;
{
    if (!del_dir_cnt && (*should_rest)(hd, path_name, name)) {
	/* Do restore of file */
	if (rest_afile(hd, name) == -1) return RST_SOME_ERR;
    } else {
	(void) tp_seek(convl(&hd->st_size), 1);
    }
    return RST_OK;
} /* pass2_afile */


/* PASS2_LFILE: Restore a link to an mfile.
 */
static int pass2_lfile(hd, name)
    struct file_info *hd;
    char *name;
{
    char full_name[MAXDIRLENGTH];
    int infd, outfd, saved_err, l;
    struct mf_info *mfi;

    /* If should restore this */
    if (!del_dir_cnt && (*should_rest)(hd, path_name, name)) {

	/* Find corresponding mfile */
	mfi = find_mfile(convl(&hd->st_size));
	if (mfi == NULL || !mfi->restore) {
	    pr("Can't restore link %s/%s due to missing file\n",
	       path_name, name);
	    return RST_OK;
	}

	/* Make full path of file to link to */
	bcopy(wd, full_name, l = wd_len);
	if (wd_len > 1) full_name[l++] = '/';
	(void) strcpy(&full_name[l], mfi->name);

	/* Try to make hard link to mfile */
	if (link(full_name, name) != -1) return RST_OK;
	if (!safe_rm(name)) {
	    if (link(full_name, name) != -1) return RST_OK;
	}

	/* If not a temp file, try to make a symlink to mfile */
	if (!mfi->istmp) {
	    saved_err = errno;
	    if (symlink(full_name, name) != -1) {
		pr("Couldn't make link from %s/%s to %s, making symlink\n",
		   path_name, name, full_name);
		return RST_OK;
	    }
	    errno = saved_err;
	    pr("Can't make link %s/%s to ", path_name, name);
	    prerror(full_name);
	}

	/* Copy mfile to this filename */
	infd = open(full_name, O_RDONLY);
	if (infd == -1) {
	    pr("Can't copy file ");
	    prerror(full_name);
	    return RST_OK;
	}

	/* Open output */
	while ((outfd = open(name, O_WRONLY|O_CREAT|O_EXCL, 0)) == -1) {
	    if (safe_rm(name)) {
		pr("Can't create ");
		prerror(name);
		(void) close(infd);
		return RST_OK;
	    }
	}
	cp(infd, outfd);
    }
    return RST_OK;
} /* pass2_lfile */


/* PASS2_SYMLINK: Restore a symlink.
 */
static int pass2_symlink(hd, name)
    struct file_info *hd;
    char *name;
{
    char link_buf[MAXPATHLEN];
    int n, sz = convl(&hd->st_size);

    if (!del_dir_cnt && (*should_rest)(hd, path_name, name)) {
	/* Read the link */
	if ((n = tp_read(link_buf, sz)) == -1) return RST_SOME_ERR;
	if (n != sz) {
	    pr("Can't restore symlink %s/%s due to tape error\n",
	       path_name, name);
	} else {
	    link_buf[sz] = '\0';
	    /* Make the link on the filesystem */
	    while(symlink(link_buf, name) == -1) {
		if (safe_rm(name)) {
		    pr("Can't create symlink %s/", path_name);
		    prerror(name);
		}
	    }
	}
    } else {
	(void) tp_seek(sz, 1);
    }
    return RST_OK;
} /* pass2_symlink */


/* PASS2_TYPE_END: All done.
 */
/*ARGSUSED*/
static int pass2_type_end(hd, name)
    struct file_info *hd;
    char *name;
{
    /* Make sure we set attributes on all dirs */
    clean_dir_list();
    return RST_ALL_DONE;
} /* pass2_type_end */


/* REST_AFILE: Do the restore of a normal file.
 */
rest_afile(hd, name)
    struct file_info *hd;
    char *name;
{
#ifdef SUN40
    char *mapadr;
#endif
    int fd, retval = RST_OK, had_err = 0;
    int size = convl(&hd->st_size);

    /* Create the file */
    while ((fd = open(name, O_RDWR|O_CREAT|O_EXCL, 0600)) == -1) {
	if (safe_rm(name)) {
	    pr("Can't create ");
	    prerror(name);
	    return RST_OK;
	}
    }

#ifdef SUN40
    if (size > IOBUFSIZ) {
	/* Trunc to right size so we can mmap it */
	(void) ftruncate(fd, (off_t)size);

	/* Mmap the file into memory */
	mapadr = mmap((caddr_t)0, size, PROT_WRITE, MAP_SHARED, fd, (off_t)0);

	/* If can't mmap */
	if (mapadr == (caddr_t)-1) {
	    prerror("Warning, mmap");
	    if (rest_afile_cp(size, fd, &had_err) == -1) retval = RST_SOME_ERR;
	} else {
	    /* Read into mmap'ed file */
	    switch(tp_read(mapadr, size)) {
		case -1:
		    retval = RST_SOME_ERR;
		    break;
		case 0:
		    had_err = 1;
		    break;
		default:
		    break;
	    }
	    if (munmap(mapadr, (int)size) == -1) {
		prerror("munmap");
		retval = RST_SOME_ERR;
	    }
	}
    } else {
	if (rest_afile_cp(size, fd, &had_err) == -1) retval = RST_SOME_ERR;
    }
#else /* Non SUN40 */
    if (rest_afile_cp(size, fd, &had_err) == -1) retval = RST_SOME_ERR;
#endif

    /* Set the attributes for this file */
    if (retval == RST_OK) {
	set_attr(hd, name, fd);
    }

    /* Close file */
    if (close(fd) == -1) {
	prerror("close");
	retval = RST_SOME_ERR;
    }

    if (had_err) {
	pr("Tape error occurred during read of %s/%s, file is damaged.\n",
	   path_name, name);
    }

    return retval;
} /* rest_afile */


/* REST_AFILE_CP: Do a copy for a rest_afile.
 */
rest_afile_cp(size, fd, had_err)
    int size, fd, *had_err;
{
    char buf[IOBUFSIZ];
    register int sz, n;

    if (!size) return 0;

    /* Do a traditional read and write */
    do {
	sz = min(size, sizeof(buf));
	if ((n = tp_read(buf, sz)) == -1) return -1;
	if (n == 0) *had_err = 1;
	if (write(fd, buf, sz) != sz) {
	    prerror("disk write");
	    return -1;
	}
    } while (size -= n);
    return 0;
} /* rest_afile_cp */


/* CLEAN_DIR_LIST: Pop all dirs off of dir_list.
 */
clean_dir_list()
{
    while (LIST_SZ(dir_list)) {
	if (pop_dir() == -1) {
	    LIST_SZ(dir_list) = 0;
	    return;
	}
    }
} /* clean_dir_list */


/* POP_DIR: Pull one dir off of dir_list, set it's attributes and chdir up.
 * If we can't cd up, return -1, else return 0.
 */
pop_dir()
{
    if (LIST_SZ(dir_list)) {
	/* Set attributes before we leave here */
	set_attr(&LIST_GET(dir_list, struct file_info, LIST_SZ(dir_list)-1),
		 ".", -1);

	/* Pop */
	LIST_DEL(dir_list);

	/* CD up */
	if (chdir("..") == -1) {
	    if (chdir(wd) == -1 || chdir(path_name) == -1) {
		pr("Can't chdir to %s/", wd);
		prerror(path_name);
		return -1;
	    }
	}
    }
    return 0;
} /* pop_dir */


/* ADD_MFILE: Add an mfile to the mf_list.
 */
add_mfile(hd, name, will_doit)
    struct file_info *hd;
    char *name;
    int will_doit;
{
    struct mf_info mfi;
    char *s;

    /* Add mfile's name to current path to get full name of file */
    add_path(name, (int)hd->name_len);
    s = NEW(char, path_len + 1 + (will_doit ? 0 : 6));
    if (s == NULL) {
	pr("Out of memory\n");
	return -1;
    }

    /* Fill in the mfi record */
    mfi.offset = (tp_offset - FILE_INFO_SIZE) - hd->name_len;
    mfi.restore = will_doit;
    mfi.istmp = 0;
    (void) sprintf(mfi.name = s, (will_doit ? "%s" : "%sXXXXXX"), path_name);

    /* Add record to list */
    if (LIST_ADD(mf_list, struct mf_info, mfi, 50) == -1) return -1;

    /* Clean up path*/
    remove_path();
    return 0;
} /* add_mfile */


/* FIND_MFILE: Find an mfile in the mfile list by searching for this offset.
 */
struct mf_info *find_mfile(offset)
    int offset;
{
    struct mf_info key;
    int find_mfile_compare();

    if (LIST_SZ(mf_list) == 0) return NULL;
    key.offset = offset;
    return (struct mf_info*)
	bsearch((char*)&key, (char*)&LIST_GET(mf_list, struct mf_info, 0),
        (unsigned)LIST_SZ(mf_list), sizeof(key), find_mfile_compare);
} /* find_mfile */


/* FIND_MFILE_COMPARE: Comparison routine for find_mfile.
 */
find_mfile_compare(mfi1, mfi2)
    register struct mf_info *mfi1, *mfi2;
{
    return mfi1->offset - mfi2->offset;
} /* find_mfile_compare */


/* CLEANUP_MFLIST: Clean up the mflist.
 */
cleanup_mflist()
{
    int i, l;
    char full_name[MAXDIRLENGTH];
    struct mf_info *mfi;

    /* Put wd into full_name */
    bcopy(wd, full_name, l = wd_len);
    if (wd_len > 1) full_name[l++] = '/';

    /* Go around and remove temporarily restored mfiles */
    for (i = 0; i < LIST_SZ(mf_list); ++i) {
	mfi = &LIST_GET(mf_list, struct mf_info, i);
	if (mfi->restore && mfi->istmp) {
	    (void) strcpy(&full_name[l], mfi->name);
	    (void) safe_rm(full_name);
	}
	free(mfi->name);
    }
    delete_list(&mf_list);
} /* cleanup_mfilst */


/* REMOVE_PATH: Remove the last directory from the path_name.
 */
remove_path()
{
    register char *s = rindex(path_name, '/');
    path_len = s ? s - path_name : 0;
    path_name[path_len] = '\0';
} /* remove_path */


/* ADD_PATH: Add a dir to the end of path_name.
 */
add_path(dir, len)
    register char *dir;
    register int len;
{
    /* Strip leading /'s from dir */
    while (len && *dir == '/') {
	++dir;
	--len;
    }

    /* Add / to end of path_name (only if path already has something in it) */
    if (path_len) path_name[path_len++] = '/';

    /* Append dir to path_name */
    bcopy(dir, &path_name[path_len], len);
    path_name[path_len += len] = '\0';
} /* add_path */


/* CP: Copy from file to file.
 */
cp(in, out)
    int in, out;
{
    int n;
    char buf[IOBUFSIZ];

    while ((n = read(in, buf, sizeof(buf))) > 0) {
	if (write(out, buf, n) != n) {
	    prerror("write");
	}
    }
    if (n == -1) prerror("read");
} /* cp */


/* SET_ATTR: Set all the attributes for a given file.  If fd is not -1, then
 * it is used for faster sys calls.
 */
set_attr(hd, name, fd)
    struct file_info *hd;
    char *name;
    int fd;
{
    time_t tv[2];

    if (hd->record_type != SYMLINK) {
	if (!uid && -1 == (fd == -1
	    ? chown(name, convs(&hd->st_uid), convs(&hd->st_gid))
	    : fchown(fd, convs(&hd->st_uid), convs(&hd->st_gid))))
	{
	    pr("Can't chown ");
	    prerror(name);
	}
	if (-1 == (fd == -1
	    ? chmod(name, convs(&hd->st_mode) & ~S_IFMT)
	    : fchmod(fd, convs(&hd->st_mode) & ~S_IFMT)))
	{
	    pr("Can't chmod ");
	    prerror(name);
	}
	tv[0] = convl(&hd->st_atime);
	tv[1] = convl(&hd->st_mtime);
	if (utime(name, tv) == -1) {
	    pr("Can't set times for ");
	    prerror(name);
	}
    }
} /* set_attr */


/* SAFE_RM: Remove something if it's not a directory.  If it is a directory,
 * return -2.  If it can't be removed, return -1, else return 0.
 */
safe_rm(name)
    char *name;
{
    extern int errno;
    int saved = errno, st = -1;
    struct stat stbuf;

    if (lstat(name, &stbuf) != -1) {
        if ((stbuf.st_mode & S_IFMT) == S_IFDIR) {
            st = -2;
        } else if (unlink(name) == -1) {
            pr("Can't remove ");
            prerror(name);
        } else {
            st = 0;
        }
    }
    errno = saved;
    return st;
} /* safe_rm */
