/*******************************************************************/
/*      Copyright (C) 1989 by                                      */
/*            The Regents of the University of California          */
/*                                                                 */
/*                             DISCLAIMER                          */
/*                                                                 */
/*      "Although   each   program   has   been  tested  by  its   */
/*      contributor, no warranty, express or implied, is made by   */
/*      the  contributor  or the University of California, as to   */
/*      the accuracy and functioning of the program and  related   */
/*      program material, nor shall the fact of the distribution   */
/*      constitute any such warranty, and no  responsibility  is   */
/*      assumed   by   the  contributor  or  the  University  of   */
/*      California in connection herewith."                        */
/*******************************************************************/

/* DUMPER.C: Simple save dumper
 *
 * $Header: /cd/ua/sam/src/ss/rest/RCS/dumper.c,v 1.3 89/06/09 15:51:06 sam REL-6-9-89 $
 *
 * $Log:	dumper.c,v $
 * Revision 1.3  89/06/09  15:51:06  sam
 * Fixed to exit after an end record.
 * 
 * Revision 1.2  89/04/03  21:30:59  sam
 * No changes
 * 
 * Revision 1.1  89/03/08  20:38:29  sam
 * Initial revision
 * 
 */
#ifndef lint
static char rcsid[] = "$Header: /cd/ua/sam/src/ss/rest/RCS/dumper.c,v 1.3 89/06/09 15:51:06 sam REL-6-9-89 $";
#endif

#include "../global.h"
#include "rest.h"
#include <stdio.h>
#include <sys/stat.h>
#include <sys/mman.h>

#define INFO(p) ((struct file_info*)p)

main(argc, argv)
    int argc;
    char **argv;
{
    char *p, *q, *end, *adr, *malloc();
    int type, fd, offset, len, fromtape;
    struct stat stbuf;
    struct save_header *hd;

    if (argc > 1) {
	fd = open(argv[1], 0);
	if (fd == -1) {
	    perror(argv[1]);
	    exit(1);
	}
    } else {
	fd = 0;
    }

    fstat(fd, &stbuf);

#ifndef SUN40
    adr = malloc(stbuf.st_size);
    if (adr == NULL) {
	fputs("out of memory", stderr);
	exit(1);
    }
    if (read(fd, adr, stbuf.st_size) != stbuf.st_size) {
	perror("read");
	exit(1);
    }
#else
    if ((adr = mmap(0, stbuf.st_size, PROT_READ, MAP_SHARED, fd, 0))
	== (char*)-1)
    {
	perror("mmap");
	exit(1);
    }
#endif

    hd = (struct save_header *)adr;
    fromtape = !(hd->type & SAVE_SLIST);
    if (fromtape) {
	fputs("Assuming file is full tape dump\n", stderr);
    }

    p = adr + SAVE_HEADER_SIZE;
    end = adr+(unsigned)stbuf.st_size;

    offset = SAVE_HEADER_SIZE;
    while (p < end) {
	type = INFO(p)->record_type;
	if (type < 0 || type >= NTYPES) {
	    (void) fprintf(stderr, "Bad type %d\n", type);
	    exit(1);
	}
	(void) printf("TYPE=%c", "EAUDRSLIMP"[type]);
	if (type != CD_UP && type != CD_ROOT && type != TYPE_END) {
	    (void) printf(", NAME=");
	    fwrite(INFONAME(p), INFO(p)->name_len, 1, stdout);
	}
	(void) printf(type == L_FILE ? ", A_FILE_OFF=%d, " : ", SIZE=%d, ",
	    convl(&INFO(p)->st_size));
	(void) printf("OFF=%d\n", offset);
	switch(type) {
	    case TYPE_END:
		exit(0);
	    case CD_UP:
	    case CD_ROOT:
		offset += FILE_INFO_SIZE;
		p = INFONAME(p);
		break;
	    case CD_DOWN:
	    case PATH_DIR:
	    case L_FILE:
	    case I_FILE:
		offset += FILE_INFO_SIZE + INFO(p)->name_len;
		p = INFONAME(p) + INFO(p)->name_len;
		break;
	    case SYMLINK:
	    case M_FILE:
	    case A_FILE:
		offset += FILE_INFO_SIZE + INFO(p)->name_len +
			    (unsigned)convl(&INFO(p)->st_size);
		p = INFONAME(p) + INFO(p)->name_len +
		    (fromtape ? convl(&INFO(p)->st_size) : 0);
		break;
	    default:
		fputs("Unknown record type %d=n", type);
		exit(1);
		break;
	}
    }
} /* main */
