/*******************************************************************/
/*      Copyright (C) 1989 by                                      */
/*            The Regents of the University of California          */
/*                                                                 */
/*                             DISCLAIMER                          */
/*                                                                 */
/*      "Although   each   program   has   been  tested  by  its   */
/*      contributor, no warranty, express or implied, is made by   */
/*      the  contributor  or the University of California, as to   */
/*      the accuracy and functioning of the program and  related   */
/*      program material, nor shall the fact of the distribution   */
/*      constitute any such warranty, and no  responsibility  is   */
/*      assumed   by   the  contributor  or  the  University  of   */
/*      California in connection herewith."                        */
/*******************************************************************/

/* INIT.C: Initialization Routines.
 *
 * $Header: /cd/ua/sam/src/ss/rest/RCS/init.c,v 1.3 89/06/09 15:52:00 sam REL-6-9-89 $
 *
 * $Log:	init.c,v $
 * Revision 1.3  89/06/09  15:52:00  sam
 * No changes
 * 
 * Revision 1.2  89/04/03  21:04:00  sam
 * Hashing now on by default in interactive mode.
 * 
 * Revision 1.1  89/03/08  20:38:38  sam
 * Initial revision
 * 
 */
#ifndef lint
static char rcsid[] = "$Header: /cd/ua/sam/src/ss/rest/RCS/init.c,v 1.3 89/06/09 15:52:00 sam REL-6-9-89 $";
#endif

#include "../global.h"
#include "rest.h"
#include <sys/stat.h>
#include <stdio.h>
#include <signal.h>
#include <netdb.h>

extern int
    bytes_per_tape,		/* Number of bytes per tape */
    blk_size,			/* Tape block size */
    rest_mode,			/* Type of restore being done */
    remote,			/* Is restore from remote tape? */
    nrest_paths,		/* Number of paths in rest_paths */
    shell_port,			/* Port for doing rsh */
    wd_len,			/* strlen(wd) */
    beg_file_num,		/* File num of restore file on tape */
    chat,			/* Talk w/the user? */
    hash,			/* Print hash marks? */
    uid,
    pid,		
    pagesize,
    dtablesize;
extern char
    *tp_buf,			/* Tape buffer */
    *tapehost,			/* Host where remote tape is */
    *tapedev,			/* Name of tape device */
    *save_list,			/* Name of savelist file */
    *save_list_dir,		/* Dir where savelists are */
    **rest_paths,		/* Paths to restore */
    wd[];			/* Our starting wd */

/* Function decls */
extern char
    *getwd(), *index(), *malloc();

/* INITIALIZE: Set up vars, read command line and config file.
 */
initialize(argc, argv)
    int argc;
    char **argv;
{
    struct servent *sp, *getservbyname();
    char *s;

    /* Miscellaneous stuff */
    dtablesize = getdtablesize();
#if defined(sequent) && defined(i386)
    /* getpagesize() doesn't work for mmaps on symmetry, but this does */
    pagesize = 8192;
#else
    pagesize = getpagesize();
#endif
    uid = getuid();
    pid = getpid();
    (void) umask(022);
    (void) getwd(wd);
    wd_len = strlen(wd);

    /* Parse command line */
    parse_args(argc, argv);

    /* Give default values to vars not set on the command line */
    if (!tapedev) tapedev = TAPEDEV;
    if (!save_list_dir) save_list_dir = SAVELISTDIR;
    if (!blk_size) blk_size = BLKSIZ;
    bytes_per_tape = BYTESPT;

    /* Check for mutually exclusive options */
    if (save_list && beg_file_num) {
	pr("Use only one of the 's' and 'l' options.\n");
	exit(1);
    }

    /* Set chat if we should chat with user */
    chat = isatty(0) && isatty(1);

    /* Make a tape buffer with size from blk_size */
    if ((tp_buf = NEW(char, blk_size)) == NULL) {
	pr("Out of memory\n");
	exit(1);
    }

    /* Get shell port */
    if ((sp = getservbyname("shell", "tcp")) == NULL) {
	prerror("getservbyname(shell,tcp)");
	exit(1);
    }
    shell_port = sp->s_port;

    /* Look at tapedev to see if restore is remote, and start rmt if so */
    if (s = index(tapedev, ':')) {
	remote = 1;
	*s = '\0';
	if (start_rmt(tapedev) == -1) exit(1);
	tapedev = s + 1;
    }

    /* Set up signals */
    reset_sigs();
} /* initialize */


/* PARSE_ARGS: Parse up the command line.
 */
parse_args(argc, argv)
    char **argv;
{
    char *opt;

    rest_mode = -1;
    tapedev = NULL;
    blk_size = 0;
    beg_file_num = 0;
    save_list = NULL;
    save_list_dir = NULL;
    hash = 0;

    if (argc <= 1) usage();

    opt = &argv[1][0];
    if (*opt == '-') ++opt;
    argc -= 2;
    argv += 2;

    switch(opt ? *opt++ : '\0') {
	case 'i':
	    if (rest_mode != -1) usage();
	    rest_mode = MODE_INTER;
	    hash = 1;
	    break;
	case 'r':
	    if (rest_mode != -1) usage();
	    rest_mode = MODE_FULL;
	    break;
	case 't':
	    if (rest_mode != -1) usage();
	    rest_mode = MODE_TOC;
	    break;
	case 'x':
	    if (rest_mode != -1) usage();
	    rest_mode = MODE_EXTRACT;
	    break;
	case 'X':
	    if (rest_mode != -1) usage();
	    rest_mode = MODE_SAVELIST;
	    break;
	default:
	    usage();
    }

    while (*opt) switch(*opt++) {
	case 'f':
	    if (!argc--) usage();
	    tapedev = *argv++;
	    break;
	case 's':
	    if (!argc--) usage();
	    beg_file_num = atoi(*argv++);
	    break;
	case 'L':
	    if (!argc--) usage();
	    save_list_dir = *argv++;
	    break;
	case 'l':
	    if (!argc--) usage();
	    save_list = *argv++;
	    break;
	case 'b':
	    if (!argc--) usage();
	    blk_size = atoi(*argv++);
	    break;
	case 'h':
	    hash = !hash;
	    break;
	default:
	    usage();
    }
    rest_paths = argv;
    nrest_paths = argc;
} /* parse_args */


/* USAGE: Print a nasty message and exit.
 */
usage()
{
    pr("\
usage: rest [-]irtxX[bfsLlh] [block-size] [dump-file] [file-skip]\n\
       [savelist-dir] [savelist-name] [paths]\n");
    exit(1);
} /* usage */


/* RESET_SIGS: Reset all the signals.
 */
reset_sigs()
{
    sigtype coredumper();

    (void) signal(SIGHUP, SIG_IGN);
    (void) signal(SIGINT, SIG_IGN);
    (void) signal(SIGQUIT, SIG_IGN);
    (void) signal(SIGINT, SIG_DFL);
    (void) signal(SIGSEGV, coredumper);
} /* reset_sigs */


/* COREDUMPER: Cd back to our starting directory and dump core.
 */
sigtype coredumper()
{
    (void) chdir(wd);
    (void) signal(SIGSEGV, SIG_DFL);
    (void) kill(getpid(), SIGSEGV);
} /* coredumper */
