/*******************************************************************/
/*      Copyright (C) 1989 by                                      */
/*            The Regents of the University of California          */
/*                                                                 */
/*                             DISCLAIMER                          */
/*                                                                 */
/*      "Although   each   program   has   been  tested  by  its   */
/*      contributor, no warranty, express or implied, is made by   */
/*      the  contributor  or the University of California, as to   */
/*      the accuracy and functioning of the program and  related   */
/*      program material, nor shall the fact of the distribution   */
/*      constitute any such warranty, and no  responsibility  is   */
/*      assumed   by   the  contributor  or  the  University  of   */
/*      California in connection herewith."                        */
/*******************************************************************/

/* INTER: Interactive restore
 *
 * $Header: /cd/ua/sam/src/ss/rest/RCS/inter.c,v 1.2 89/06/09 15:54:24 sam REL-6-9-89 $
 *
 * $Log:	inter.c,v $
 * Revision 1.2  89/06/09  15:54:24  sam
 * Implemented extract command.
 * 
 * Revision 1.1  89/04/03  21:04:26  sam
 * Initial revision
 * 
 */
#ifndef lint
static char rcsid[] = "$Header: /cd/ua/sam/src/ss/rest/RCS/inter.c,v 1.2 89/06/09 15:54:24 sam REL-6-9-89 $";
#endif

#include "../global.h"
#include "rest.h"
#include <stdio.h>

#ifdef SUN40
#include <dirent.h>
#else
#include <sys/dir.h>
#endif

/* DENT_TYPE: type of structure returned by readdir.
 */
#ifdef SUN40
    typedef struct dirent dent_type;
#else
    typedef struct direct dent_type;
#endif

/* ISDOTFILE: Return true if the string given is . or ..
 */
#define ISDOTFILE(s) \
	 (*s == '.' && (s[1] == '\0' || (s[1] == '.' && s[2] == '\0')))

#define DIR_INC 30		/* Increments to add to a dir */
#define FE_POOL_SZ 1000
#define DE_POOL_SZ 100

/* Local vars */
struct gen_list
    dir_stack,			/* Stack of dirs saved during restores */
    slist_list;			/* List of slist_info structures */
struct file_ent
    **lookup_begin,		/* Vars for the lookup_fi routine */
    **lookup_end,
    **lookup_cur,
    *rest_root,			/* Root of fs being restored */
    *cur_dir;			/* Track cur_dir during restore */
struct file_info
    *cur_dir_fi;		/* Fi of cur_dir for use in should_rest */

/* Extern vars */
extern char
    *save_list_dir;		/* Directory where savelists are */

/* Function decls */
static int
    list_cmd(), search_cmd(), browse_cmd(), extract_cmd(), help_cmd(),
    quit_cmd();
extern char
    *ctime(), *malloc(), *sprintf();
struct file_ent
    *lookup_fi();


static struct cmd_rec cmds[] = {
    {"list       - list all saves", list_cmd},
    {"search     - search saves for a file", search_cmd},
    {"browse     - browse through a save", browse_cmd},
    {"extract    - extract marked files", extract_cmd},
    {"help       - prints this message", help_cmd},
    {"?          - prints this message", help_cmd},
    {"quit       - quit restore", quit_cmd},
};

/* INTERACTIVE: Start interactive mode.
 */
interactive()
{
    if (read_slist_dir() == -1) finish_up();
    shell("restore> ", NUM_ARR(cmds), cmds, (int(*)())NULL);
} /* interactive */


/* HELP_CMD: Print the help message
 */
/*ARGSUSED*/
static int help_cmd(argc, argv, out)
    int argc;
    char **argv;
    FILE *out;
{
    int i;

    for (i = 0; i < NUM_ARR(cmds); ++i) {
	(void) fprintf(out, "%s\n", cmds[i].name);
    }
    return 0;
} /* help_cmd */


/* LIST_CMD: List all of the savelists.
 */
/*ARGSUSED*/
static int list_cmd(argc, argv, out)
    int argc;
    char **argv;
    FILE *out;
{
    int i, t;
    char *s;
    struct slist_info *info;

    (void) fputs(
	" Num  Type           Began                        Name\n", out);
    (void) fputs(
	" ===  ====  ========================  =============================\n",
	out);
    for (i = 0; i < LIST_SZ(slist_list); ++i) {
	info = &LIST_GET(slist_list, struct slist_info, i);
	t = convl(&info->hd->begin_time);
	(s = ctime((long*)&t))[24] = '\0';
	(void) fprintf(out, "%c%3d  %s  %s  %s\n", info->num_marked ? '+' : ' ',
	        i+1,
		(info->hd->type & SAVE_INC) ? "Inc " : "Full",
		s, info->hd->slistname);
    }
    return 0;
} /* list_cmd */


/* QUIT_CMD: Quit.
 */
/*ARGSUSED*/
static int quit_cmd(argc, argv, out)
    int argc;
    char **argv;
    FILE *out;
{
    return -1;
} /* quit_cmd */


/*ARGSUSED*/
static int search_cmd(argc, argv, out)
    int argc;
    char **argv;
    FILE *out;
{
    fputs("Not implemented.\n", out);
} /* search_cmd */


/* BROWSE_CMD: Browse through a savelist
 */
/*ARGSUSED*/
static int browse_cmd(argc, argv, out)
    int argc;
    char **argv;
    FILE *out;
{
    int i;
    struct slist_info *info;

    if (argc != 2 || argv[1][0] < '0' || argv[1][0] > '9') {
	pr("Usage: browse number\n");
    } else {
	i = atoi(argv[1]);
	if (i <= 0 || i > LIST_SZ(slist_list)) {
	    pr("Number out of range\n");
	} else {
	    info = &LIST_GET(slist_list, struct slist_info, i-1);
	    if (get_slist(info) != -1) {
		browse(info);
		if (!info->num_marked) unget_slist(info);
	    }
	}
    }
} /* browse_cmd */


/*ARGSUSED*/
static int extract_cmd(argc, argv, out)
    int argc;
    char **argv;
    FILE *out;
{
    int i, n;

    if (argc > 1) {
	for (i = 1; i < argc; ++i) {
	    if (argv[i][0] < '0' || argv[i][0] > '9') {
		pr("Usage: extract [save numbers]\n");
		return;
	    }
	}
	for (i = 1; i < argc; ++i) {
	    n = atoi(argv[i]);
	    if (n < 1 || n > LIST_SZ(slist_list)) {
		pr("%d: out of range\n", n);
	    } else {
		if (LIST_GET(slist_list, struct slist_info, n-1).num_marked) {
		    inter_rest(&LIST_GET(slist_list, struct slist_info, n-1));
		} else {
		    pr("%d: no files marked\n", n);
		}
	    }
	}
    } else {
	for (n = 0; n < LIST_SZ(slist_list); ++n) {
	    if (LIST_GET(slist_list, struct slist_info, n).num_marked) {
		inter_rest(&LIST_GET(slist_list, struct slist_info, n));
	    }
	}
    }
} /* extract_cmd */


/* READ_SLIST_DIR: Read all savelists in the savelist directory.
 */
read_slist_dir()
{
    struct slist_info info;
    int fd, dirnmlen, read_slist_dir_cmp();
    dent_type *dent;
    DIR *dirfp;

    init_list(&slist_list);
    dirnmlen = strlen(save_list_dir);
    dirfp = opendir(save_list_dir);
    if (dirfp == NULL) {
	perror(save_list_dir);
	return -1;
    }

    while (dent = readdir(dirfp)) {
	/* If good entry */
	if (dent->d_fileno != 0 && ! ISDOTFILE(dent->d_name)) {

	    /* Set up the info record */
	    info.ld_info.ld_point = NULL;
	    info.num_marked = 0;
	    info.file_name = NEW(char, dirnmlen + strlen(dent->d_name) + 2);
	    info.hd = NEW(struct save_header, 1);
	    if (info.hd == NULL || info.file_name == NULL) {
		pr("Out of memory\n");
		return -1;
	    }
	    (void)sprintf(info.file_name, "%s/%s", save_list_dir, dent->d_name);

	    /* Open file and read in header */
	    if ((fd = open(info.file_name, 0)) == -1 ||
	        read(fd, (char*)info.hd, SAVE_HEADER_SIZE) == -1)
	    {
		prerror(info.file_name);
		continue;
	    }
	    (void) close(fd);

	    /* Add this to the slist_list */
	    if (LIST_ADD(slist_list, struct slist_info, info, 5) == -1) {
		return -1;
	    }
	}
    }
    (void) closedir(dirfp);

    /* Sort the slist_list by time */
    qsort((char*)&LIST_GET(slist_list, struct slist_info, 0),
	  LIST_SZ(slist_list), sizeof(struct slist_info), read_slist_dir_cmp);
    return 0;
} /* read_slist_dir */


/* READ_SLIST_DIR_CMP: Compare routine to sort slist_info recs by time.
 */
read_slist_dir_cmp(p1, p2)
    struct slist_info *p1, *p2;
{
    return p1->hd->begin_time - p2->hd->begin_time;
} /* read_slist_dir_cmp */


/* GET_SLIST: Load up a savelist.
 */
get_slist(info)
    struct slist_info *info;
{
    /* Check if already loaded */
    if (info->ld_info.ld_point) return 0;

    /* Get the slist loaded */
    if (ld_slist(&info->ld_info, info->file_name) == -1) return -1;

    /* Set up rest of slist_info struct */
    info->root = NULL;
    info->num_marked = 0;
    init_pool(&info->pool);

    /* Build up the fs around the slist */
    return build_fs(info);
} /* get_slist */


/* UNGET_SLIST: Free up everything in the slist_info given
 */
unget_slist(info)
    struct slist_info *info;
{
    if (!info->ld_info.ld_point) return;

    /* Free up the fs */
    free_dir(info->root);

    /* Free up the mem pools */
    delete_pool(&info->pool);

    /* Unmap the savelist */
    unld_slist(&info->ld_info);
    info->ld_info.ld_point = NULL;
} /* unget_slist */


/* FREE_DIR: Free all the dir's under this one, then free this dir.
 */
free_dir(dir)
    struct file_ent *dir;
{
    struct gen_list *list = &dir->info_u.dir->file_list;
    register struct file_ent **fe, **end_fe;

    end_fe = &LIST_GET(*list, struct file_ent*, LIST_SZ(*list));
    for (fe = &LIST_GET(*list, struct file_ent*, 0); fe < end_fe; ++fe)
    {
	if ((*fe)->flags & FE_ISDIR) {
	    free_dir(*fe);
	}
    }
    delete_list(list);
} /* free_dir */


/* BUILD_FS: Build the internal fs structure around an slist.
 */
build_fs(info)
    struct slist_info *info;
{
    register struct file_info *file_hd;
    register struct file_ent *cdir_ent;
    register struct dir_ent *de, *cdir;
    register struct pool *pool = &info->pool;
    struct file_ent *fe;

    /* Set up the root */
    if ((info->root = POOL_NEW(*pool, struct file_ent)) == NULL
     || (info->root->info_u.dir = POOL_NEW(*pool, struct dir_ent)) == NULL)
    {
	return -1;
    }
    info->root->flags = FE_ISDIR;
    info->root->info_u.dir->parent = info->root;
    info->root->info_u.dir->info = NULL;
    init_list(&info->root->info_u.dir->file_list);

    /* Loop through the savelist */
    cdir_ent = info->root;
    cdir = cdir_ent->info_u.dir;
    for (file_hd = (struct file_info*)
		   (info->ld_info.ld_point + SAVE_HEADER_SIZE);
	 ;
	 file_hd = (struct file_info*)
	           (((char*)file_hd) + FILE_INFO_SIZE +
	           (NEEDS_NAME(file_hd->record_type) ? file_hd->name_len : 0)))
    {
	switch(file_hd->record_type) {

	case TYPE_END:
	    goto end_o_loop;
	case CD_UP:
	    cdir_ent = cdir->parent;
	    cdir = cdir_ent->info_u.dir;
	    break;
	case CD_ROOT:
	    cdir_ent = info->root;
	    cdir = cdir_ent->info_u.dir;
	    break;
	case I_FILE:
	    break;
	case PATH_DIR:
	case CD_DOWN:
	    fe = POOL_NEW(*pool, struct file_ent);
	    de = POOL_NEW(*pool, struct dir_ent);
	    if (fe == NULL || de == NULL) return -1;
	    fe->flags = FE_ISDIR;
	    fe->info_u.dir = de;
	    if (LIST_ADD(cdir->file_list, struct file_ent*, fe, DIR_INC) == -1)
	    {
		return -1;
	    }
	    de->parent = cdir_ent;
	    de->info = file_hd;
	    init_list(&de->file_list);
	    cdir_ent = fe;
	    cdir = de;
	    break;
	case A_FILE:
	case M_FILE:
	case L_FILE:
	case SYMLINK:
	    fe = POOL_NEW(*pool, struct file_ent);
	    if (fe == NULL) return -1;
	    fe->flags = 0;
	    fe->info_u.file = file_hd;
	    if (LIST_ADD(cdir->file_list, struct file_ent*, fe, DIR_INC) == -1)
	    {
		return -1;
	    }
	    break;
	default:
	    pr("Unkown file type %d\n", file_hd->record_type);
	    return -1;
	} /* end of switch */
    }
end_o_loop:
    return 0;
} /* build_fs */


/* INTER_REST: Start a restore on the passed slist.
 */
inter_rest(info)
    struct slist_info *info;
{
    struct save_header *tp_hd, tp_hd_store;
    int inter_should_rest();
    void inter_cd_down(), inter_cd_up(), inter_cd_root();
    static void (*(funcs[NTYPES]))() = {
	NULL, NULL, inter_cd_up, inter_cd_down, inter_cd_root, NULL, NULL,
	NULL, NULL, inter_cd_down};

    switch(start_tape_from_slist(info->hd, tp_hd = &tp_hd_store)) {
	case -2: tp_hd = NULL; break;
	case -1: return -1;
	default: break;
    }
    if (tp_hd) {
	if (cmp_save_headers(info->hd, tp_hd) == -1) return -1;
    }
    rest_root = info->root;
    change_cur_dir(rest_root);
    cur_dir_fi = rest_root->info_u.dir->info;
    init_list(&dir_stack);
    if (restore(((char*)info->ld_info.ld_point) + SAVE_HEADER_SIZE,
	funcs, funcs, inter_should_rest) == -1)
    {
	return -1;
    }
    return 0;
} /* inter_rest */


/* INTER_CD_ROOT: Track a cd_root.
 */
/*ARGSUSED*/
void inter_cd_root(fi, path, name)
    struct file_info *fi;
    char *path, *name;
{
    change_cur_dir(rest_root);
    cur_dir_fi = rest_root->info_u.dir->info;
    LIST_SZ(dir_stack) = 0;
} /* inter_cd_root */


/* INTER_CD_DOWN: Track a cd_down during the restore.
 */
/*ARGSUSED*/
void inter_cd_down(fi, path, name)
    struct file_info *fi;
    char *path, *name;
{
    if (LIST_ADD(dir_stack, struct file_ent*, cur_dir, 20) == -1) {
	pr("Out of memory\n");
	finish_up();
    }
    change_cur_dir(lookup_fi(cur_dir_fi = fi));
} /* inter_cd_down */


/* INTER_CD_UP: Track a cd_up during the restore.
 */
/*ARGSUSED*/
void inter_cd_up(fi, path, name)
    struct file_info *fi;
    char *path, *name;
{
    change_cur_dir(
	LIST_GET(dir_stack, struct file_ent*, LIST_SZ(dir_stack)-1));
    LIST_DEL(dir_stack);
} /* inter_cd_up */


/* INTER_SHOULD_REST: Return a value telling whether this file should be
 * restored.
 */
/*ARGSUSED*/
inter_should_rest(fi, path, name)
    struct file_info *fi;
    char *path, *name;
{
    struct file_ent *fe;

    fe = (fi == cur_dir_fi) ? cur_dir : lookup_fi(fi);
    return fe->flags & FE_REST;
} /* inter_should_rest */


/* LOOKUP_FI: Look up this fi in the current dir.  Start looking where the
 * last fi was found.
 */
struct file_ent *lookup_fi(fi)
    struct file_info *fi;
{
    while (FETOINFO(*lookup_cur) != fi) {
	if (++lookup_cur >= lookup_end) {
	    lookup_cur = lookup_begin;
	}
    }
    return *lookup_cur;
} /* lookup_fi */


/* CHANGE_CUR_DIR: Change the directory in which lookups take place.
 */
change_cur_dir(fe)
    struct file_ent *fe;
{
    struct gen_list *list = &fe->info_u.dir->file_list;

    cur_dir = fe;
    lookup_begin = &LIST_GET(*list, struct file_ent*, 0);
    lookup_cur = lookup_begin;
    lookup_end = &LIST_GET(*list, struct file_ent*, LIST_SZ(*list));
} /* change_cur_dir */
