/*******************************************************************/
/*      Copyright (C) 1989 by                                      */
/*            The Regents of the University of California          */
/*                                                                 */
/*                             DISCLAIMER                          */
/*                                                                 */
/*      "Although   each   program   has   been  tested  by  its   */
/*      contributor, no warranty, express or implied, is made by   */
/*      the  contributor  or the University of California, as to   */
/*      the accuracy and functioning of the program and  related   */
/*      program material, nor shall the fact of the distribution   */
/*      constitute any such warranty, and no  responsibility  is   */
/*      assumed   by   the  contributor  or  the  University  of   */
/*      California in connection herewith."                        */
/*******************************************************************/

/* LIST.C: Do table of contents and extract savelist.
 *
 * $Header: /cd/ua/sam/src/ss/rest/RCS/list.c,v 1.3 89/06/09 15:55:13 sam REL-6-9-89 $
 *
 * $Log:	list.c,v $
 * Revision 1.3  89/06/09  15:55:13  sam
 * Cleaned up save header listing
 * Fixed to close savelist file on sequent when done.
 * 
 * Revision 1.2  89/04/03  21:05:28  sam
 * Fixed some problems with loop boundaries.  Sometimes caused a SEGV.
 * Changed start_slist to ld_slist and it now fills in an slist_ld_info
 * struct instead of just returning a char pointer
 * Added unld_slist to unload a loaded slist.
 * 
 * Revision 1.1  89/03/08  20:38:46  sam
 * Initial revision
 * 
 */
#ifndef lint
static char rcsid[] = "$Header: /cd/ua/sam/src/ss/rest/RCS/list.c,v 1.3 89/06/09 15:55:13 sam REL-6-9-89 $";
#endif

#include "../global.h"
#include "rest.h"
#include <stdio.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <sys/file.h>
#include <pwd.h>
#include <grp.h>

#define NCACHE 15

/* CACHE_INFO: One entry in a cache */
struct cache_info
{
    char name[9];
    unsigned last_used;
    int id;
};

/* CACHE_REC: Record for one cache */
struct cache_rec
{
    unsigned time;
    int nused;
    struct cache_info info[NCACHE];
};

/* Local vars */
static struct cache_rec
    user_cache, group_cache;	/* Caches for group and user names */

/* Externs */
extern char
    *save_list,			/* Name of save list file */
    *save_list_dir;		/* Name of directory holding savelists */
extern int
    chat,			/* Chat with the user? */
    pagesize,			/* System page size */
    beg_file_num;		/* First file on tape to use */

/* Function decls */
char
    *mk_slist_name(), *start_slist();
struct cache_info 
    *look_up();
extern char
    *ctime(), *rindex(), *malloc(), *sprintf(), *strcpy();


/* TOC: List the table of contents from the savelist file or the tape file.
 */
toc()
{
    char *s, *slist_ptr, path[MAXDIRLENGTH], name[MAXFILENAME];
    int path_len = 0;
    struct save_header *hd;
    struct slist_ld_info ld_info;
    struct file_info fi_buf, *fi;

    /* Open files and get header */
    if (save_list) {
	/* Get header from savelist in memory */
	s = mk_slist_name();
	if (ld_slist(&ld_info, s) == -1) {
	    prerror(s);
	    finish_up();
	}
	free(s);
	slist_ptr = ld_info.ld_point;
	hd = (struct save_header*)slist_ptr;
	slist_ptr += sizeof(*hd);
    } else {
	/* Get header from tape */
	if (mount_tape("", 0, beg_file_num) == -1) finish_up();
	if (tp_bootstrap(&hd) == -1) finish_up();
	fi = &fi_buf;
    }

    /* Print out header stuff */
    pr_save_header(hd, (int)save_list);

    /* Init cache's for print_info */
    init_cache(&user_cache);
    init_cache(&group_cache);

    for (;;) {
	/* Get the file_info record */
	if (save_list) {
	    fi = (struct file_info*)slist_ptr;
	    slist_ptr += sizeof(*fi);
	} else {
	    toc_read((char*)fi, sizeof(*fi));
	}

	/* Get a name */
	if (NEEDS_NAME(fi->record_type)) {
	    if (save_list) {
		bcopy(slist_ptr, name, (int)fi->name_len);
		slist_ptr += fi->name_len;
	    } else {
		toc_read(name, (int)fi->name_len);
	    }
	    name[fi->name_len] = '\0';
	}

	/* Print out file information. Skip file data if needed. */
	switch (fi->record_type) {
	    case PATH_DIR:
	    case CD_DOWN:
		print_info(path, name, fi);
		if (path_len) path[path_len++] = '/';
		bcopy(name, &path[path_len], (int)fi->name_len);
		path_len += fi->name_len;
		path[path_len] = '\0';
		break;
	    case L_FILE:
		print_info(path, name, fi);
	    case I_FILE:
		break;
	    case SYMLINK:
	    case A_FILE:
	    case M_FILE:
		print_info(path, name, fi);
		if (!(hd->type & SAVE_SLIST)) {
		    if (save_list) {
			slist_ptr += convl(&fi->st_size);
		    } else {
			(void) tp_seek(convl(&fi->st_size), 1);
		    }
		}
		break;
	    case CD_ROOT:
		path_len = 0;
		path[path_len] = '\0';
		break;
	    case CD_UP:
		s = rindex(path, '/');
		path_len = s ? s - path : 0;
		path[path_len] = '\0';
		break;
	    case TYPE_END:
		finish_up();
		break;
	    default:
		pr("Illegal restore file type %d\n", fi->record_type);
		finish_up();
		break;
	}
    }
} /* toc */


/* PR_SAVE_HEADER: Print info from the save header in a readable fashion.
 * If pr_end_time is true, print the ending time of the save, otherwise, don't.
 */
pr_save_header(hd, pr_end_time)
    struct save_header *hd;
    int pr_end_time;
{
    int i;
    char *s;

    pr("SS version %c.%c, tape name is \"%s\", savelist name is \"%s\"\n",
	hd->version[0], hd->version[1], hd->tapename, hd->slistname);
    i = convl(&hd->begin_time);
    (s = ctime((long*)&i))[24] = '\0';
    pr("%s, started %s, ",
       (hd->type & SAVE_INC) ? "Incremental" : "Fullsave", s);
    if (pr_end_time) {
	i = convl(&hd->end_time);
	(s = ctime((long*)&i))[24] = '\0';
	pr("ended %s\n", s);
    }
    pr("save %s, tape number %d, file number %d, %d bytes left on tape\n",
       (hd->type & SAVE_SLIST) ? "list" : "data",
       hd->tapenum, convl(&hd->file_offset), convl(&hd->bytes_left));
} /* pr_save_header */


/* TOC_READ: Read from a block from tape, exit if error.
 */
static toc_read(buf, size)
    char *buf;
    int size;
{
    if (tp_read(buf, size) != size) {
	pr("Tape error\n");
	finish_up();
    }
} /* toc_read */


/* PRINT_INFO: Print out one line of info for the toc listing.
 */
print_info(path, name, fi)
    char *path, *name;
    struct file_info *fi;
{
    struct passwd *pwent, *getpwuid();
    struct group *grent, *getgrgid();
    struct cache_info *ci;
    char *uname, *gname, *s, *p;
    int i;

    /* Get uname */
    i = convs(&fi->st_uid);
    ci = look_up(&user_cache, i);
    if (ci) {
	uname = ci->name;
    } else {
	pwent = getpwuid(i);
	if (pwent) {
	    uname = pwent->pw_name;
	    add_cache(&user_cache, uname, i);
	} else {
	    uname = NULL;
	}
    }

    /* Print uname */
    if (uname) {
	pr("%-9s", uname);
    } else {
	pr("%-9d", i);
    }

    /* Get gname */
    i = convs(&fi->st_gid);
    ci = look_up(&group_cache, i);
    if (ci) {
	gname = ci->name;
    } else {
	grent = getgrgid(i);
	if (grent) {
	    gname = grent->gr_name;
	    add_cache(&group_cache, gname, i);
	} else {
	    gname = NULL;
	}
    }

    /* Print gname */
    if (gname) {
	pr("%-9s", gname);
    } else {
	pr("%-9d", i);
    }

    /* Print mtime, path and file name */
    i = convl(&fi->st_mtime);
    (s = ctime((long*)&i))[24] = '\0';
    p = &s[19];
    s[16] = '\0';
    if (*path) {
	pr("%s%s %s/%s\n", s, p, path, name);
    } else {
	pr("%s%s %s\n", s, p, name);
    }
} /* print_info */


/* LOOK_UP: Look up something in a cache.
 */
struct cache_info *look_up(cache, id)
    struct cache_rec *cache;
    int id;
{
    register struct cache_info *ci, *end_ci;

    end_ci = &cache->info[cache->nused];
    for (ci = cache->info; ci < end_ci; ++ci) {
	if (ci->id == id) {
	    ci->last_used = cache->time++;
	    return ci;
	}
    }
    return NULL;
} /* look_up */


/* ADD_TO_CACHE: Add something to the cache.
 */
add_cache(cache, name, id)
    struct cache_rec *cache;
    char *name;
    int id;
{
    int min_val = cache->time + 1;
    register struct cache_info *ci, *end_ci, *min_ci = NULL;

    if (cache->nused < NCACHE) {
	ci = &cache->info[cache->nused++];
    } else {
	end_ci = &cache->info[cache->nused];
	for (ci = cache->info; ci < end_ci; ++ci) {
	    if (ci->last_used <= min_val) {
		min_ci = ci;
		min_val = ci->last_used;
	    }
	}
	ci = min_ci;
    }
    ci->id = id;
    (void) strcpy(ci->name, name);
    ci->last_used = cache->time++;
} /* add_to_cache */


/* INIT_CACHE: Initialize the cache rec.
 */
init_cache(cache)
    struct cache_rec *cache;
{
    cache->nused = cache->time = 0;
} /* init_cache */


/* EXTR_SLIST: Extract a savelist from a save file.
 */
extr_slist()
{
    char buf[10], *s, name[MAXFILENAME];
    FILE *fp;
    struct save_header *hd;
    struct file_info fi;

    /* Open tape */
    if (mount_tape("", 0, beg_file_num) == -1) finish_up();
    if (tp_bootstrap(&hd) == -1) finish_up();

    /* Exit if it's already a savelist */
    if (hd->type & SAVE_SLIST) {
	pr("Already a savelist, try extract anyways (y/n)? ");
	input_line(buf, sizeof(buf));
	if (*buf != 'Y' && *buf != 'y') finish_up();
    }

    /* Make up save_list name if user hasn't */
    if (!save_list) {
	save_list = hd->slistname;
	if (chat) {
	    pr("Getting savelist %s\n", save_list);
	}
    }
    s = mk_slist_name();

    /* Open up */
    if (access(s, F_OK) != -1) {
	pr("Overwrite %s (y/n)? ", s);
	input_line(buf, sizeof(buf));
	if (*buf != 'Y' && *buf != 'y') finish_up();
    }
    fp = fopen(s, "w");
    if (fp == NULL) {
	prerror(s);
	finish_up();
    }

    /* Set the SLIST bit in the header and write it out */
    hd->type |= SAVE_SLIST;
    extr_write(fp, (char*)hd, sizeof(*hd));

    for (;;) {
	/* Get the file_info record */
	toc_read((char*)&fi, sizeof(fi));
	if (fi.record_type != I_FILE) {
	    extr_write(fp, (char*)&fi, sizeof(fi));
	}

	/* Get a name */
	if (NEEDS_NAME(fi.record_type)) {
	    toc_read(name, (int)fi.name_len);
	    if (fi.record_type != I_FILE) {
	        extr_write(fp, name, (int)fi.name_len);
	    }
	}

	/* Skip data on tape if needed */
	switch (fi.record_type) {
	    case CD_UP:
	    case CD_ROOT:
	    case CD_DOWN:
	    case L_FILE:
	    case I_FILE:
		break;
	    case SYMLINK:
	    case A_FILE:
	    case M_FILE:
		(void) tp_seek(convl(&fi.st_size), 1);
		break;
	    case TYPE_END:
		finish_up();
		break;
	    default:
		pr("Illegal file type %d\n", fi.record_type);
		finish_up();
		break;
	}
    }
} /* extr_slist */


/* EXTR_WRITE: Do an fwrite to the fp, exit on error.
 */
static extr_write(fp, buf, size)
    FILE *fp;
    char *buf;
    int size;
{
    if (fwrite(buf, size, 1, fp) != 1) {
	prerror("write");
	finish_up();
    }
} /* extr_write */


/* MK_SLIST_NAME: Return a malloc'd char array with the full filename of the
 * savelist file in it.
 */
char *mk_slist_name()
{
    char *s = NEW(char, strlen(save_list)+strlen(save_list_dir)+2);
    if (s == NULL) {
	pr("Out of memory\n");
	exit(1);
    }
    (void) sprintf(s, "%s/%s", save_list_dir, save_list);
    return s;
} /* mk_slist_name */


/* LD_SLIST: Read the savelist into memory.  Return it's address.  Return
 * NULL if slist can't be opened or read.
 */
int ld_slist(info, name)
    struct slist_ld_info *info;
    char *name;
{
    struct stat stbuf;
    char *ptr;
    int fd;
#ifdef sequent
    int maplen;
#endif

#ifdef sequent
    if ((fd = open(name, O_RDWR)) == -1) {
	if ((fd = open(name, 0)) == -1) return -1;
    }
    info->fd = fd;
#else
    if ((fd = open(name, 0)) == -1) return -1;
#endif

    (void) fstat(fd, &stbuf);

#ifdef sequent
    ptr = (caddr_t) (((int)sbrk(0)+(pagesize-1)) & ~(pagesize-1));
    if ((maplen = stbuf.st_size) % pagesize) {
	maplen = (maplen + pagesize) & ~(pagesize-1);
    }
    info->ismapped = (mmap(ptr, maplen, PROT_READ, MAP_SHARED, fd, 0) != -1);
    if (info->ismapped) {
	info->length = maplen;
    } else {
	/* mmap failed so do a malloc */
	ptr = malloc(stbuf.st_size);
	if (ptr == NULL) {
	    pr("Out of memory\n");
	    return -1;
	}
	if (read(fd, ptr, stbuf.st_size) != stbuf.st_size) {
	    prerror("read");
	    ptr = NULL;
	}
	info->length = stbuf.st_size;
    }
#else /* sun */
#ifdef SUN3X
    ptr = malloc(stbuf.st_size);
    if (ptr == NULL) {
	pr("Out of memory\n");
	return -1;
    }
    if (read(fd, ptr, stbuf.st_size) != stbuf.st_size) {
	prerror("read");
	ptr = NULL;
    }
    info->ismapped = 0;
    info->length = stbuf.st_size;
#else /* SUN40 */
    if ((ptr = mmap((caddr_t)0, (int)stbuf.st_size,
	PROT_READ, MAP_SHARED, fd, (off_t)0)) == (char*)-1)
    {
	prerror("mmap");
	ptr = NULL;
    }
    info->ismapped = 1;
    info->length = stbuf.st_size;
#endif
#endif /* sun */

    /* We can't do the close now on the sequent, otherwise we'll get a SEGV.
     */
#ifndef sequent
    (void) close(fd);
#endif

    return (info->ld_point = ptr) == NULL ? -1 : 0;
} /* ld_slist */


/* UNLD_SLIST: Unload the savelist from memory.
 */
unld_slist(info)
    struct slist_ld_info *info;
{
    if (info->ismapped) {
	if (munmap(info->ld_point, info->length) == -1) {
	    perror("munmap");
	}
    } else {
	free(info->ld_point);
    }
#ifdef sequent
    (void) close(info->fd);
#endif
} /* unld_slist */
