/*******************************************************************/
/*      Copyright (C) 1989 by                                      */
/*            The Regents of the University of California          */
/*                                                                 */
/*                             DISCLAIMER                          */
/*                                                                 */
/*      "Although   each   program   has   been  tested  by  its   */
/*      contributor, no warranty, express or implied, is made by   */
/*      the  contributor  or the University of California, as to   */
/*      the accuracy and functioning of the program and  related   */
/*      program material, nor shall the fact of the distribution   */
/*      constitute any such warranty, and no  responsibility  is   */
/*      assumed   by   the  contributor  or  the  University  of   */
/*      California in connection herewith."                        */
/*******************************************************************/

/* REST.C: Start rest, do full restore and extract.
 *
 * $Header: /cd/ua/sam/src/ss/rest/RCS/rest.c,v 1.3 89/06/09 15:57:02 sam REL-6-9-89 $
 *
 * $Log:	rest.c,v $
 * Revision 1.3  89/06/09  15:57:02  sam
 * Cleaned up the way tape is started when we have a savelist
 * 
 * Revision 1.2  89/04/03  21:07:49  sam
 * Now using new memory routines
 * Fixed some loop boundary problems.
 * Fixed to use ld_slist instead of start_slist
 * 
 * Revision 1.1  89/03/08  20:38:58  sam
 * Initial revision
 * 
 */
#ifndef lint
static char rcsid[] = "$Header: /cd/ua/sam/src/ss/rest/RCS/rest.c,v 1.3 89/06/09 15:57:02 sam REL-6-9-89 $";
#endif

#include "../global.h"
#include "rest.h"
#include <sys/types.h>
#include <sys/param.h>
#include <stdio.h>
#include <time.h>
#include <fcntl.h>
#include <netdb.h>

/* Local vars */
int
    rest_dir;			/* Are we in a restored dir? */
struct gen_list
    pre_list;			/* Prefixes of paths to be restored */


/* Global vars */
int
    pagesize,			/* System page size */
    uid,			/* Effective uid */
    dtablesize,			/* FD table size */
    pid,			/* Our pid */
    chat,			/* Talk to the user? */
    hash,			/* Print hash marks? */
    shell_port,			/* Port for doing rsh */
    wd_len,			/* strlen(wd) */
    blk_size,			/* Tape block size */
    remote,			/* From a remote machine? */
    nrest_paths,		/* Number of rest_paths */
    bytes_per_tape,		/* Number of bytes on one tape */
    beg_file_num,		/* File num on tape to begin at */
    rest_mode;			/* Type of restore to do */
char
    *tp_buf,			/* Tape block buffer */
    *save_list,			/* Name of the save list file */
    *save_list_dir,		/* The savelist directory */
    *tapedev,			/* Name of tape device */
    **rest_paths,		/* Paths to restore */
    wd[MAXPATHLEN];		/* Current working dir */

/* Function decls */
extern char
    *mk_slist_name(), *bsearch(), *malloc(), *strcpy();
int
    extr_cmp() , extr_should_rest();
void
    extr_cd_down(), extr_cd_up();


/* Better arg checking for lint */
#ifdef lint
convs(p) u_short *p; { return (u_short)*p; }
convl(p) u_long *p; { return (u_long)*p; }
#endif


/* MAIN
 */
main(argc, argv)
    char **argv;
{
    initialize(argc, argv);
    switch(rest_mode) {
	case MODE_INTER:
	    interactive();
	    break;
	case MODE_FULL:
	    rest_full();
	    break;
	case MODE_EXTRACT:
	    extr();
	    break;
	case MODE_TOC:
	    toc();
	    break;
	case MODE_SAVELIST:
	    extr_slist();
	    break;
    }
    finish_up();
} /* main */


/* EXTR: Extract the paths in the rest_paths array.
 */
extr()
{
    static void (*(funcs[NTYPES]))() = {
	NULL, NULL, extr_cd_up, extr_cd_down, NULL, NULL, NULL, NULL, NULL,
	extr_cd_down};
    int i;
    char *s, **p, **end_p;

    /* Take slashes off the front of the paths */
    for (i = 0; i < nrest_paths; ++i) {
	if (rest_paths[i][0] == '/') {
	    ++rest_paths[i];
	}
    }

    /* Add all the dirs that are in the path to restored files to a list */
    init_list(&pre_list);
    end_p = &rest_paths[nrest_paths];
    for (p = rest_paths; p < end_p; ++p) {
	for (i = 0; (*p)[i]; ++i) {
	    if ((*p)[i] == '/') {
		if (!(s = NEW(char, i+1))) {
		    pr("Out of memory\n");
		    finish_up();
		}
		(*p)[i] = '\0';
		(void) strcpy(s, *p);
		(*p)[i] = '/';
		if (LIST_ADD(pre_list, char*, s, 20) == -1) finish_up();
	    }
	}
    }

    /* Sort the prefix dirs */
    qsort((char*)&LIST_GET(pre_list, char*, 0), LIST_SZ(pre_list),
	  sizeof(char*), extr_cmp);

    /* Sort the paths */
    qsort((char*)rest_paths, nrest_paths, sizeof(char*), extr_cmp);

    /* Do the restore */
    rest_dir = 0;
    do_rest(extr_should_rest, funcs);
} /* extr */


/* EXTR_CMP: Compare for the sorts and searches in extr routines.
 */
extr_cmp(p1, p2)
    char **p1, **p2;
{
    return strcmp(*p1, *p2);
} /* extr_cmp */


/* EXTR_CD_UP: Cd up.
 */
/*ARGSUSED*/
void extr_cd_up(fi, path, name)
    struct file_info *fi;
    char *path, *name;
{
    if (rest_dir) --rest_dir;
} /* extr_cd_up */


/* EXTR_CD_DOWN: Cd into a dir.
 */
/*ARGSUSED*/
void extr_cd_down(fi, path, name)
    struct file_info *fi;
    char *path, *name;
{
    if (rest_dir || extr_look_up(rest_paths, nrest_paths, path, name)) {
	++rest_dir;
    }
} /* extr_cd_down */


/* EXTR_SHOULD_REST: should_rest function for extract.
 */
extr_should_rest(fi, path, name)
    struct file_info *fi;
    char *path, *name;
{
    if (rest_dir) {
	return 1;
    } else if (fi->record_type == CD_DOWN || fi->record_type == PATH_DIR) {
        /* Look in the prefixes to see if we'll need this dir */
	return extr_look_up(&LIST_GET(pre_list, char*, 0), LIST_SZ(pre_list),
			    path, name);
    } else {
	return extr_look_up(rest_paths, nrest_paths, path, name);
    }
} /* extr_should_rest */


/* EXTR_LOOK_UP: Do a bsearch on a path/name.
 */
extr_look_up(search_paths, npaths, path, name)
    char **search_paths;
    int npaths;
    char *path, *name;
{
    register int oldl = strlen(path), l, retval;

    if (l = oldl) path[l++] = '/';
    (void) strcpy(&path[l], name);
    retval = bsearch((char*)&path, (char*)search_paths, (unsigned)npaths,
		     sizeof(char*), extr_cmp) != NULL;
    path[oldl] = '\0';
    return retval;
} /* extr_look_up */


/* REST_FULL: Do a full restore.
 */
rest_full()
{
    int full_should_rest();
    static void (*(null_funcs[NTYPES]))();

    do_rest(full_should_rest, null_funcs);
} /* rest_full */


/* FULL_SHOULD_REST: On full restore, should restore everything.
 */
/*ARGSUSED*/
full_should_rest(fi, path_name, name)
    struct file_info *fi;
    char *path_name, *name;
{
    return 1;
} /* full_should_rest */


/* DO_REST: Start a restore and use the passed function for a should_restore
 * function.
 */
do_rest(should_rest, funcs)
    int (*should_rest)();
    void (*(funcs[]))();
{
    char *s, *slist_ptr;
    struct save_header *tp_hd, *sl_hd, tp_hd_store;
    struct slist_ld_info ld_info;

    /* If save list given, open it and open tape */
    if (save_list) {
	s = mk_slist_name();
	if (ld_slist(&ld_info, s) == -1) {
	    prerror(s);
	    finish_up();
	}
	free(s);
	slist_ptr = ld_info.ld_point;
	sl_hd = (struct save_header*)slist_ptr;
	tp_hd = &tp_hd_store;
	switch(start_tape_from_slist(sl_hd, tp_hd)) {
	    case -2:
		tp_hd = NULL; break;
	    case -1:
		finish_up(); break;
	    default:
		break;
	}
    } else {
	if (mount_tape("", 0, beg_file_num) == -1) finish_up();
	if (tp_bootstrap(&tp_hd) == -1) finish_up();
	save_list = tp_hd->slistname;
	s = mk_slist_name();
	if (ld_slist(&ld_info, s) == -1) {
	    pr("Warning, can't read savelist %s, continuing\n", s);
	    slist_ptr = NULL;
	    sl_hd = NULL;
	} else {
	    slist_ptr = ld_info.ld_point;
	    sl_hd = (struct save_header*)slist_ptr;
	}
	free(s);
    }

    /* Compare header from tape and header on savelist */
    if (tp_hd && sl_hd) {
	if (cmp_save_headers(sl_hd, tp_hd) == -1) finish_up();
    }

    /* Start the restore */
    if (slist_ptr) slist_ptr += SAVE_HEADER_SIZE;
    (void) restore(slist_ptr, funcs, funcs, should_rest);
} /* do_rest */


/* CMP_SAVE_HEADERS: Compare a tape save_header and a savelist save_header.
 * If they differ, ask the user if they want to continue.  If the answer is no,
 * return -1, else return 0.
 */
cmp_save_headers(sl_hd, tp_hd)
    struct save_header *tp_hd, *sl_hd;
{
    char buf[10];

    if (bcmp(tp_hd->version, sl_hd->version, sizeof(sl_hd->version))
     || tp_hd->tapenum != sl_hd->tapenum
     || convl(&tp_hd->begin_time) != convl(&sl_hd->begin_time)
     || convl(&tp_hd->file_offset) != convl(&sl_hd->file_offset)
     || convl(&tp_hd->bytes_left) != convl(&sl_hd->bytes_left))
    {
	pr("Header on tape and header on savelist don't match.\n");
	pr("\nHeader in savelist file is:\n\n");
	pr_save_header(sl_hd, 1);
	pr("\nHeader on tape is:\n\n");
	pr_save_header(tp_hd, 0);
	pr("\nContinue (y/n)? ");
	input_line(buf, sizeof(buf));
	if (*buf != 'y' && *buf != 'Y') return -1;
    }
    return 0;
} /* cmp_save_headers */


/* START_TAPE_FROM_SLIST: Given a savelist header, get the right tape mounted
 * and read the save header from it into tp_hd and check against the savelist's
 * header.  If we can't mount the tape, return -1.  If we can't read the
 * header from the tape return -2.  Else return 0.
 */
start_tape_from_slist(sl_hd, tp_hd)
    struct save_header *sl_hd, *tp_hd;
{
    if (mount_tape(sl_hd->tapename, (int)sl_hd->tapenum,
	convl(&sl_hd->file_offset)) == -1 || tp_init(sl_hd) == -1)
    {
	return -1;
    }
    if (tp_read((char*)tp_hd, SAVE_HEADER_SIZE) != SAVE_HEADER_SIZE) {
	pr("Warning: can't read header from tape\n");
	return -2;
    }
    return 0;
} /* start_tape_from_slist */


/* FINISH_UP: Clean up and exit.
 */
finish_up()
{
    if (remote) {
	stop_rmt();
    }
    exit(0);
} /* finish_up */
