/* REST.H: Header for rest.
 *
 * $Header: /cd/ua/sam/src/ss/rest/RCS/rest.h,v 1.3 89/06/09 15:57:54 sam REL-6-9-89 $
 *
 * $Log:	rest.h,v $
 * Revision 1.3  89/06/09  15:57:54  sam
 * No changes
 * 
 * Revision 1.2  89/04/03  21:09:06  sam
 * Everything has been rearranged.  Added new memory macros.  Added
 * savelist, and fs structures.
 * 
 * Revision 1.1  89/03/08  20:39:09  sam
 * Initial revision
 * 
 */


#ifdef lint
u_short	ntohs(), htons();
u_long	ntohl(), htonl();
#endif



/* Constants */


#define TAPEDEV		"/dev/rst0"	/* Tape device */
#define LS_CUTOFF	(182*24*60*60)	/* Cutoff in ls before printing years */
#define MAXLSEEK	(800*64512)	/* Max tape lseek allowed */

/* Restore modes */
#define MODE_INTER	0		/* Interactive */
#define MODE_FULL	1		/* Full restore */
#define MODE_TOC	2		/* List table of contents */
#define MODE_EXTRACT	3		/* Extract some files */
#define MODE_SAVELIST	4		/* Extrace a savelist */

/* file_ent flags (can be or'ed) */
#define FE_ISDIR	1	/* It's a directory */
#define FE_REST		2	/* Being restored */

/* Params for pool alloc's */
#define POOL_INC	50	/* How many buf ptrs to allocate at once */
#define POOL_SZ		4096	/* Size of one pool buffer */

/* Defs for mode bits in stat's st_mode */
#ifndef S_IXUSR
#define S_IXUSR 0000100         /* execute/search permission, owner */
#endif
#ifndef S_IXGRP
#define S_IXGRP 0000010         /* execute/search permission, group */
#endif
#ifndef S_IXOTH
#define S_IXOTH 0000001         /* execute/search permission, other */
#endif



/* Structures */


/* GEN_LIST: A generic list (or dynamically allocated array).
 */
struct gen_list {
    int count;		/* Number of things in the list */
    int num_alloced;	/* Number allocated */
    char *list;		/* The list */
}; /* gen_list */

/* POOL: List of big buffers to allocate smaller things.  Space allocated
 * from a pool can't be resized or freed individually.
 */
struct pool {
    struct gen_list bufs;	/* List of buffers that hold things */
    char *cur;			/* Pointer to open spot in current pool */
    char *end;			/* One plus end of current pool */
}; /* pool */

/* DIR_ENT: A directory (for internal fs)
 */
struct dir_ent {
    struct file_ent *parent;		/* Our parent dir */
    struct file_info *info;		/* Our file_info record */
    struct gen_list file_list;		/* List of files in this dir */
}; /* dir_ent */

/* FILE_ENT: One file entry (for internal fs).  Can be directory or a normal
 * file.
 */
struct file_ent {
    union _info_u {
	struct file_info *file;		/* Info record if normal file*/
	struct dir_ent *dir;		/* Directory info for dirs */
    } info_u;
    char flags;				/* Flags as given above */
}; /* file_ent */

/* SLIST_LD_INFO: Info for a loaded savelist.
 */
struct slist_ld_info {
    char *ld_point;		/* Where it's loaded, NULL = not loaded */
    int length;			/* Length of loaded part */
    short ismapped;		/* Is this mmap'd */
#ifdef sequent
    int fd;			/* FD to close after unmapping on sequent */
#endif
}; /* slist_ld_info */

/* SLIST_INFO: Information about a savelist.
 */
struct slist_info {
    char *file_name;
    struct save_header *hd;		/* Header */
    struct slist_ld_info ld_info;	/* Info about where this is loaded */
    struct file_ent *root;		/* Root of f/s built around this */
    int num_marked;			/* Number of files marked for restore */
    struct pool pool;			/* Pool for alloc'ing fs struct's */
}; /* slist_info */

/* MF_INFO: Saved info for an mfile.
 */
struct mf_info {
    int offset;			/* Offset on tape */
    short restore;		/* Will it be restored (so far)? */
    short istmp;		/* Is this a temporary filename */
    char *name;			/* Full pathname of file to be restored */
}; /* mf_info */

/* CMD_REC: Pairs a command's name with a function to call.  Used by shell.
 */
struct cmd_rec {
    char *name;
    int (*func)();
}; /* cmd_rec */



/* Functional macros */


/* NEEDS_NAME: Does the passed type of file need a name?
 */
#define NEEDS_NAME(t) ("\0y\0y\0yyyyy"[t])

/* FETOINFO: Convert a file_ent pointer to the corresponing file_info pointer.
 */
#define FETOINFO(fep) \
    (((fep)->flags & FE_ISDIR) ? (fep)->info_u.dir->info : (fep)->info_u.file)

/* INFONAME: Giving a pointer to a struct file_info, return a pointer to
 * the name after it.
 */
#define INFONAME(p) (((char*)p) + FILE_INFO_SIZE)

/* ISDOTFILE: Return true if the string given is . or ..
 */
#define ISDOTFILE(s) \
	 (*s == '.' && (s[1] == '\0' || (s[1] == '.' && s[2] == '\0')))

#ifndef lint

/* CONVS: Convert from a short in the savelist to the host's short.  Takes
 * care of alignment problems.
 */
#ifdef sparc
#define cs2(p) ((p[0] << 8) | p[1])
#define convs(p) ((u_short)ntohs((u_short)cs2(((unsigned char *)p))))
#else
#define convs(p) ntohs(*(p))
#endif

/* CONVL: Convert from a long in the savelist to the hosts's long.  Takes
 * care of alignment problems.
 */
#ifdef sparc
#define cl2(p) ((p[0] << 24) | (p[1] << 16) | (p[2] << 8) | p[3])
#define convl(p) ((u_long)ntohl((u_long)cl2(((unsigned char *)p))))
#else
#define convl(p) ntohl(*(p))
#endif

#endif /* not lint */

/* List macros */

/* LIST_ADD: Add one element to the end of the list.
 */
#define LIST_ADD(lst, typ, what, alloc_inc) \
    (((lst).count == (lst).num_alloced) \
	? _list_add(&(lst), sizeof(typ), (char*)&(what), (alloc_inc)) \
	: ((((typ*)(lst).list)[(lst).count++] = (what)), 0))

/* LIST_NEW: Return a pointer to the next free spot in the list.
 */
#define LIST_NEW(lst, typ, alloc_inc) \
    (((lst).count == (lst).num_alloced) \
	? (typ*)_list_new(&(lst), sizeof(typ), (alloc_inc)) \
	: &((typ*)(lst).list)[(lst).count++])
extern char *_list_new();

/* LIST_GET: Retrieve an element from the list.  Evaluates to that element.
 */
#define LIST_GET(lst, typ, idx) (((typ*)(lst).list)[idx])

/* LIST_SZ: Return the number of items in the list.
 */
#define LIST_SZ(lst) ((lst).count)

/* LIST_DEL: Delete one element from the end of the list.
 */
#define LIST_DEL(lst) ((lst).count = (lst).count ? (lst).count - 1 : 0)


/* POOL_NEW: Allocate something from a pool.
 */
#define POOL_NEW(pl, typ) \
    ((((pl).end - (pl).cur) >= sizeof(typ)) \
	? ((typ*)(pl).cur)++ : (typ*)_pool_new(&(pl), sizeof(typ)))
extern char *_pool_new();
