/*******************************************************************/
/*      Copyright (C) 1989 by                                      */
/*            The Regents of the University of California          */
/*                                                                 */
/*                             DISCLAIMER                          */
/*                                                                 */
/*      "Although   each   program   has   been  tested  by  its   */
/*      contributor, no warranty, express or implied, is made by   */
/*      the  contributor  or the University of California, as to   */
/*      the accuracy and functioning of the program and  related   */
/*      program material, nor shall the fact of the distribution   */
/*      constitute any such warranty, and no  responsibility  is   */
/*      assumed   by   the  contributor  or  the  University  of   */
/*      California in connection herewith."                        */
/*******************************************************************/

/* ROUTINES.C: Misc subroutines.
 *
 * $Header: /cd/ua/sam/src/ss/rest/RCS/routines.c,v 1.3 89/06/09 15:58:17 sam REL-6-9-89 $
 *
 * $Log:	routines.c,v $
 * Revision 1.3  89/06/09  15:58:17  sam
 * No changes
 * 
 * Revision 1.2  89/04/03  21:10:26  sam
 * Changed the way the list routines are called -- through macros now.
 * Added routines to support pool alloc macros.
 * 
 * Revision 1.1  89/03/08  20:39:16  sam
 * Initial revision
 * 
 */
#ifndef lint
static char rcsid[] = "$Header: /cd/ua/sam/src/ss/rest/RCS/routines.c,v 1.3 89/06/09 15:58:17 sam REL-6-9-89 $";
#endif

#include "../global.h"
#include "rest.h"
#include <stdio.h>
#include <varargs.h>
#include <errno.h>

/* Externs */
extern int
    sys_nerr;
extern char
    *sys_errlist[];

/* Function decls */
extern char
    *malloc(), *realloc();
    

/* _LIST_ADD: Do a malloc on a list and add an element to it.
 */
int _list_add(list, elem_size, what, alloc_inc)
    register struct gen_list *list;
    register int elem_size, alloc_inc;
    register char *what;
{
    char *empty = _list_new(list, elem_size, alloc_inc);

    if (empty == NULL) return -1;
    bcopy(what, empty, elem_size);
    return 0;
} /* _list_add */


/* _LIST_NEW: Do a malloc and return a pointer to an empty spot in the list.
 */
char *_list_new(list, elem_size, alloc_inc)
    register struct gen_list *list;
    register int elem_size, alloc_inc;
{
    list->num_alloced += alloc_inc;
    if (list->list) {
	list->list = realloc(list->list,
			     (unsigned)(list->num_alloced * elem_size));
    } else {
	list->list = malloc((unsigned)(list->num_alloced * elem_size));
    }
    if (list->list == NULL) {
	pr("Out of memory\n");
	return NULL;
    }
    return list->list + (elem_size * list->count++);
} /* _list_alloc */


/* INIT_LIST: Initialize a list.
 */
init_list(list)
    struct gen_list *list;
{
    list->count = list->num_alloced = 0;
    list->list = NULL;
} /* init_list */


/* DELETE_LIST: Delete everything in a list.
 */
delete_list(list)
    struct gen_list *list;
{
    if (list->num_alloced) free(list->list);
} /* delete_list */


/* INIT_POOL: Initialize a pool.
 */
init_pool(pool)
    struct pool *pool;
{
    init_list(&pool->bufs);
    pool->cur = pool->end = NULL;
} /* init_pool */


/* DELETE_POOL: Free the contents of a pool.
 */
delete_pool(pool)
    struct pool *pool;
{
    int i;

    for (i = LIST_SZ(pool->bufs)-1; i >= 0; --i) {
	free(LIST_GET(pool->bufs, char*, i));
    }
    delete_list(&pool->bufs);
} /* delete_pool */

#define POOL_NEW(pl, typ) \
    ((((pl).end - (pl).cur) >= sizeof(typ)) \
	? ((typ*)(pl).cur)++ : (typ*)_pool_new(&(pl), sizeof(typ)))
    
/* _POOL_NEW:  Allocate another buf for the pool.
 */
char *_pool_new(pool, sz)
    register struct pool *pool;
    int sz;
{
    char *s = NEW(char, POOL_SZ);

    if (s == NULL
	|| LIST_ADD(pool->bufs, char*, s, POOL_INC) == -1)
    {
	pr("Out of memory\n");
	return NULL;
    }
    pool->cur = s + sz;
    pool->end = s + POOL_SZ;
    return s;
} /* _pool_new */


/* READ_ALL: Like read(2), but read until everything is in.
 */
read_all(fd, buf, nbytes)
    int fd, nbytes;
    char *buf;
{
    int retval = nbytes, res;

    while (nbytes) {
	res = read(fd, buf, nbytes);
	if (res == -1 || res == 0) {
	    if (res == -1) {
		prerror("read");
	    } else {
		fputs("eof on read", stderr);
	    }
	    return -1;
	}
	nbytes -= res;
	buf += res;
    }
    return retval;
} /* read_all */


#ifdef sequent
/* BSEARCH: Do a binary search in an array.
 */
char *bsearch(key, base, nel, key_size, compar)
    register char *key, *base;
    register unsigned nel;
    register int key_size, (*compar)();
{
    register int low = 0, hi = nel-1, mid, cmp;

    while (low <= hi) {
	mid = (low + hi) / 2;
	cmp = compar(key, base + mid * key_size);
	/* Key is lower */
	if (cmp < 0) {
	    hi = mid-1;
	    continue;
	/* Key is higher */
	} else if (cmp > 0) {
	    low = mid+1;
	    continue;
	/* Found a match */
	} else {
	    return base + mid * key_size;
	}
    }
    return NULL;
} /* bsearch */
#endif


/* PR: Tell the user something.  Uses same args as printf.
 */
/*VARARGS0*/
pr(va_alist)
    va_dcl
{
    va_list args;
    char *fmt;

    va_start(args);
    fmt = va_arg(args, char *);
#ifdef sequent
    (void) _doprnt(fmt, args, stdout);
#else
    (void) vfprintf(stdout, fmt, args);
#endif
    va_end(args);
} /* pr */


/* PR_FLUSH: Flush whats waiting to be output.
 */
pr_flush()
{
    (void) fflush(stdout);
} /* pr_flush */


/* PRERROR: Like perror, but make output go to the right place.
 */
prerror(buf)
    char *buf;
{
    pr("%s: %s\n", buf,
        errno >= sys_nerr ? "Unknown error" : sys_errlist[errno]);
} /* prerror */


/* INPUT_LINE: Get a line of input from the user.
 */
input_line(buf, size)
    char *buf;
    int size;
{
    int i, c;

    if (feof(stdin)) clearerr(stdin);
    for (i = 0; i < size-1; ++i) {
	if ((c = getchar()) == EOF || c == '\n') {
	    break;
	}
        buf[i] = c;
    }
    buf[i] = '\0';
} /* input_line */
