/*******************************************************************/
/*      Copyright (C) 1989 by                                      */
/*            The Regents of the University of California          */
/*                                                                 */
/*                             DISCLAIMER                          */
/*                                                                 */
/*      "Although   each   program   has   been  tested  by  its   */
/*      contributor, no warranty, express or implied, is made by   */
/*      the  contributor  or the University of California, as to   */
/*      the accuracy and functioning of the program and  related   */
/*      program material, nor shall the fact of the distribution   */
/*      constitute any such warranty, and no  responsibility  is   */
/*      assumed   by   the  contributor  or  the  University  of   */
/*      California in connection herewith."                        */
/*******************************************************************/

/* SHELL.C: An interactive shell.
 *
 * $Header: /cd/ua/sam/src/ss/rest/RCS/shell.c,v 1.2 89/06/09 15:58:52 sam REL-6-9-89 $
 *
 * $Log:	shell.c,v $
 * Revision 1.2  89/06/09  15:58:52  sam
 * Fixed to compile on sequent
 * 
 * Revision 1.1  89/04/03  21:13:43  sam
 * Initial revision
 * 
 */
#ifndef lint
static char rcsid[] = "$Header: /cd/ua/sam/src/ss/rest/RCS/shell.c,v 1.2 89/06/09 15:58:52 sam REL-6-9-89 $";
#endif

#include "../global.h"
#include "rest.h"
#include <stdio.h>

extern char
    *bsearch();


/* SHELL: Parse a command line, execute a command and loop.
 */
shell(prompt, ncmds, cmds, exp_func)
    char *prompt;
    struct cmd_rec *cmds;
    int (*exp_func)(), ncmds;
{
    register char *s, *p;
    char buf[MAXFILENAME*2], *argv[100], **argvp, prev;
    int argc, inspc, cmd_rec_cmp(), retval, nbest, nmatch, ambig;
    FILE *out, *popen();
    struct cmd_rec *crec, *best;

    for (;;) {
	/* Prompt */
	pr_flush();
	pr("%s", prompt);
	pr_flush();

	/* Get a line */
	input_line(buf, sizeof(buf));
	if (!buf[0]) continue;

	/* Break line into argv array, and check for pipe output */
	for (s = buf, prev = argc = 0, argvp = argv, inspc = 1, out = stdout;
	     *s;
	     prev = *s++)
	{
	    if (*s == '|') {
		if (inspc) {
		    out = popen(s+1, "w");
		    if (out == NULL) {
			pr("Can't pipe to %s\n", s+1);
			out = stdout;
		    }
		    *s = '\0';
		    break;
		}
	    } else if (*s == '\t' || *s == ' ') {
		if (prev != '\\') {
		    if (!inspc) {
			inspc = 1;
			*s = '\0';
		    }
		    continue;
		}
	    }
	    if (inspc) {
		*argvp++ = s;
		++argc;
		inspc = 0;
	    }
	}

	/* Do filename expansion (done by caller) */
	if (exp_func) {
	    argc = (*exp_func)(argc, argv);
	    if (argc == -1) continue;
	}

	/* Get rid of all escapes */
	for (argvp = &argv[argc-1]; argvp >= argv; --argvp) {
	    for (s = p = *argvp, prev = '\0'; *s; s++) {
		if (*s == '\\') {
		    if (prev == '\\') {
			*p++ = '\\';
			prev = '\0';
			continue;
		    }
		} else {
		    *p++ = *s;
		}
		prev = *s;
	    }
	    *p = '\0';
	}

	/* Look up the command */
	for (crec = &cmds[ncmds-1], nbest = 0; crec >= cmds; --crec) {
	    for (s = crec->name, nmatch = 0, p = argv[0];
		 *s && *s != ' ' && *s == *p;
		 ++s, ++p)
	    {
		nmatch++;
	    }
	    if (*p) nmatch = 0;
	    if (nmatch > nbest) {
		best = crec;
		nbest = nmatch;
		ambig = 0;
	    } else if (nmatch == nbest) {
		ambig = 1;
	    }
	}
	crec = best;
	if (!nbest) {
	    pr("%s: No such command\n", argv[0]);
	    continue;
	} else if (ambig) {
	    pr("\"%s\" is ambiguous\n", argv[0]);
	    continue;
	}

	/* Call the command */
	retval = (*(crec->func))(argc, argv, out);
	if (out != stdout) (void) pclose(out);
	if (retval == -1) return;
    }
} /* shell */


#ifdef TEST

int help_cmd(), quit_cmd(), yearn_cmd(), yellow_cmd();

static struct cmd_rec test_cmds[] = {
    {"?           - Print a help message", help_cmd},
    {"help        - Print a help message", help_cmd},
    {"quit        - Exit restore", quit_cmd,},
    {"yearn       - Synonyms for yearn", yearn_cmd},
    {"yellow      - Something yellow", yellow_cmd},
};

quit_cmd()
{
    return -1;
}

yearn_cmd(argc, argv, out)
    int argc;
    char **argv;
    FILE *out;
{
    int i;
    static char yearn[4][10] = {"crave\n", "lust\n", "desire\n", "wish\n"};

    if (argc == 1) {
	srand(time(0));
	fputs(yearn[(rand()&12)>>2], out);
    } else {
	for (i = 0; i < argc; ++i) {
	    (void) fprintf(out, "arg(%d) is \"%s\"\n", argv[i]);
	}
    }
    return 0;
}

yellow_cmd(argc, argv, out)
    int argc;
    char **argv;
    FILE *out;
{
    static char yellow_things[4][10] =
	{"bannana\n", "rose\n", "bee\n", "sun\n"};
    srand(time(0));
    fputs(yellow_things[(rand()&12)>>2], out);
    return 0;
} /* yellow_cmd */


help_cmd(argc, argv, out)
    int argc;
    char **argv;
    FILE *out;
{
    int i;

    for (i = 0; i < NUM_ARR(test_cmds); ++i) {
	printf("%s\n", test_cmds[i].name);
    }
} /* help_cmd */

    
main()
{
    shell("test> ", NUM_ARR(test_cmds), test_cmds, (int (*)())NULL);
}
#endif TEST
