/*******************************************************************/
/*      Copyright (C) 1989 by                                      */
/*            The Regents of the University of California          */
/*                                                                 */
/*                             DISCLAIMER                          */
/*                                                                 */
/*      "Although   each   program   has   been  tested  by  its   */
/*      contributor, no warranty, express or implied, is made by   */
/*      the  contributor  or the University of California, as to   */
/*      the accuracy and functioning of the program and  related   */
/*      program material, nor shall the fact of the distribution   */
/*      constitute any such warranty, and no  responsibility  is   */
/*      assumed   by   the  contributor  or  the  University  of   */
/*      California in connection herewith."                        */
/*******************************************************************/

/* TPIO.C: Tape I/O (mostly I)
 *
 * $Header: /cd/ua/sam/src/ss/rest/RCS/tpio.c,v 1.3 89/06/09 15:59:37 sam REL-6-9-89 $
 *
 * $Log:	tpio.c,v $
 * Revision 1.3  89/06/09  15:59:37  sam
 * Fixed some byte ordering problems that came up on the sequent.
 * 
 * Revision 1.2  89/04/03  21:14:31  sam
 * Added many constants to replace built-in numbers.
 * Fixed so that if a savelist header has a zero bytes_left field, bytes_left
 * is assumed to be bytes_per_tape.
 * 
 * Revision 1.1  89/03/08  20:39:23  sam
 * Initial revision
 * 
 */
#ifndef lint
static char rcsid[] = "$Header: /cd/ua/sam/src/ss/rest/RCS/tpio.c,v 1.3 89/06/09 15:59:37 sam REL-6-9-89 $";
#endif

#include "../global.h"
#include "rest.h"
#include <sys/param.h>
#include <sys/file.h>
#include <sys/ioctl.h>
#include <sys/mtio.h>
#include <sys/time.h>
#include <errno.h>
#include <stdio.h>

/* MAX_RMT_CMD: Max length of one rmt command
 */
#define MAX_RMT_CMD 100

/* Locals */
static int
    in_err_blk,			/* Set if we're in the middle of a bad blk */
    left_on_tape,		/* Number of bytes left on tape. */
    left_in_buf,		/* Number of bytes left in the buffer */
    real_pos,			/* Our position in this tape file */
    log_pos,			/* Our logical position inside save file */
    tapefd = -1,		/* Fd to read from tape */
    tape_count,			/* Number of tapes into set */
    seek_pos;			/* Where the user has seeked to */
static char
    tapename[MAXTAPENAME],	/* Name of the tape */
    *tp_buf_ptr;		/* Points to next read point in tape buffer */

/* Externs */
extern int
    blk_size,			/* Size of tape block */
    bytes_per_tape,		/* Number of bytes on a tape */
    dtablesize,			/* Val from getdtablesize() */
    remote,			/* Is tape remote? */
    hash,			/* Print hash marks as blocks are read? */
    chat,			/* Chat with the user? */
    shell_port;			/* Port for doing an rsh */
extern char
    *tp_buf,			/* Points to tape buffer */
    *tapedev;			/* Name of tape device */

/* Function Decl's */
extern off_t
    lseek();
extern char
    *strcat(), *strcpy(), *strncpy(), *sprintf();


/* CHECK_ERR: Do a perror if code is -1.  Always return code.
 */
check_err(code, name)
    int code;
    char *name;
{
    if (code == -1) prerror(name);
    return code;
} /* check_err */


/* TP_IOCTL: Do an ioctl on the tape.
 */
int tp_ioctl(cmd, count)
    int cmd, count;
{
    char buf[MAX_RMT_CMD];
    struct mtop op;

    if (remote) {
	(void) sprintf(buf, "I%d\n%d\n", cmd, count);
	return write_get_ack(buf, strlen(buf));
    } else {
	op.mt_op = cmd;
	op.mt_count = count;
	return check_err(ioctl(tapefd, MTIOCTOP, &op), "ioctl");
    }
} /* tp_ioctl */
	

/* TP_CLOSE: Close the tape device.
 */
int tp_close()
{
    int retval;

    if (remote) {
	return write_get_ack("C\n", 2);
    } else {
	retval = (tapefd == -1) ? 0 : check_err(close(tapefd), "close");
	tapefd = -1;
	return retval;
    }
} /* tp_close */


/* TP_OPEN: Open the tape device.
 */
int tp_open(mode)
    int mode;
{
    char buf[MAX_RMT_CMD];

    if (remote) {
	(void) sprintf(buf, "O%s\n%d\n", tapedev, mode);
	return write_get_ack(buf, strlen(buf));
    } else {
	return tapefd = check_err(open(tapedev, mode, 0), "open");
    }
} /* tp_open */


/* TP_SEEK_LOW: Do an lseek on the tape.
 */
int tp_seek_low(offset, whence)
    int offset, whence;
{
    char buf[MAX_RMT_CMD];

    if (remote) {
	(void) sprintf(buf, "L%d\n%d\n", offset, whence);
	return write_get_ack(buf, strlen(buf));
    } else {
	return check_err((int)lseek(tapefd, (off_t)offset, whence), "lseek");
    }
} /* tp_seek_low */


/* TP_READ_LOW: Low level read of tape.  Should only read full blocks here.
 */
static int tp_read_low(blk, size)
    char *blk;
    int size;
{
    char buf[MAX_RMT_CMD];
    int retval;

    if (remote) {
	(void) sprintf(buf, "R%d\n", size);
	retval = write_get_ack(buf, strlen(buf));
	if (retval > 0 && read_rmt(blk, retval) != retval) {
	    pr("Rmt protocol screwed up\n");
	    return -1;
	}
	return retval;
    } else {
	return check_err(read(tapefd, blk, size), "read");
    }
} /* tp_read_low */


/* TP_READ: Buffered tape read.  Can use any size here you want.
 */
int tp_read(blk, sz)
    char *blk;
    int sz;
{
    int n, size = sz, had_err;

    if (tp_sync() == -1) return -1;
    had_err = in_err_blk && left_in_buf;
    while (size) {
	if (left_in_buf) {
	    n = min(left_in_buf, size);
	    bcopy(tp_buf_ptr, blk, n);
	    left_in_buf -= n;
	    left_on_tape -= n;
	    real_pos += n;
	    seek_pos += n;
	    log_pos += n;
	    tp_buf_ptr += n;
	    blk += n;
	    size -= n;
	} else {
	    if (tp_read_buf() == -1) return -1;
	    had_err |= in_err_blk;
	}
    }
    return had_err ? 0 : sz;
} /* tp_read */


/* TP_READ_BUF: Read into buffer from tape.
 */
tp_read_buf()
{
    int n, ret;

    if (hash) {
	pr(".");
	pr_flush();
    }
    in_err_blk = 0;
    if (!left_on_tape && tp_change() == -1) return -1;
    n = min(blk_size, left_on_tape);
    ret = tp_read_low(tp_buf, n);
    if (ret == 0) {
	pr("Unexpected eof on tape\n");
	return -1;
    }
    if (ret < 0) {
	left_in_buf = n;
	if (tp_seek_low(n, 1) == -1) return -1;
	bzero(tp_buf, n);
	in_err_blk = 1;
    } else {
	left_in_buf = ret;
    }
    tp_buf_ptr = tp_buf;
    return 0;
} /* tp_read_buf */


/* TP_SEEK: Do a logical seek on the tape.
 */
tp_seek(offset, whence)
    int offset, whence;
{
    switch (whence) {
	/* Absolute offset (offset is bytes into backup file) */
	case 0:
	    if (offset < log_pos) return -1;
	    seek_pos += offset - log_pos;
	    break;
	/* Relative offset */
	case 1:
	    if (offset < 0) return -1;
	    seek_pos += offset;
	    break;
	default:
	    return -1;
    }
    return seek_pos;
} /* tp_seek */


/* TP_SYNC: Sync up the tape with seek_pos.
 */
tp_sync()
{
    int diff, n;

    for (;;) {
	if (!(diff = seek_pos - real_pos)) return 0;

	/* Consume anything in the buffer */
	if (left_in_buf) {
	    n = min(left_in_buf, diff);
	    left_in_buf -= n;
	    left_on_tape -= n;
	    real_pos += n;
	    log_pos += n;
	    tp_buf_ptr += n;
	    if (!(diff = seek_pos - real_pos)) return 0;
	}
	/* Do an lseek */
	if (seek_pos/blk_size != real_pos/blk_size) {
	    n = (seek_pos - (seek_pos % blk_size)) - real_pos;
	    if (n >= left_on_tape) {
		seek_pos = n - left_on_tape;
		if (tp_change() == -1) return -1;
		continue;
	    }
	    n = min(n, MAXLSEEK);	/* Sun driver can't lseek too far */
	    if (tp_seek_low(n, 1) == -1) {
		return -1;
	    }
	    left_on_tape -= n;
	    real_pos += n;
	    log_pos += n;
	    if (!(diff = seek_pos - real_pos)) return 0;
	}
	/* Do a read */
	if (tp_read_buf() == -1) return -1;
    }
} /* tp_sync */


/* TP_CHANGE: Change to next tape.
 */
tp_change()
{
    if (tp_close() == -1) return -1;
    if (mount_new_tape() == -1) return -1;
    real_pos = 0;
    left_in_buf = 0;
    left_on_tape = bytes_per_tape;
    return 0;
} /* tp_change */


/* TP_INIT: Initialize variables from the file header.
 */
tp_init(slist_hd)
    struct save_header *slist_hd;
{
    int retval = 0;
    char buf[10];

    /* Init vars */
    left_in_buf = 0;
    real_pos = 0;
    log_pos = 0;
    seek_pos = 0;
    left_on_tape = convl(&slist_hd->bytes_left);
    if (!left_on_tape) left_on_tape = bytes_per_tape;

    if (!*tapename) {
	(void) NCOPY(tapename, slist_hd->tapename, sizeof(tapename));
    }
    if (bcmp(slist_hd->version, VERSION, sizeof(slist_hd->version))) {
	pr("Warning: version on the tape (%c.%c) is wrong\n",
	    slist_hd->version[0], slist_hd->version[1]);
	if (chat) {
	    pr("Continue (y/n)?");
	    input_line(buf, sizeof(buf));
	    retval = (*buf == 'Y' || *buf == 'y') ? 0 : -1;
	}
    }
    return retval;
} /* tp_init */


/* TP_BOOTSTRAP: Do an init by reading slist header from current file on tape.
 * Return the slist header if caller wants it.
 */
tp_bootstrap(ret_hd)
    struct save_header **ret_hd;
{
    static struct save_header slist_hd;
    char buf[10];
    int retval = 0, n;

    /* Init vars */
    left_on_tape = bytes_per_tape;
    left_in_buf = 0;
    real_pos = 0;
    log_pos = 0;
    seek_pos = 0;
    if (ret_hd) *ret_hd = &slist_hd;

    if (tp_read((char*)&slist_hd, SAVE_HEADER_SIZE) != SAVE_HEADER_SIZE) {
	return -1;
    }
    if (n = convl(&slist_hd.bytes_left)) {
	left_on_tape = n - SAVE_HEADER_SIZE;
    }
    if (!*tapename) {
	(void) NCOPY(tapename, slist_hd.tapename, sizeof(tapename));
    }
    if (bcmp(slist_hd.version, VERSION, sizeof(slist_hd.version))) {
	pr("Warning: version on the tape (%c.%c) is wrong\n",
	    slist_hd.version[0], slist_hd.version[1]);
	if (chat) {
	    pr("Continue (y/n)?");
	    input_line(buf, sizeof(buf));
	    retval = (*buf == 'Y' || *buf == 'y') ? 0 : -1;
	}
    }
    return retval;
} /* tp_bootstrap */


/* START_RMT: Start up the remote tape daemon.
 */
start_rmt(tape_host)
    char *tape_host;
{
    char hostbuf[MAXHOSTNAMELEN], *hb = hostbuf;

    /* Start rmt */
    (void) strcpy(hostbuf, tape_host);
    tapefd = rcmd(&hb, shell_port, "root", "root", RMT_CMD, (int*)0);
    if (tapefd == -1) {
	pr("Can't execute rsh on %s\n", hostbuf);
	return -1;
    }
    return 0;
} /* start_rmt */


/* STOP_RMT: Stop the rmt program.
 */
stop_rmt()
{
    if (tapefd != -1) {
	(void) write(tapefd, "Q\n", 2);
    }
} /* stop_rmt */


/* WRITE_GET_ACK: Write a command and get the acknowledgement.
 */
write_get_ack(buf, size)
    char *buf;
    int size;
{
    char rmtbuf[MAX_RMT_CMD];

    if (write(tapefd, buf, size) != size) {
	prerror("write");
	return -1;
    }
    if (read_rmt_line(rmtbuf, sizeof(rmtbuf)) == -1) return -1;
    switch(rmtbuf[0]) {
	case 'A':
	    return atoi(&rmtbuf[1]);
	case 'E':
	    if (read_rmt_line(rmtbuf, sizeof(rmtbuf)) == -1) return -1;
	    pr("%s", rmtbuf);
	    return -1;
	default:
	    pr("Rmt, protocol screwed up\n");
	    return -1;
    }
} /* write_get_ack */


/* READ_RMT_LINE: Read a line from rmt.
 */
read_rmt_line(buf, size)
    char *buf;
    int size;
{
    while (size--) {
	if (read_rmt(buf, 1) != 1) break;
	if (*buf++ == '\n') {
	    *buf = '\0';
	    return 0;
	}
    }
    pr("Rmt, protocol screwed up\n");
    return -1;
} /* read_rmt_line */


/* READ_RMT: Do a buffered read of size bytes from rmt.
 */
read_rmt(buf, sz)
    char *buf;
    int sz;
{
    static int buf_sz = 0;
    static char iobuf[MAX_RMT_CMD], *iobuf_ptr;
    int n, size = sz;

    while (size) {
	if (buf_sz) {
	    n = min(buf_sz, size);
	    bcopy(iobuf_ptr, buf, n);
	    iobuf_ptr += n;
	    buf_sz -= n;
	    buf += n;
	    size -= n;
	} else if (size < sizeof(iobuf)) {
	    n = read(tapefd, iobuf, sizeof(iobuf));
	    if (n <= 0) return -1;
	    buf_sz = n;
	    iobuf_ptr = iobuf;
	} else {
	    n = read(tapefd, buf, size);
	    if (n <= 0) return -1;
	    buf += n;
	    size -= n;
	}
    }
    return sz;
} /* read_rmt */


/* MOUNT_TAPE: Mount the nth tape from a restore.  Fsf to the right file on
 * the tape.
 */
mount_tape(tname, tapenum, filenum)
    char *tname;
    int tapenum, filenum;
{
    tape_count = tapenum;
    if (!tname || tname[0] == '\0') {
	tapename[0] = '\0';
    } else {
	(void) NCOPY(tapename, tname, sizeof(tapename));
    }

    /* Open the tape */
    if (mount_new_tape() == -1) return -1;

    /* Fsf to the right file */
    if (filenum) {
	return tp_ioctl(MTFSF, filenum);
    }
    return 0;
} /* mount_tape */


/* MOUNT_NEW_TAPE: Change the tape.
 */
mount_new_tape()
{
    static char nice_num[][10] = {
	"first", "second", "third", "fourth", "fifth"};
    char buf[10];
    char name[MAXTAPENAME+2];

    if (chat && (tape_count || *tapename)) {
	if (*tapename) {
	    (void) NCOPY(name, tapename, MAXTAPENAME);
	    (void) strcat(name, " ");
	} else {
	    *name = '\0';
	}
	if (tape_count > NUM_ARR(nice_num)-1) {
	    pr("Insert the %dth %stape into the tape drive\n",
		tape_count+1, name);
	} else {
	    pr("Insert the %s %stape into the tape drive\n",
		nice_num[tape_count], name);
	}
	pr("Hit return when it's ready\n");
	input_line(buf, sizeof(buf));
    }

    /* Open the tape */
    while (tp_open(O_RDONLY) == -1) {
	if (chat) {
	    pr("The tape can't be opened for reading\n");
	    pr("Try again (y/n)? ");
	    input_line(buf, sizeof(buf));
	    if (buf[0] != 'y' && buf[0] != 'Y') return -1;
	} else {
	    return -1;
	}
    }
    ++tape_count;
    return 0;
} /* mount_new_tape */
