/*******************************************************************/
/*      Copyright (C) 1989 by                                      */
/*            The Regents of the University of California          */
/*                                                                 */
/*                             DISCLAIMER                          */
/*                                                                 */
/*      "Although   each   program   has   been  tested  by  its   */
/*      contributor, no warranty, express or implied, is made by   */
/*      the  contributor  or the University of California, as to   */
/*      the accuracy and functioning of the program and  related   */
/*      program material, nor shall the fact of the distribution   */
/*      constitute any such warranty, and no  responsibility  is   */
/*      assumed   by   the  contributor  or  the  University  of   */
/*      California in connection herewith."                        */
/*******************************************************************/

/* SAVER.C: Main program for saver.
 *
 * $Header: /cd/ua/sam/src/ss/saver/RCS/saver.c,v 1.3 89/06/09 15:17:21 sam REL-6-9-89 $
 *
 * $Log:	saver.c,v $
 * Revision 1.3  89/06/09  15:17:21  sam
 * Got rid of I_FILES
 * 
 * Revision 1.2  89/04/03  20:52:36  sam
 * Changed to use the VERSION constant
 * 
 * Revision 1.1  89/03/08  20:53:49  sam
 * Initial revision
 * 
 */
#ifndef lint
static char rcsid[] = "$Header: /cd/ua/sam/src/ss/saver/RCS/saver.c,v 1.3 89/06/09 15:17:21 sam REL-6-9-89 $";
#endif

/*
 * Saver: A filesave program
 *
 * Tim Morgan 8/23/88
 *
 * Flags:
 *      -a <n>		        Give an after argument for incrementals
 *      -i <filename>		ignore file this file (like core)
 *      -x <filenum> <bytes-left> <tape-num> <tape-name>
 *      -c			Don't cross devices
 *	-s			Print statistics to stderr when finished
 *	-v			List directories saved
 *	-V			List all files saved
 *	-f <file>		Do dump to <file>; otherwise, to stdout
 *	-t <listfile>		Write savelist to <listfile>
 *	-b <n>			Set blocking factor to <n>
 *	-d			Interpose ddd on the output; after -f, -h, -b
 *
 * Names of files and directories to save may also be listed on the
 * command line.
 */

#include "../global.h"
#include "saver.h"
#include <stdio.h>
#ifdef SUN40
#include <dirent.h>
#else
#include <sys/dir.h>
#endif
#include <signal.h>
#include <sys/time.h>
#include <sys/file.h>
#include <errno.h>
#include <ctype.h>
#include <sys/mman.h>
#include <netdb.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <sys/wait.h>


/* ISDOTFILE: Return true if the string given is . or ..
 */
#define ISDOTFILE(s) \
	(*s == '.' && (s[1] == '\0' || (s[1] == '.' && s[2] == '\0')))


/* WRITE_FILE: Write out a file being saved either by sending info to our slave
 * or by doing it ourself.
 */
#if NSLAVES == 0
#   define WRITE_FILE(nm, nml, st, type)\
	   write_file(curdir, nm, nml, st, type, tape) 
#else
#   define WRITE_FILE(nm, nml, st, type)\
	   write_slave(nm, nml, st, type)
#endif


/* DENT_TYPE: type of structure returned by readdir.
 */
#ifdef SUN40
    typedef struct dirent dent_type;
#else
    typedef struct direct dent_type;
#endif


/* SAVE_PATH: Structure of what to store for each path to save.
 */
struct save_path {
    char *name;
    int after;
};


/* DUP_INFO: Info to keep for mfiles.
 */
typedef struct dup_info {
    dev_t st_dev;	/* Device this file is on */
    ino_t st_ino;	/* Inode of this file */
    off_t offset;	/* Offset of the file on the tape */
} *dups_ptr;


/* Our global vars */
static char
    *savelist_name="tape.list",	/* Name of savelist file */
    *ignore_file="",		/* Name of a file not to save */
    curdir[MAXDIRLENGTH],	/* Current path during save */
    wd[MAXPATHLEN];		/* CWD when prog was started */
static int
    curdir_len,			/* # of chars in curdir string */
    tape,			/* fd of socket or tape drive (or file) */
    devnum,			/* Device number we're on */
    isinc,			/* Is this an incremental? */
    npaths,			/* Number of entries in the paths array */
    ndups,			/* Number of entries in the dups array */
    pid,			/* Our (the master's) pid */
#if NSLAVES > 0
    slavepid[NSLAVES],		/* Slave pid's */
#endif
    output_buf_size=63*1024;	/* Block size when writing directly to tape */
static int
    data_saved=0,		/* # of bytes we've written out so far */
    symlinks_saved=0,		/* # of symlinks we've processed */
    files_saved=0,		/* # of files we've processed */
    dirs_saved=0,		/* # number of directories we've saved */
    hard_links=0,		/* # of duplicate hardlinked files seen */
    files_avoided=0,		/* # of files we skipped as too old */
    since = 0;			/* Time after which a file should be saved */
    print_stats = FALSE,	/* Print statistics at end */
    verbose = FALSE,
    very_verbose = FALSE,
    dontcross;			/* Don't cross filesystems */
static double
    start_time;			/* Time when we started for statistics */
#if NSLAVES > 0
static FILE
    *slavefp;			/* Slave pipe fp */
#endif
static FILE
    *savelist;			/* File for savelist */
static struct stat
    stbuf;			/* Stat from current file */
static struct save_header
    hd;				/* Header for savelist */
static struct save_path
    *paths;			/* List of paths to save (from command line) */
static dups_ptr
    dups;			/* List of inodes with refcount > 1 */

/* Global vars */
int
    dtablesize,			/* Value from getdtablesize() */
    sys_page_size;		/* System's page size */

/* Extern functions */
extern char
    *sprintf(), *malloc(), *realloc(), *rindex(), *index(), *getwd(),
    *sbrk(), *mktemp(), *strcpy(), *strncpy();
extern time_t
    time();
extern void
    slave(), write_file();

#ifndef htonl
extern unsigned long
    htonl();
extern unsigned short
    htons();
#endif

/* Forward decls */
static void
    flush_slave(), init(), parse_args(), write_headers(), finish_slist(),
    start_slaves(), stop_slaves(), save_path(), myexit(),
    add_to_curdir(), curdir_up();


/* WRITE_SLIST: Write something to the savelist file.
 */
static void write_slist(buf, size)
    char *buf;
{
    if (fwrite(buf, size, 1, savelist) != 1) {
	perror(savelist_name);
	myexit(1);
    }
} /* write_slist */


/* WRITE_SLAVE: Write something to the slave(s).
 */
#if NSLAVES > 0
static void write_slave(name, name_len, st, type)
    char *name;
    int name_len;
    struct stat *st;
    int type;
{
    static char *perr = "pipe write";
    struct slave_info si;	/* Info to send to slave */

    /* Copy everything into si */
    bcopy((char*)st, (char*)&si.st, sizeof(si.st));
    si.path_len = name_len ? curdir_len : 0;
    si.type = type;
    si.name_len = name_len;

    /* Write out the si record */
    if (fwrite((char*)&si, sizeof(si), 1, slavefp) != 1) {
	perror(perr);
	myexit(1);
    }

    /* Now write the pathname and filename if needed */
    if (name_len) {
	if (fwrite(curdir, curdir_len+1, 1, slavefp) != 1) {
	    perror(perr);
	    myexit(1);
	}
	if (fwrite(name, name_len+1, 1, slavefp) != 1) {
	    perror(perr);
	    myexit(1);
	}
    }
} /* write_slave */
#endif


/* FLUSH_SLAVE: Make sure the slave's output buf is flushed.
 */
#if NSLAVES > 0
static void flush_slave()
{
    if (fflush(slavefp) != 0) {
	perror("fflush");
	myexit(1);
    }
} /* flush_slave */
#endif


/* WRITE_ALL: Do a write(2), but iterate until everything is written or error.
 */
static int write_all(fd, buf, size)
    int fd, size;
    char *buf;
{
    int res;

    while (size) {
	if ((res = write(fd, buf, size)) == -1 || res == 0) {
	    return -1;
	}
	size -= res;
	buf += res;
    }
    return 0;
} /* write_all */


/* WRITE_INFO: Write info about a file to the slaves and to the savelist.
 */
static void write_info(type, name)
    int type;
    char *name;
{
    struct file_info info;	/* Info to put into savelist */

    /* Make info to put into savelist */
    info.st_size = htonl((u_long) stbuf.st_size);
    info.st_atime = htonl((u_long) stbuf.st_atime);
    info.st_mtime = htonl((u_long) stbuf.st_mtime);
    info.st_mode = htons(stbuf.st_mode);
    info.st_uid = htons((u_short) stbuf.st_uid);
    info.st_gid = htons((u_short) stbuf.st_gid);
    info.record_type = type;
    info.name_len = name ? strlen(name) : 0;

    /* Don't write I_FILE's to the savelist */
    if (type != I_FILE) {
	/* Write the info rec to the savelist */
	write_slist((char*)&info, FILE_INFO_SIZE);

	/* Write file name to the savelist */
	if (info.name_len) {
	    write_slist(name, (int)info.name_len);
	}
    }

    /* Update data_saved */
    switch(type) {
	case SYMLINK:
	case A_FILE:
	case M_FILE:
	    data_saved += FILE_INFO_SIZE + info.name_len + stbuf.st_size;
	    break;
	default:
	    data_saved += FILE_INFO_SIZE + info.name_len;
	    break;
    }

    /* Go write the file */
    WRITE_FILE(name, (int)info.name_len, &stbuf, type);
} /* write_info */


/* SAVE_SYMLINK: Save a symlink onto the tape
 */
static void save_symlink(name)
    char *name;
{
    /* Don't save if this is an incremental and file hasn't changed */
    if (stbuf.st_ctime < since) {
	++files_avoided;
	return;
    }
    write_info(SYMLINK, name);
    ++symlinks_saved;
} /* save_symlink */


/* SAVE_DIR: Recursively save directory dir
 */
static int save_dir(dir)
    char *dir;
{
    DIR *d;
    dent_type *dent, *dlist;
    int num_alloced, n, retval = 0;

    if (chdir(dir) < 0) {
	(void) fprintf(stderr, "Cannot chdir to %s/", curdir);
	perror(dir);
	return retval;	/* Plunge ahead anyway */
    }

    d = opendir(".");
    if (!d) {
	(void) fprintf(stderr, "Can't open directory %s/", curdir);
	perror(dir);
    } else {
	/* Write a record for this dir */
	write_info(CD_DOWN, dir);

	/* Keep track of pathname of curdir */
	add_to_curdir(dir);

	/* Collect the entries from this dir */
	for (dlist = NULL, num_alloced = n = 0; dent = readdir(d);) {

	    /* Check whether we should ignore this. */
	    if (!ISDOTFILE(dent->d_name) && strcmp(dent->d_name, ignore_file))
	    {
		/* Alloc some space if needed */
		if (n >= num_alloced) {
		    num_alloced = num_alloced ? num_alloced*2 : 10;
		    dlist = ALLOC(dlist, dent_type, num_alloced);
		    if (dlist == NULL) {
			fputs("Out of memory.\n", stderr);
			myexit(1);
		    }
		}

		/* Copy into the array */
		bcopy((char*)dent, (char*)&dlist[n], sizeof(dent_type));
		++n;
	    }
	}
	(void) closedir(d);

	/* Save everthing in this directory */
	if (n) {
	    for (dent = &dlist[n-1]; dent >= dlist; --dent) {
		if (save_something(dent->d_name)) {
		    /*
		     * Couldn't return to this dir, so give up on saving it
		     */
		    retval = -1;
		    break;
		}
	    }
	    free((char*)dlist);
	}

	if (verbose) {
	    (void) fprintf(stderr, "Directory %s\n", curdir);
	}

	/* Remove this dir from curdir */
	curdir_up();

	write_info(CD_UP, (char*)NULL);
	++dirs_saved;
    }

    if (retval == -1 || chdir("..") == -1) {
	if (chdir(curdir) == -1) {
	    (void) fprintf(stderr, "Warning: can't chdir back to %s\n", curdir);
	    retval = -1;
	}
    }
    return retval;
} /* save_dir */


/* ADD_DUP: This routine adds one more file to the list of possible duplicates
 * (those whose ref count is > 1).  It expands the dups list if necessary.
 */
static void add_dup()
{
    static maxdups = 0;

    if (ndups >= maxdups) {
	maxdups += DUP_INCREMENT;
	dups = ALLOC(dups, struct dup_info, maxdups);
	if (!dups) {
	    (void) fprintf(stderr, "Out of memory\n");
	    myexit(1);
	}
    }
    dups[ndups].st_dev = stbuf.st_dev;
    dups[ndups].st_ino = stbuf.st_ino;
    dups[ndups++].offset = data_saved;
} /* add_dup */


/* IN_DUP_LIST: Return index in dups list of this file if there, else -1
 */
static int in_dup_list()
{
    register struct dup_info *p;

    if (ndups) {
	for (p = &dups[ndups-1]; p >= dups; --p) {
	    if (stbuf.st_ino == p->st_ino && stbuf.st_dev == p->st_dev) {
		return p - dups;
	    }
	}
    }
    return(-1);
} /* in_dup_list */


/* DUPLICATE_FILE: Return TRUE if this is a file we've already seen
 */
static int duplicate_file(name)
char *name;
{
    int i;

    if (stbuf.st_nlink < 2) return(FALSE);
    if ((i = in_dup_list()) < 0) {	/* We've not seen this one before */
	add_dup();
	return(FALSE);
    }
    stbuf.st_size = dups[i].offset;	/* special encoding */
    write_info(L_FILE, name);
    return(TRUE);
} /* duplicate_file */


/* SAVE_FILE: Save a file onto tape
 */
static void save_file(name)
    char *name;
{
    /* Don't save if this is an incremental and file hasn't changed */
    if (stbuf.st_ctime < since) {
	++files_avoided;
	return;
    }

    /* Check for hard to an already saved file */
    if (duplicate_file(name)) {
	++hard_links;
	return;
    }

    /* Write out the file */
    write_info(stbuf.st_nlink > 1 ? M_FILE : A_FILE, name);

    if (very_verbose) {
	(void) fprintf(stderr, "%s/%s %d\n", curdir, name, stbuf.st_size);
    }
    ++files_saved;
} /* save_file */


/* SAVE_SOMETHING: Figure out what something is and save it.
 */
static int save_something(name)
    char *name;
{
    int retval = 0;

    if (lstat(name, &stbuf) < 0) {   /* It must have been deleted already */
	if (verbose) {
	    (void) fprintf(stderr, "Can't stat %s/", curdir);
	    perror(name);
	}
	return retval;
    }

    switch (stbuf.st_mode & S_IFMT) {
    case S_IFDIR:	/* Directory */
	/* Stop if crossing devices */
	if (stbuf.st_dev != devnum) {
	    if (devnum == -1 || !dontcross) {
		devnum = stbuf.st_dev;
	    } else {
		return retval;
	    }
	}
	retval = save_dir(name);
	break;
    case S_IFCHR:	/* Character special file */
    case S_IFBLK:	/* Block special file */
	/* Skip it */
	break;
    case S_IFREG:	/* Regular file */
	save_file(name);
	break;
    case S_IFLNK:	/* Symbolic link */
	save_symlink(name);
	break;
    case S_IFSOCK:	/* Unix-domain socket */
    case S_IFIFO:	/* FIFO file */
	/* Skip it */
	break;
    default:
	(void) fprintf(stderr, "file of unknown type %d: %s\n",
		stbuf.st_mode & S_IFMT, name);
    }
    return retval;
} /* save_something */


/* SIGHANDLER: Handle a signal from a pipe
 */
sigtype sighandler()
{
    (void) fprintf(stderr, "\nProgram error occurred, aborting.\n");
    (void) fprintf(stderr, "Save incomplete!!\n");
    myexit(1);
} /* sighandler */


/* COREDUMPER: Dump a core into our starting directory.
 */
sigtype coredumper()
{
    /* Secret encoding for error status */
    (void) fputs("~1\n", stderr);
    (void) fflush(stderr);

    (void) chdir(wd);
    (void) signal(SIGSEGV, SIG_DFL);
    (void) kill(getpid(), SIGSEGV);
} /* coredumper */


/* INVOKE_DDD: Fork/exec ddd and make "tape" the fd for sending to ddd.
 */
static void invoke_ddd()
{
    int fds[2];
    char bufsiz[50];

    if (pipe(fds) < 0)
	perror("pipe"), myexit(1);
    switch (fork()) {
    default:	/* parent */
	(void) close(tape);	/* Don't need that fd any more */
	tape = fds[1];		/* Our output now goes into the pipe */
	(void) close(fds[0]);	/* Which we won't be reading from */
	return;
    case 0:	/* child */
	(void) close(fds[1]);
	if (dup2(fds[0], 0) < 0 || dup2(tape, 1) < 0)
	    perror("dup2"), myexit(1);
	if (tape > 2) (void) close(tape);
	if (fds[0] > 2) (void) close(fds[0]);
	(void) sprintf(bufsiz, "bs=%d", output_buf_size);
	(void) execl(DDD, "ddd", bufsiz, (char *) 0);
	perror(DDD);
	exit(1);
    case -1:
	perror("fork");
	myexit(1);
    }
} /* invoke_ddd */


/* ROUND_OFF: Round a floating point value to two decimal places
 */
static double round_off(x)
double x;
{
    int i;
    i = x * 100.0 + 0.5;
    return(i/100.0);
} /* round_off */


/* MAIN:
 */
main(argc, argv)
    char *argv[];
{
    int i;
    char name[MAXPATHLEN];
    struct timeval tp;
    struct timezone tz;
    double end_time;

    /* Initialize */
    init(argc, argv);

#if NSLAVES > 0
    /* Start up the slaves */
    start_slaves();
#endif

    /* Write headers to savelist and tape */
    write_headers();

    /* Save any files specified on the command line */
    for (i = 0; i < npaths; ++i) {
	devnum = -1;
	since = paths[i].after;
	if (access(paths[i].name, R_OK) < 0) {
	    perror(paths[i].name);
	} else {
	    save_path(paths[i].name);
	}
    }

    /* If no paths on command line, read filenames from stdin */
    if (!npaths) {
	while (fgets(name, sizeof(name), stdin)) {
	    devnum = -1;
	    name[strlen(name)-1] = '\0';
	    if (access(name, R_OK) < 0) {
		perror(name);
	    } else {
		save_path(name);
	    }
	}
    }

    /* Finish up savelist */
    finish_slist();

#if NSLAVES > 0
    /* Tell the slaves to quit */
    stop_slaves();
#endif

    /* Print statistics */
    if (print_stats) {
	double kb;
	if (gettimeofday(&tp, &tz) < 0)
	    perror("gettimeofday"),myexit(1);
	end_time = tp.tv_sec + tp.tv_usec/1000000.0 - start_time;
	kb = data_saved / 1024.0;
	(void) fprintf(stderr,
"%ld files, %ld directories, %ld symlinks, %ld hard links, %ld files skipped\n",
		files_saved, dirs_saved, symlinks_saved, hard_links,
		files_avoided);
	(void) fprintf(stderr, "%g KB, %g seconds, %g KB/sec\n",
		round_off(kb), round_off(end_time), round_off(kb/end_time));
    }

    if (chdir(wd) == -1) {
	perror("Sorry, can't chdir back to original directory");
    }
    myexit(0);
} /* main */


/* INIT: Initialize and parse command line.
 */
static void init(argc, argv)
    char **argv;
{
    int i;

    tape = 1;
    if ((int)signal(SIGPIPE, sighandler) == -1
	|| (int)signal(SIGSEGV, coredumper) == -1)
    {
	perror("signal");
	myexit(1);
    }
    (void) umask(022);
    if (getwd(wd) == 0) {
	(void) fprintf(stderr, "getwd: %s\n", wd);
	myexit(1);
    }
    pid = getpid();
    dtablesize = getdtablesize();
#if defined(sequent) && defined(i386)
    /* getpagesize doesn't work for mmap's on symmetry, but this does */
    sys_page_size = 8192;
#else
    sys_page_size = getpagesize();
#endif sequent

#ifdef	SO_SNDBUF
    /* If the output is a TCP connection, make it more efficient */
    for (i = SENDSIZE;
         setsockopt(tape, SOL_SOCKET, SO_SNDBUF, (caddr_t)&i, sizeof(i)) == -1
	 && i > 0; i -= 1024);
#endif
    parse_args(argc, argv);
} /* init */


/* GET_NUM_ARG: Get a numeric arg for the opt passed, exit if non-numeric.
 */
static int get_num_arg(opt, arg)
    int opt;
    char *arg;
{
    char *s = arg;

    while (*arg) {
	if (!isdigit(*arg++)) {
	    (void) fprintf(stderr, "Argument to %c is not an integer\n", opt);
	    myexit(1);
	}
    }
    return atoi(s);
} /* get_num_arg */


/* ADD_PATH: Add a path to be saved and an after value to the paths array.
 */
static void add_path(path, after)
    char *path;
    int after;
{
    static num_alloced = 0;

    if (npaths >= num_alloced) {
	num_alloced += 10;
	paths = ALLOC(paths, struct save_path, num_alloced);
	if (paths == NULL) {
	    fputs("Out of memory\n", stderr);
	    myexit(1);
	}
    }
    paths[npaths].name = path;
    paths[npaths].after = after;
    ++npaths;
} /* add_path */


/* MISSING: Tell the user an arg is missing for the passed option and exit.
 */
static void missing(opt)
    int opt;
{
    (void) fprintf(stderr, "Argument to -%c is missing\n", opt);
    myexit(1);
} /* missing */


/* PARSE_ARGS: Parse up the command line
 */
static void parse_args(argc, argv)
    int argc;
    char **argv;
{
    int i;
    struct timeval tp;
    struct timezone tz;

    isinc = 0;
    for (i = 1; i < argc; i++) {
	if (argv[i][0] == '-') switch (argv[i][1]) {
	    case 'a':
		if (++i >= argc) {
		    missing('a');
		}
		if (since = get_num_arg('a', argv[i])) {
		    isinc = 1;
		}
		break;
	    case 'i':
		if (++i >= argc) {
		    missing('i');
		}
		ignore_file = argv[i];
		break;
	    case 'x':
		if (++i >= argc-3) {
		    missing('x');
		}
		hd.file_offset = htonl((u_long)get_num_arg('x', argv[i++]));
		hd.bytes_left = htonl((u_long)get_num_arg('x', argv[i++]));
		hd.tapenum = htonl((u_long)get_num_arg('x', argv[i++]));
		(void) NCOPY(hd.tapename, argv[i], sizeof(hd.tapename));
		break;
	    case 'v':
		verbose = TRUE;
		break;
	    case 'V':
		very_verbose = TRUE;
		break;
	    case 'f':
		if (++i >= argc) {
		    missing('f');
		}
		tape = open(argv[i], O_CREAT|O_WRONLY|O_TRUNC, 0666);
		if (tape < 0) {
		    perror(argv[i]);
		    myexit(1);
		}
		break;
	    case 's':
		print_stats = TRUE;
		if (gettimeofday(&tp, &tz) < 0) {
		    perror("gettimeofday");
		    myexit(1);
		}
		start_time = tp.tv_sec + tp.tv_usec/1000000.0;
		break;
	    case 't':
		if (++i >= argc) {
		    missing('t');
		}
		savelist_name = argv[i];
		break;
	    case 'd':
		invoke_ddd();
		break;
	    case 'b':
		if (++i >= argc) {
		    missing('b');
		}
		output_buf_size = get_num_arg('b', argv[i]);
		break;
	    case 'c':
		dontcross = TRUE;
		break;
	    default:
		(void) fprintf(stderr, "%s: unknown flag %s\n", *argv, argv[i]);
	} else {
	    add_path(argv[i], since);
	}
    }
} /* parse_args */


/* WRITE_HEADERS: Open the savelist, write a header to the savelist and tape.
 */
static void write_headers()
{
    char *s;

    /* Open the file */
    savelist = fopen(savelist_name, "w+");
    if (!savelist) {
	perror(savelist_name);
	myexit(1);
    }

    /* Put version, begin time, end time, tape file #, tape block # in hd */
    bcopy(VERSION, hd.version, sizeof(hd.version));
    hd.type = isinc ? SAVE_INC : 0;
    hd.begin_time = htonl((u_long)time((time_t*)0));
    hd.end_time = 0; /* Unknown now */
    s = rindex(savelist_name, '/');
    (void) NCOPY(hd.slistname, s ? s+1 : savelist_name, sizeof(hd.slistname));
    /* Rest of the fields were set by -x or are zero */

    /* Write hd to tape */
    (void) write_all(tape, (char*)&hd, SAVE_HEADER_SIZE);

    /* Write hd to savelist */
    hd.type |= SAVE_SLIST;
    write_slist((char*)&hd, SAVE_HEADER_SIZE);

    data_saved = SAVE_HEADER_SIZE;
} /* write_headers */


/* FINISH_SLIST: Write end time into header of savelist and close.
 */
static void finish_slist()
{
    write_info(TYPE_END, (char*)NULL);
    hd.end_time = htonl((u_long)time((time_t*)0));
    if (fseek(savelist, 0L, 0) == -1) {
	fputs("fseek failed", stderr);
    } else {
	 write_slist((char*)&hd, SAVE_HEADER_SIZE);
    }
    if (fclose(savelist) != 0) {
	perror("fclose");
	myexit(1);
    }
} /* finish_slist */


/* START_SLAVES: Start up the slaves.
 */
#if NSLAVES > 0
static void start_slaves()
{
    int i, j, maxsinfo = MAXSINFO, pipes[2], outbfsiz;
    int fds[2][NSLAVES];
    char tmpname[100], *buf;

    /* Start receiving signals from the children */
    if (signal(SIGCHLD, sighandler)) {
	perror("Warning: signal");
    }

#if NSLAVES > 1
    /* Make lock files */
    (void) strcpy(tmpname, "/tmp/saverlockXXXXXX");
    (void) mktemp(tmpname);
    for (i = 0; i < 2; ++i) {
	fds[i][0] = open(tmpname, O_RDONLY|O_CREAT|O_EXCL, 0600);
	if (fds[i][0] == -1) {
	    fputs("Can't make temp file ", stderr);
	    perror(tmpname);
	    myexit(1);
	}
	for (j = 1; j < NSLAVES; ++j) {
	    fds[i][j] = open(tmpname, O_RDONLY);
	    if (fds[i][j] == -1)  {
		fputs("Can't find temp file ", stderr);
		perror(tmpname);
		myexit(1);
	    }
	}
	if (unlink(tmpname) == -1) {
	    fputs("Can't remove temp file ", stderr);
	    perror(tmpname);
	    myexit(1);
	}
    }
#endif

    /* Make pipe for sending commands */
    if (pipe(pipes) == -1) {
	perror("pipe");
	myexit(1);
    }

#ifdef	SO_SNDBUF
    /* Make sure pipe will hold one complete record */
    outbfsiz = setsockopt(pipes[W_END], SOL_SOCKET, SO_SNDBUF,
		          (caddr_t)&maxsinfo, sizeof(maxsinfo))
	       == -1 ? 4096 : maxsinfo;
#else
    outbfsiz = 4096;
#endif

    /* Start up the slaves */
    for (i = 0; i < NSLAVES; ++i) {
	switch(slavepid[i] = fork()) {
	    case -1:
		perror("fork");
		if (i == 0) myexit(1);
	    case 0:
		(void) signal(SIGSEGV, coredumper);
#ifdef SO_RCVBUF
		(void) setsockopt(pipes[R_END], SOL_SOCKET, SO_RCVBUF,
				  (caddr_t)&maxsinfo, sizeof(maxsinfo));
#endif
		slave(pipes[R_END], tape, fds[0][i], fds[1][i]);
		fputs("Argh! slave returned\n", stderr);
	    default:
		break;
	}
    }
    (void) close(pipes[R_END]);

    /* Make an fp for the slave pipe */
    slavefp = fdopen(pipes[W_END], "w");
    if ((buf = NEW(char, outbfsiz)) == NULL) {
	perror("Out of memory");
	myexit(1);
    }
    setbuffer(slavefp, buf, outbfsiz);

#if NSLAVES > 1
    for (i = 0; i < 2; ++i) {
	for (j = 0; j < NSLAVES; ++j) {
	    (void) close(fds[i][j]);
	}
    }
#endif
} /* start_slaves */
#endif


/* KILL_SLAVES: Kill off our slaves.
 */
#if NSLAVES > 0
static void kill_slaves()
{
    int i;

    /* Stop getting signals from our children */
    if ((int)signal(SIGCHLD, SIG_DFL) == -1) {
	perror("signal");
	myexit(1);
    }
    for (i = 0; i < NSLAVES; ++i) {
	if (slavepid[i] > 1 && kill(slavepid[i], SIGTERM) != -1) {
	    (void) wait((union wait *)0);
	}
    }
} /* kill_slaves */
#endif


/* STOP_SLAVES: Stop our slaves in a nice way.
 */
#if NSLAVES > 0
static void stop_slaves()
{
    int i;

    /* Stop receiving signals from our children */
    if ((int)signal(SIGCHLD, SIG_DFL) == -1) {
	perror("signal");
	myexit(1);
    }

    /* Send quit record to all slaves */
    for (i = 0; i < NSLAVES; ++i) {
        WRITE_FILE((char*)NULL, 0, &stbuf, TYPE_QUIT);
    }

    /* Flush the pipe to the slaves */
    flush_slave();

    /* Wait for the slaves to exit */
    for (i = 0; i < NSLAVES; ++i) {
	(void) wait((union wait *)0);
    }
} /* stop_slaves */
#endif


/* SAVE_PATH: Save a pathname by breaking it down and saving each part.
 */
static void save_path(path)
    char *path;
{
    char *p, *pnm = path;

    /* Starting at root (not necessarily the system's root) */
    write_info(CD_ROOT, (char*)NULL);

    /* Get a starting directory */
    curdir_len = 0;
    if (*pnm == '/') {
	while (*pnm == '/') ++pnm;
	add_to_curdir("/");
    } else {
	add_to_curdir(wd);
    }

    /* Change to starting directory */
    if (chdir(curdir) == -1) {
	fputs("Can't cd to ", stderr);
	perror(curdir);
	myexit(1);
    }

    /* Cd into each part of the path, writing PATH_DIR's as we go */
    for (; p = index(pnm, '/'); pnm = p+1) {
	*p = '\0';
	if (stat(pnm, &stbuf) == -1 || chdir(pnm) == -1) {
	    perror(path);
	    myexit(1);
	}
	write_info(PATH_DIR, pnm);
	add_to_curdir(pnm);
	*p = '/';
    }
    (void) save_something(pnm);
} /* save_path */


/* MYEXIT: Exit, but first send our exit status over stderr
 */
static void myexit(status)
{
#if NSLAVES > 0
    /* Kill off our slaves */
    if (status) kill_slaves();
#endif

    /* Secret encoding for exit status */
    (void) fprintf(stderr, "~%d\n", status);
    (void) fflush(stderr);
    exit(status);
} /* myexit */


/* ADD_TO_CURDIR: Add a directory to the curdir.
 */
static void add_to_curdir(dir)
    char *dir;
{
    int len = strlen(dir);

    if ((curdir_len && curdir[curdir_len-1] == '/')
	|| (len && dir[len-1] == '/'))
    {
	(void) strcpy(curdir+curdir_len, dir);
    } else {
	curdir[curdir_len] = '/';
	(void) strcpy(curdir+curdir_len+1, dir);
	++len;
    }
    curdir_len += len;
} /* add_to_curdir */


/* CURDIR_UP: Remove last dir from curdir.
 */
static void curdir_up()
{
    char *p = rindex(curdir, '/');

    curdir_len = p ? p - curdir : 0;
    curdir[curdir_len] = '\0';
} /* curdir_up */
