/*******************************************************************/
/*      Copyright (C) 1989 by                                      */
/*            The Regents of the University of California          */
/*                                                                 */
/*                             DISCLAIMER                          */
/*                                                                 */
/*      "Although   each   program   has   been  tested  by  its   */
/*      contributor, no warranty, express or implied, is made by   */
/*      the  contributor  or the University of California, as to   */
/*      the accuracy and functioning of the program and  related   */
/*      program material, nor shall the fact of the distribution   */
/*      constitute any such warranty, and no  responsibility  is   */
/*      assumed   by   the  contributor  or  the  University  of   */
/*      California in connection herewith."                        */
/*******************************************************************/

/* SLAVE.C: Slave process, takes in stat info and writes on output.
 *
 * $Header: /cd/ua/sam/src/ss/saver/RCS/slave.c,v 1.3 89/06/09 15:17:47 sam REL-6-9-89 $
 *
 * $Log:	slave.c,v $
 * Revision 1.3  89/06/09  15:17:47  sam
 * Changed check for read error to detect EFAULT.
 * 
 * Revision 1.2  89/04/03  20:54:16  sam
 * Fixed so that it tries to skip over a file when there is an error while
 * writing from an mmap'ed file.
 * Also changed so that files w/size < IOBLKSIZ do a read/write instead of
 * an mmap/write
 * 
 * Revision 1.1  89/03/08  20:54:06  sam
 * Initial revision
 * 
 */
#ifndef lint
static char rcsid[] = "$Header: /cd/ua/sam/src/ss/saver/RCS/slave.c,v 1.3 89/06/09 15:17:47 sam REL-6-9-89 $";
#endif

#include "../global.h"
#include "saver.h"
#include <stdio.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/mman.h>
#include <sys/file.h>
#include <sys/uio.h>
#include <netinet/in.h>
#include <signal.h>
#include <errno.h>


/* HDSIZE: Maximum size for one file's header */
#define HDSIZE (FILE_INFO_SIZE+MAXFILENAME)

/* IOBLKSIZ: Size of blocks we'll do I/O with */
#ifdef SUN3X
#define IOBLKSIZ 64512
#else
#define IOBLKSIZ 8192
#endif

/* Local vars */
static char
    *mapadr,				/* Where mmap took place */
    outbuf[IOBLKSIZ];			/* I/O buffer */
static int
    maplen,				/* Number of bytes mmaped */
    rd_fd,				/* FD to read in files */
    debug = 0,				/* Do debugging */
    nvecs,				/* Number of items in outvecs */
    wrbytes,				/* Number of bytes left to write */
    fakeit;				/* Start zero-fill on output? */
static struct iovec
    outvecs[5];				/* I/O vectors for writing */

/* Externs */
extern int
    dtablesize,
    sys_page_size;

/* Function decls */
static void
    prepare_write(), prep_symlink(), prep_afile(), do_write(), add_wvec(),
    slave_abort(), file_err(), get_lock(), give_lock();
extern char
    *strcpy();
extern caddr_t
    sbrk();


/* SLAVE: Main slave process loop, controls command reading and
   reading/writing the file.
 */
#if NSLAVES > 0
void slave(cmd_fd, wr_fd, rdlock, wrlock)
    int cmd_fd, wr_fd, rdlock, wrlock;
{
    char *path, *name;
    int name_len, type;
    struct stat *st;

    for (;;) {
#if NSLAVES > 1
	/* Get perm to read a cmd */
	get_lock(rdlock);

	if (debug) {
	    (void) fprintf(stderr, "(%d) Got ok to read\n", getpid());
	}
#endif

	/* Get a cmd, exit if told to */
	if (read_cmd(cmd_fd, &path, &name, &name_len, &type, &st) == -1) {
	    if (debug) {
		(void) fprintf(stderr, "(%d) exiting normally\n", getpid());
	    }
#if NSLAVES > 1
	    give_lock(rdlock);
#endif
	    exit(0);
	}

	/* Make sure we're in the right directory to access this file */
	fakeit = do_chdir(path) == -1;

	/* Get everything ready to write */
	prepare_write(path, name, name_len, st, type);

#if NSLAVES > 1
	/* Make sure we'll go next before allowing next proc to get cmd */
	get_lock(wrlock);

	if (debug) {
	    (void) fprintf(stderr, "(%d) Giving lock to read\n", getpid());
	}

	/* Allow cmd reading */
	give_lock(rdlock);
#endif

	/* Write */
	do_write(path, name, type, wr_fd, st);

#if NSLAVES > 1
	/* Allow next process to write */
	give_lock(wrlock);
#endif
    }
} /* slave */
#endif


/* WRITE_FILE: Called to do writes when there are no slaves.
 */
#if NSLAVES == 0
void write_file(path, name, name_len, st, type, wr_fd)
    char *path, *name;
    int wr_fd;
    struct stat *st;
{
    if (name_len != -1) {
	fakeit = 0;
	if (debug) {
	    (void) fprintf(stderr, "doing %s/", path);
	    (void) fwrite(name, name_len, 1, stderr);
	}
	prepare_write(path, name, name_len, st, type);
	do_write(path, name, type, wr_fd, st);
    }
} /* write_file */
#endif


/* GET_LOCK: Aquire an exclusive lock for fd.
 */
#if NSLAVES > 1
static void get_lock(fd)
    int fd;
{
    if (flock(fd, LOCK_EX) == -1) {
	perror("flock");
	slave_abort();
    }
} /* get_lock */
#endif


/* GIVE_LOCK: Give away our lock on fd.
 */
#if NSLAVES > 1
static void give_lock(fd)
    int fd;
{
    if (flock(fd, LOCK_UN) == -1) {
	perror("flock");
	slave_abort();
    }
} /* give_lock */
#endif


/* BUF_READ: Do a buffered read of the command channel.
 */
#if NSLAVES == 1
buf_read(fd, buf, size)
    int fd, size;
    char *buf;
{
    static int num_in_buf = 0;
    static char iobuf[MAXSINFO], *iobuf_ptr;
    int n;

    while (size) {
	if (num_in_buf) {
	    n = min(num_in_buf, size);
	    bcopy(iobuf_ptr, buf, n);
	    iobuf_ptr += n;
	    num_in_buf -= n;
	    buf += n;
	    size -= n;
	} else {
	    n = read(fd, iobuf, sizeof(iobuf));
	    if (n <= 0) return -1;
	    num_in_buf = n;
	    iobuf_ptr = iobuf;
	}
    }
    return 0;
} /* buf_read */
#endif


/* READ_CMD: Get a command from the pipe.
 */
#if NSLAVES > 0
static int read_cmd(cmd_fd, path, name, name_len, type, st)
    int cmd_fd, *name_len, *type;
    char **path, **name;
    struct stat **st;
{
    static struct slave_info si;
    static char namebuf[MAXFILENAME+MAXDIRLENGTH+2];
    int len;

    /* Read in a slave_info record */
#if NSLAVES == 1
    if (buf_read(cmd_fd, (char*)&si, sizeof(si)) == -1) {
#else
    if (read_all(cmd_fd, (char*)&si, sizeof(si)) == -1) {
#endif
	perror("pipe read");
	slave_abort();
    }

    if (debug) {
	(void) fprintf(stderr, "\n(%d) got command packet\n", getpid());
	(void) fprintf(stderr,
	        "st_size is %d, name_len is %d, path_len is %d, type is %d\n",
		si.st.st_size, si.name_len, si.path_len, si.type);
    }

    /* Return quit code if master tells us to quit */
    if (si.type == TYPE_QUIT) return -1;

    /* Calc length of path and file names */
    len = si.name_len + si.path_len + 2;

    /* Read in path and file names */
    if (len > 2) {
#if NSLAVES == 1
	if (buf_read(cmd_fd, namebuf, len) == -1) {
#else
	if (read_all(cmd_fd, namebuf, len) == -1) {
#endif
	    perror("pipe read");
	    slave_abort();
	}
	if (debug) {
	    (void) fprintf(stderr,
		       "got file '%s/%s'\n", namebuf, namebuf+si.path_len+1);
	}
    }

    /* Copy out all the parameters */
    *path = namebuf;
    *name = namebuf + si.path_len + 1;
    *name_len = si.name_len;
    *type = si.type;
    *st = &si.st;

    return 0;
} /* read_cmd */
#endif


/* PREP_HEAD: Make a header and add it to the list of things to write.
 */
static void prep_head(name, name_len, st, type)
    char *name;
    int name_len, type;
    struct stat *st;
{
    static union {
	char hdbuf[HDSIZE];
	struct file_info fi;
    } hd;

    /* Copy in all the fields */
    hd.fi.st_size = htonl((u_long) st->st_size);
    hd.fi.st_atime = htonl((u_long) st->st_atime);
    hd.fi.st_mtime = htonl((u_long) st->st_mtime);
    hd.fi.st_mode = htons(st->st_mode);
    hd.fi.st_uid = htons((u_short) st->st_uid);
    hd.fi.st_gid = htons((u_short) st->st_gid);
    hd.fi.record_type = (char)type;
    hd.fi.name_len = (char)name_len;

    /* Append the name to the end */
    bcopy(name, &hd.hdbuf[FILE_INFO_SIZE], name_len);

    /* Put in a vector to write the header */
    add_wvec(hd.hdbuf, FILE_INFO_SIZE + name_len);
} /* prep_head */


/* PREPARE_WRITE: Get ready to write.
 */
static void prepare_write(path, name, name_len, st, type)
    char *path, *name;
    int name_len, type;
    struct stat *st;
{
    /* Prepare to write the file header */
    prep_head(name, name_len, st, type);

    switch(type) {
	case A_FILE:
	case M_FILE:
	    prep_afile(path, name, st);
	    break;
	case SYMLINK:
	    prep_symlink(path, name, st);
	    break;
	default:
	    break;
    }
} /* prepare_write */


/* PREP_SYMLINK: Prepare for writing a symlink.
 */
static void prep_symlink(path, name, st)
    char *path, *name;
    struct stat *st;
{
    int res;

    /* Read the symlink */
    if (fakeit) {
	res = 0;
    } else {
	if ((res = readlink(name, outbuf, sizeof(outbuf))) == -1) {
	    file_err("Warning: can't read symlink", path, name);
	    res = 0;
	}
    }

    /* If can't read it, or it changed length, lie about its contents */
    if (res < st->st_size) {
	bzero(&outbuf[res], (int)(st->st_size - res));
    }

    /* Add to list of things to write out */
    add_wvec(outbuf, (int)st->st_size);
} /* prep_symlink */


/* PREP_AFILE: Prepare for writing an A_FILE (or M_FILE).
 */
static void prep_afile(path, name, st)
    char *path, *name;
    struct stat *st;
{
    int n;

    maplen = 0;
    rd_fd = -1;
    wrbytes = st->st_size;

    /* Return if we're not supposed to try to read the file */
    if (fakeit || !wrbytes) return;

    /* Open the file */
    rd_fd = open(name, 0);
    if (rd_fd == -1) {
	file_err("Warning: can't open file", path, name);
	fakeit = 1;
    } else {
#ifdef SUN3X
	/* On 3.X, do all read/writes, no mmap */
	maplen = 0;
#else /* sequent or SUN40 */
	if (wrbytes <= IOBLKSIZ) {
	    /* We'd rather do one read than an mmap */
	    maplen = 0;
	} else {
#ifdef SUN40
	    maplen = wrbytes;
	    mapadr = mmap((caddr_t)0, (int)maplen, PROT_READ, MAP_SHARED,
			  rd_fd, 0L);
	    if (mapadr == (caddr_t)-1) {
		perror("mmap");
		maplen = 0;
	    }
#else /* sequent */
	    maplen = wrbytes & ~(sys_page_size-1);
	    if (maplen > 0) {
		mapadr = (caddr_t) (((int)sbrk(0)+(sys_page_size-1)) &
				   ~(sys_page_size-1));
		if (mmap(mapadr, maplen, PROT_READ, MAP_PRIVATE, rd_fd, 0)
		    == -1)
		{
		    maplen = 0;
		}
	    }
#endif
	}
#endif
	/* Add mapped in area as a buffer to write out */
	if (maplen) {
	    add_wvec(mapadr, maplen);
	    wrbytes -= maplen;
	}

	/* If need to do a regular read/write, read the first chunk now */
	if (wrbytes) {
	    /* Lseek past what we mapped in */
	    if (maplen && lseek(rd_fd, (off_t)maplen, 0) != (off_t)maplen) {
		file_err("Warning: can't lseek on", path, name);
		fakeit = 1;
	    }
	    /* Read in one chunk */
	    n = min(wrbytes, IOBLKSIZ);
	    if (read_all(rd_fd, outbuf, n) == -1) {
		file_err("Warning: read error on", path, name);
		fakeit = 1;
	    } else {
		wrbytes -= n;
		/* Add it to the list of things to write out */
		add_wvec(outbuf, n);
	    }
	}
    }
} /* prep_afile */


/* DO_WRITE: Carry out the write.
 */
static void do_write(path, name, type, wr_fd, st)
    char *path, *name;
    int wr_fd, type;
    struct stat *st;
{
    int n;

    /* All the buffers should be full, so flush them now. */
    if (flush_wvec(wr_fd) == -1) {
	/* Guess that the write was actually due to a read during the mmap */
	if (maplen && (errno == EIO || errno == EFAULT)) {
	    file_err("This could be a read error, trying to skip",
	             path, name);
	    fakeit = 1;
	    wrbytes = st->st_size;
	} else {
	    file_err("Aborting on", path, name);
	    exit(1);
	}
    }

    /* If reading from a normal file */
    if (type == A_FILE || type == M_FILE) {
	/* Munmap if needed */
	if (maplen) {
	    if (munmap(mapadr, maplen) == -1) {
		perror("munmap");
		slave_abort();
	    }
	}

	/* Write whatever didn't get written from mmap'ed part */
	while (wrbytes) {
	    if (fakeit) {
		/* Write blocks of zeroes to fill up space */
		for (bzero(outbuf, IOBLKSIZ); wrbytes; wrbytes -= n) {
		    n = min(wrbytes, IOBLKSIZ);
		    add_wvec(outbuf, n);
		    if (flush_wvec(wr_fd) == -1) {
			file_err("Aborting on", path, name);
			exit(1);
		    }
		}
	    } else {
		/* Do read/write cycle */
		for (; wrbytes; wrbytes -= n) {
		    n = min(wrbytes, IOBLKSIZ);
		    if (read_all(rd_fd, outbuf, n) == -1) {
			file_err("Read error on", path, name);
			fakeit = 1;
			break;
		    }
		    add_wvec(outbuf, n);
		    if (flush_wvec(wr_fd) == -1) {
			file_err("Aborting on", path, name);
		    }
		}
	    }
	}
	if (rd_fd != -1) {
	    if (close(rd_fd) == -1) {
		file_err("Can't close", path, name);
		slave_abort();
	    }
	}
    }
} /* do_write */


/* ADD_WVEC: Add a buffer to write to the list of buffers to write.
 */
static void add_wvec(buf, size)
    char *buf;
    int size;
{
    outvecs[nvecs].iov_base = buf;
    outvecs[nvecs++].iov_len = size;
} /* add_wvec */


/* FLUSH_WVEC: Write out the write vectors.  Do multiple writes if needed.
 */
static int flush_wvec(fd)
    int fd;
{
    int res;
    struct iovec *iv = outvecs;

    while (nvecs) {
	/* Do a write */
	res = writev(fd, iv, nvecs);

	/* Return -1 on error.  Make sure we flush the vecs */
	if (res == -1 || res == 0) {
	    perror("write");
	    nvecs = 0;
	    return -1;
	}

	/* Skip over vecs that were written */
	while (res) {
	    if (iv->iov_len <= res) {
		res -= iv->iov_len;
		++iv;
		--nvecs;
	    } else {
		iv->iov_len -= res;
		res = 0;
	    }
	}
    }
    return 0;
} /* flush_wvec */


/* READ_ALL: Read all the bytes passed by issuing multiple read()'s.
 */
static int read_all(fd, buf, size)
    int fd, size;
    char *buf;
{
    int res;

    while (size) {
	if ((res = read(fd, buf, size)) == -1 || res == 0) {
	    return -1;
	}
	size -= res;
	buf += res;
    }
    return 0;
} /* read_all */


/* SLAVE_ABORT: Abort the slave proc.
 */
static void slave_abort()
{
    exit(1);
} /* slave_abort */


/* FILE_ERR: Report some error about a file.
 */
static void file_err(errmsg, path, name)
    char *errmsg, *path, *name;
{
    (void) fprintf(stderr, "%s %s/", errmsg, path);
    perror(name);
} /* file_err */


/* DO_CHDIR: Make sure we've cd'ed into the path given.
   Future Plans: change this to something that does a "maximal" chdir -- it
   does a chdir to either a relative pathname it creates or the absolute
   pathname, depending on which has the least number of dirs in it.
 */
#if NSLAVES > 0
static int do_chdir(path)
    char *path;
{
    static char curdir[MAXDIRLENGTH];

    if (!curdir[0] || strcmp(path, curdir)) {
	if (chdir(path) == -1) {
	    fputs("Can't chdir to ", stderr);
	    perror(path);
	    return -1;
	}
	(void) strcpy(curdir, path);
    }
    return 0;
} /* do_chdir */
#endif
